// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description: This file realizes the methods of the PBS_Accounting class
//               providing access to the PBS server's accounting records.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#include <PBS_Accounting.h>
#include <PBS_Server.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <dirent.h>


#ifdef __linux__ 
	#define DIRENT_SORT_PARM const void *
#else	
	#define DIRENT_SORT_PARM const struct dirent **
#endif

// *****************************************************************************
// * time_t scanTime :
// *	This value is used as the baseline for the scandirSelect function.  Any
// *	file that contains a date before this scanTime will be disregarded.
// *****************************************************************************
static time_t scanTime = 0;
	
// *****************************************************************************
// * stringToDuration :
// *    This method converts a time string to a time_t. The defaultValue 
// *    provided by the caller will be used in the event that the str is NULL
// *    or blank.
// *****************************************************************************
static time_t stringToDuration ( const char * str, time_t defaultValue )
        {
        time_t result=defaultValue;

        if(str && *str)
                {
                int multx = 1;
                for(const char *ptr=str; multx<3600 && (ptr=strchr(ptr, ':'))!=NULL; ptr++)
                        {
                        multx*=60;
                        }

                result    = 0;
                while(*str && multx)
                        {
                        result += atoi(str)*multx;
                        if(*(str+=strcspn(str, ":"))) str++;
                        multx/=60;
                        }
                }

        return result;
        }

// *****************************************************************************
// * scandirSelect :
// *	This function selects which files found in the dataPath will be read 
// * 	into the PBS_ServerDatabase.  Each PBS log file is named YYYYMMDD to 
// *	reflect the month day and year that it was opened.  This method ensures 
// *	that the filename complies with this format and that it is within a 
// * 	certain time period.
// *
// *	This method will only accept files that specify a date later than the
// *	specified scanTime.
// *****************************************************************************
static int scandirSelect ( const struct dirent * dir )
	{
	time_t           fileTime  = 0;
	int              result    = 0;
	struct tm        fileTm;
	
	if(sscanf(dir->d_name, "%04d%02d%02d", &fileTm.tm_year, &fileTm.tm_mon, &fileTm.tm_mday)==3 && !dir->d_name[8])
		{
		if(scanTime>0)
			{
			fileTm.tm_mon -= 1;
        	 	fileTm.tm_year-= 1900;
			fileTm.tm_sec  = 59;
			fileTm.tm_min  = 59;
			fileTm.tm_hour = 23;
			fileTime = mktime(&fileTm);
			if(fileTime>=scanTime) result = 1;
			}
		else result = 1;
		}

	return result;
	}
	
// *****************************************************************************
// * scandirCompare :
// *	This method compares the names of two log files for sorting.
// *****************************************************************************
static int scandirCompare ( DIRENT_SORT_PARM a, DIRENT_SORT_PARM b )
	{
	const struct dirent * A = *(const struct dirent **)a;
	const struct dirent * B = *(const struct dirent **)b;
	
	return atoi(B->d_name)-atoi(A->d_name);
	}

// *****************************************************************************
// * stringPointerCompare :
// *	This method compares two pointers to character strings.
// *****************************************************************************	
static int stringPointerCompare ( const void * ptr1, const void * ptr2 )
	{
	char * str1 = *(char **)ptr1;
	char * str2 = *(char **)ptr2;
	return strcmp(str1, str2);
	}

// *****************************************************************************
// * strndup :
// *	This method will allocate a buffer and copy the first n characters of
// *	the string identified by prt to a new string and NULL terminate it.
// *****************************************************************************
static char * strndup ( const char * ptr, size_t len )
	{
	char * result = NULL;

	if(ptr && *ptr)
		{
		if(len<=0) len = strlen(ptr);
		result = new char[len+1];
		strncpy(result, ptr, len);
		result[len] = 0;
		}
	return result;
	}

// *****************************************************************************
// * substrdup :
// *	This function will allocate a buffer and then copy the characters from 
// *	the string identified by ptr to the new string until it reaches one
// *	of the specified token characters.
// *****************************************************************************
static char * substrdup ( const char * ptr, const char * tokens )
	{
	char * result = NULL;
	size_t len;
	
	if(ptr && *ptr && (len=strcspn(ptr, tokens))!=0)
		{
		result = new char[len+1];
		strncpy(result, ptr, len);
		result[len] = 0;
		}
	return result;
	}


// *****************************************************************************
// * substrcpy :
// *	This function will copy all of the characters from the specified ptr
// *	until it reaches one of the specified tokens.
// *****************************************************************************
static char * substrcpy ( char * result, const char * ptr, const char * tokens )
	{
	if(result)
		{
		size_t len;
		*result=0;
		if(ptr && *ptr && (len=strcspn(ptr, tokens))!=0)
			{
			strncpy(result, ptr, len);
			result[len] = 0;
			}
		}
	return result;
	}

// *****************************************************************************
// * getNextLogFile :
// *	This function will evaluate the current log file name and will generate
// *	a new log file name that represents the following day. A time_t that
// *	corresponds to the file date will be returned to the caller.  If the
// *	time_t value is 0, then an error ocurred in processing.
// *****************************************************************************
static time_t getNextLogFile ( const char * filename, char * nextFile )
	{
	time_t result = 0;
	int    len    = 0;
	int    dfYear, dfMon, dfDay;
	
	if(nextFile && filename && *filename && (len=strlen(filename))>=8)
		{
		const char * fPtr = filename+len-8; 
	   	if(sscanf(fPtr, "%04d%02d%02d", &dfYear, &dfMon, &dfDay)==3)
			{
			struct tm t;
			struct tm * tPtr;
			t.tm_year = dfYear-1900;
			t.tm_mon  = dfMon-1;
			t.tm_mday = dfDay;
			t.tm_hour = 0;
			t.tm_min  = 0;
			t.tm_sec  = 0;
			result    = mktime(&t);
			result   += (24*60*60);
			tPtr      = localtime(&result);
			strcpy(nextFile, filename);
			sprintf(nextFile+len-8, "%04d%02d%02d", tPtr->tm_year+1900, tPtr->tm_mon+1, tPtr->tm_mday);
			}
		}
	return result;
	}


// #############################################################################
// # START COLLECTION OF METHODS ASSOCIATED WITH PBS_AccountingRecord CLASS.
// #############################################################################

// *****************************************************************************
// * PBS_AccountingRecord::now :
// *	This value represents the time that the startCycle method was called
// *	in the PBS_AccountingRecord.  This value is used to provide a consistent
// *	representation of the current time throughout the load process.
// *****************************************************************************
time_t PBS_AccountingRecord::now = 0;

// *****************************************************************************
// * PBS_AccontingRecord::PBS_AccountingRecord :
// *	This is the constructor for the PBS_AccountingRecord class.  It is 
// *	responsible for initializing all of the internal variables.
// *****************************************************************************
PBS_AccountingRecord::PBS_AccountingRecord ( void )
	: job(NULL),     user(NULL),  group(NULL), 
	  jobname(NULL), queue(NULL), exec_host(NULL),
	  nodeList(NULL), nodeAcct(NULL), userAcct(NULL), 
	  queueAcct(NULL)
	{
	clear();
	}
	
// *****************************************************************************
// * PBS_AccountingRecord::startCycle:
// *	The static startCycle method is called at the beginning of record
// *	processing to set the value of the static now variable.  The now
// *	variable is uysed to provide a consistent notion of the current time
// *	throughout record processing.
// *****************************************************************************
void PBS_AccountingRecord::startCycle ( void )
	{
	now = time(NULL);
	}
	
// *****************************************************************************
// * PBS_AccountingRecord::endCycle :
// *	The endCycle method is called at the end of the record processing cycle
// *	to clear the static now variable.
// *****************************************************************************
void PBS_AccountingRecord::endCycle ( void )
	{
	now = 0;
	}
	
// *****************************************************************************
// * PBS_AccountingRecord::create :
// *	This is a factory method for the PBS_AccountingRecord object. It 
// *	provides a mechanism to sense when object construction has failed and
// *	return a NULL to the caller rather than throwing an exception.
// *****************************************************************************
PBS_AccountingRecord * PBS_AccountingRecord::create ( 
	PBS_AccountingDatabase & db, 
	const char * buf )
	{
	static PBS_AccountingRecord * rec    = new PBS_AccountingRecord();
	PBS_AccountingRecord        * result = NULL;
	
	if(rec->scanRecord(db, buf)==0)
		{
		result = rec;
		rec    = new PBS_AccountingRecord();
		}
	return result;	
	}		


// *****************************************************************************
// * PBS_AccountingRecord::PBS_AccountingRecord :
// *	This is the destructor for the PBS_AccountingRecord class.  It will
// *	call the clear method to delete all of the internally allocated 
// *	variables and remove the object5 from any lists where it may be stored
// *	in the PBS_AccountingDatabase.
// *****************************************************************************
PBS_AccountingRecord::~PBS_AccountingRecord ( void )
	{
	clear();
	}


// *****************************************************************************
// * PBS_AccountingRecord::clear :
// *	This methdo will remove the entry from the queue and user lists that
// *	are stored in the PBS_AccountingDatabase and then will delete all of the
// *	internal variables that are associated with the object. Typically,
// *	this method will be called prior to destruction.
// *****************************************************************************
void PBS_AccountingRecord::clear ( void )
	{
	int idx;
	
	if(nodeList)
		{
		for(int idx=0; nodeList[idx]!=NULL && idx<nodeCnt; idx++)
			{
			free(nodeList[idx]);
			}
		delete [] nodeList;
		}
	if(nodeAcct) 
		{
		for(int idx=0; nodeAcct[idx]!=NULL; idx++)
			{
			nodeAcct[idx]->detachRecord(this);
			nodeAcct[idx] = NULL;
			}
		delete[] nodeAcct;
		}
	if(userAcct)  userAcct->detachRecord(this);
	if(queueAcct) queueAcct->detachRecord(this);

	
	nullify(job);
	nullify(user);
	nullify(group);
	nullify(jobname);
	nullify(queue);
	nullify(exec_host);
	nodeAcct    = NULL;
	userAcct    = NULL;
	queueAcct   = NULL;
	nodeList    = NULL;
	ctime       = 0;
	qtime       = 0;
	etime       = 0;
	start       = 0;
	end         = 0;
	exit_status = 0;
	cputime     = 0;
	walltime    = 0;
	mem         = 0;
	vmem        = 0;
	procCnt     = 0;
	jobID       = 0;
	}


// *****************************************************************************
// * PBS_AccountingRecord::nullify :
// *	This method will delete a previously allocated pointer and then will
// *	set the pointer to NULL to prevent subsequent access.
// *****************************************************************************
void PBS_AccountingRecord::nullify ( char * &ptr )
	{
	if(ptr) delete[] ptr;
	ptr = NULL;
	}


// *****************************************************************************
// * PBS_AccountingRecord::scanRecord :
// *	This method is called by the PBS_AccountingDatabase to scan the
// *	components of a character string and extract the critical data elements.
// *	If sufficient data elements were read to create a valid record, then
// *	0 will be returned.  Otherwise, a value of -1 will be returned and the
// *	data object will be cleared.
// *****************************************************************************
int PBS_AccountingRecord::scanRecord ( PBS_AccountingDatabase & db, const char * buf )
	{
	int result = -1;
	
	clear();
	
	if(now==0) startCycle();
	
	if(buf && *buf && strlen(buf)>22 && buf[20]=='S' || buf[20]=='E')
		{
		char * ptr, * ePtr;
		int    hour, min, sec;
		size_t aIdx;
		
		job = substrdup(buf+22, "; \t\n");
		if((jobID = atoi(job)) <= 0) jobID = 0;
		
		if((ptr = strstr(buf, "user="))!=NULL)
			user = substrdup(ptr+5, " \t\n");
		if((ptr = strstr(buf, "group="))!=NULL)
			group = substrdup(ptr+6, " \t\n");
		if((ptr = strstr(buf, "jobname="))!=NULL)
			jobname = substrdup(ptr+8, " \t\n");
		if((ptr = strstr(buf, "queue="))!=NULL)
			queue = substrdup(ptr+6, " \t\n");
		if((ptr = strstr(buf, "exec_host="))!=NULL)
			exec_host = substrdup(ptr+10, " \t\n");
		if((ptr = strstr(buf, "ctime="))!=NULL)
			ctime = atoi(ptr+6);
		if((ptr = strstr(buf, "qtime="))!=NULL)
			qtime = atoi(ptr+6);
		if((ptr = strstr(buf, "etime="))!=NULL)
			etime = atoi(ptr+6);
		if((ptr = strstr(buf, "start="))!=NULL)
			start = atoi(ptr+6);
		if((ptr = strstr(buf, "end="))!=NULL)
			end = atoi(ptr+4);
		if((ptr = strstr(buf, "Exit_status="))!=NULL)
			exit_status = atoi(ptr+12);
		if((ptr = strstr(buf, "resources_used.cput="))!=NULL)
			{
			sscanf(ptr+20, "%d:%d:%d", &hour, &min, &sec);
			cputime = hour*3600 + min*60 + sec;
			if(cputime<0) cputime = 0;
			}
		if((ptr = strstr(buf, "resources_used.walltime="))!=NULL)
			{
			sscanf(ptr+24, "%d:%d:%d", &hour, &min, &sec);
			walltime = hour*3600 + min*60 + sec;
			}
		if((ptr = strstr(buf, "resources_used.mem="))!=NULL)
			{
			mem = strtoul(ptr+19, &ePtr, 10);
			if(toupper(*ePtr)=='K') mem*=(1024);
			if(toupper(*ePtr)=='M') mem*=(1024*1024);
			if(toupper(*ePtr)=='G') mem*=(1024*1024*1024);
			}
		if((ptr = strstr(buf, "resources_used.vmem="))!=NULL)
			{
			vmem = strtoul(ptr+20, &ePtr, 10);
			if(toupper(*ePtr)=='K') vmem*=(1024);
			if(toupper(*ePtr)=='M') vmem*=(1024*1024);
			if(toupper(*ePtr)=='G') vmem*=(1024*1024*1024);
			}
		
		if(walltime <= 0)
			{
			time_t endTime = end>=start?end:now;
			walltime = endTime-start;
			}

		if(exec_host!=NULL)
			{
			for(procCnt=1, ptr=exec_host; *ptr; ptr++)
				if(*ptr=='+') procCnt++;
			
			PBSStringHash nodeNames(procCnt*2);
			
			char * tBuf = strdup(exec_host);
			char * nPtr = tBuf;
			char * ePtr = tBuf;
			aIdx = 1;
			
			while(*ePtr)
				{
				nPtr = ePtr;
				while(*ePtr && *ePtr!='+' && *ePtr!='/') ePtr++;
				if(*ePtr=='/') *(ePtr++) = 0;
				while(*ePtr && *ePtr!='+') ePtr++;
				if(*ePtr=='+') *(ePtr++) = 0;

				if(nodeNames.find(nPtr)==NULL)
					{
					nodeNames.insert(nPtr, (void *)aIdx++);
					}
				}
			
			if((nodeCnt = aIdx-1) > 0)
				{
				nodeList = new char * [nodeCnt+1];
				memset(nodeList, 0, (nodeCnt+1)*sizeof(char *));
				
				PBSStringHashIter nIter(nodeNames);
				for(nIter.first(); nIter.key() && (aIdx=(int)nIter.data())!=0; nIter++)
					{
					if((--aIdx)<nodeCnt) nodeList[aIdx] = strdup(nIter.key());
					}
				for(nodeCnt=0; nodeList[nodeCnt]!=NULL; nodeCnt++);
				}
			
			free(tBuf);
			}

		if(jobID && user && queue ) 
			{
			if((userAcct  = db.findUser(user))!=NULL)
				userAcct->attachRecord(this);
			if((queueAcct = db.findQueue(queue))!=NULL)
				queueAcct->attachRecord(this);
			if(nodeCnt>0)
				{
				size_t nIdx = 0, aIdx = 0;
				nodeAcct = new PBS_NodeAccounting * [nodeCnt+1];
				memset(nodeAcct, 0, (nodeCnt+1)*sizeof(PBS_Accounting *));

				for(nIdx=0, aIdx=0; nIdx<nodeCnt; nIdx++)
					if((nodeAcct[aIdx] = db.findNode(nodeList[nIdx]))!=NULL)
						nodeAcct[aIdx++]->attachRecord(this);
				}
			result = 0;
			}
		else 	{
			clear();
			result = -1;
			}
		}
	return result;
	}


// *****************************************************************************
// * PBS_AccountingRecord::updateRecord :
// *	This method is called to update the volatile components of the 
// *	PBS_AccountingRecord... These are: end time, cpu time, wall time, 
// *	physical memory and virtual memory.  This is a protected function and
// *	is only called by the PBS_AccountingDatabase when it is being updated
// *	externally.
// *****************************************************************************
void PBS_AccountingRecord::updateRecord ( time_t endTime, time_t cpuTime, time_t wallTime )
	{
	end      = (endTime>start)?endTime:0;
	cputime  = (cpuTime>0)?cpuTime:0;
	walltime = (wallTime>0)?wallTime:((end>=start)?(end-start):(now-start));
	}


// *****************************************************************************
// * PBS_AccountingRecord::getNodeProcCnt :
// *	This method will return the number of processors from a specified node
// *	that are used by a job.
// *****************************************************************************
size_t PBS_AccountingRecord::getNodeProcCnt ( const char * node )
	{
	size_t procCnt = 0;
	char * ptr     = exec_host;
	char   buf[128];

	sprintf(buf, "%s/", node);
	while((ptr = strstr(ptr, buf))!=NULL)
		{
		procCnt++;
		ptr++;
		}
	
	return procCnt;
	}
		
	
	
// *****************************************************************************
// * PBS_AccountingRecord::asciiDump :
// *	This method generates an ascii report of the contents of the 
// *	PBS_AccountingRecord object.
// *****************************************************************************
void PBS_AccountingRecord::asciiDump ( FILE * fp , char * prefix )
	{
	char c = '\0';
	if(prefix==NULL)  prefix=&c;
	fprintf(fp, "%sPBS Accounting Record for Job : %s\n", prefix, job?job:"unknown");
	fprintf(fp, "%s    - User            : %s\n", prefix, user?user:"unknown");
	fprintf(fp, "%s    - Group           : %s\n", prefix, group?group:"unknown");
	fprintf(fp, "%s    - Job Name        : %s\n", prefix, jobname?jobname:"unknown");
	fprintf(fp, "%s    - Queue           : %s\n", prefix, queue?queue:"unknown");
	fprintf(fp, "%s    - Create Time     : %s", prefix, ::ctime(&ctime));
	fprintf(fp, "%s    - Queue Time      : %s", prefix, ::ctime(&qtime));
	fprintf(fp, "%s    - Eligible Time   : %s", prefix, ::ctime(&etime));
	fprintf(fp, "%s    - Start Time      : %s", prefix, ::ctime(&start));
	fprintf(fp, "%s    - End Time        : %s", prefix, end?::ctime(&end):"unknown\n");
	fprintf(fp, "%s    - Exit Status     : %i\n", prefix, exit_status);
	fprintf(fp, "%s    - CPU Time        : %d:%02d:%02d\n", prefix, cputime/3600, (cputime%3600)/60, cputime%60); 
	fprintf(fp, "%s    - Wall Time       : %d:%02d:%02d\n", prefix, walltime/3600, (walltime%3600)/60, walltime%60); 
	fprintf(fp, "%s    - Physical Memory : %d Kb\n", prefix, mem/1024);
	fprintf(fp, "%s    - Virtual Memory  : %d Kb\n", prefix, vmem/1024);
	fprintf(fp, "%s    - Processor Count : %d\n", prefix, procCnt);
	fprintf(fp, "%s    - Processors      : %s\n", prefix, exec_host);
	}
	
// #############################################################################
// # START COLLECTION OF METHODS ASSOCIATED WITH PBS_Accounting CLASS.
// #############################################################################

// *****************************************************************************
// * PBS_Accounting::PBS_Accounting :
// *	This is the constructor for the PBS_Accounting object.  It will 
// *	preallocate an array of PBS_AccountingRecord pointers that will be 
// *	populated as new record are added.
// *****************************************************************************
PBS_Accounting::PBS_Accounting ( const char * Name )
	: name       ((Name && *Name)?strdup(Name):NULL),
	  maxRecs    (16),
	  numRecs    (0),
	  recs       ((PBS_AccountingRecord **)malloc(16*sizeof(PBS_AccountingRecord *)))
	{
	}
	

// *****************************************************************************
// * PBS_Accounting::~PBS_Accounting:
// *	This is the destructor for the PBS_Accounting object.  It will delete 
// *	the object name and the ARRAY used to store the PBS_AccountingRecord
// *	objects.  The actual PBS_AccountingRecord objects are deleted from
// *	within the PBS_AccountingDatabase objects destructor.
// *****************************************************************************
PBS_Accounting::~PBS_Accounting ( void )
	{
	if(name) free(name);
	if(recs) free(recs);
	}
	

// *****************************************************************************
// * PBS_Accounting::operator [] :
// *	This is the indexing operator for the PBS_Accounting object. It will
// *	return a pointer to the PBS_AccountingRecord associated with the
// *	index - or NULL if the index is beyond the end of the list.
// *****************************************************************************
PBS_AccountingRecord * PBS_Accounting::operator [] ( size_t idx )
	{
	return ((idx+1)>numRecs)?NULL:recs[idx];
	}

// *****************************************************************************
// * PBS_Accounting::getIndex :
// *	This method will locate the PBS_AccountingRecord associated with the 
// *	specified job name.  If no matching job is found, then the index of
// *	the location where the job should be inserted will be returned.  If
// *	the job name is invalid (NULL or doesn't convert to a positive
// *	integer) then -1 will be returned.
// *****************************************************************************
ssize_t PBS_Accounting::getIndex ( const char * job )
	{
	ssize_t idx  = -1;
	int    jobID = (job && *job)?atoi(job):0;
	
	if(jobID<=0) 
		{
		idx = -1;
		}
	else if(numRecs<16)
		{
		for(idx=0; 
		    idx<numRecs && recs[idx] && jobID>recs[idx]->getJobID(); 
		    idx++);
		}
	else	{
		int    sDirection = 1;
		size_t sFactor    = numRecs-1;
		size_t sIdx       = 0;
		
		while(idx<0)
			{
			if(jobID<recs[sIdx]->getJobID())
				{
				sDirection = -1;
				if(sIdx==0) idx=0;
				else if(jobID>recs[sIdx-1]->getJobID()) idx=sIdx;
				}
			else if (jobID>recs[sIdx]->getJobID())
				{
				sDirection = 1;
				if(sIdx==numRecs-1) idx=numRecs;
				}
			else idx = sIdx;

			int newIdx = sIdx+sDirection*sFactor;
			if(newIdx>(int)(numRecs-1))
				{
				newIdx = numRecs-1;
				sFactor = newIdx-sIdx;
				}
			else if(newIdx<0)
				{
				newIdx  = 0;
				sFactor = sIdx;
				}
			sIdx = newIdx;
			sFactor = (sFactor/2)+(sFactor%2);
			}
		}
	
	return idx;
	}

// *****************************************************************************
// * PBS_Accounting::find :
// *	This method will return the PBS_AccountingRecord associated with the
// *	specified job name.  If no matching record is found then NULL will be
// *	returned.
// *****************************************************************************
PBS_AccountingRecord * PBS_Accounting::find ( const char * job )
	{
	ssize_t idx = getIndex(job);
	return (idx>=0 && idx<numRecs && !strcmp(recs[idx]->getJob(), job))?recs[idx]:NULL; 
	}

// *****************************************************************************
// * PBS_AccountingRecord::attachRecord :
// *	This method allows the caller to insert an existing PBS_AccountingRecord
// *	object into the PBS_Accounting array.  If a record with a matching name
// *	is found, then it will be removed from the list and returned to the
// *	caller for deletion. In the event that this record is invalid or cannot
// *	be inserted, then a pointer to the record will be returned to the caller
// *	for deletion.  In either event, the user shoudl delete the 
// *	PBS_Accountingrecord that is returned from this function and should not
// *	subsequently access the pointer that was provided to the method.
// *****************************************************************************
PBS_AccountingRecord * PBS_Accounting::attachRecord ( PBS_AccountingRecord * rec )
	{
	PBS_AccountingRecord * oldRec = NULL;
	if(rec)
		{
		ssize_t idx = 0;
		
		if(numRecs>=maxRecs)
			{
			maxRecs *= 2;
			recs = (PBS_AccountingRecord **)realloc(recs, maxRecs*sizeof(PBS_AccountingRecord *));
			}
		if((idx = getIndex(rec->getJob()))>=0)
			{
			if(idx>=numRecs)
				{
				idx = numRecs;
				recs[numRecs++] = rec;
				}
			else if(rec->getJobID()==recs[idx]->getJobID())
				{
				oldRec    = recs[idx];
				recs[idx] = rec;
				}
			else	{
				memmove(&recs[idx+1], &recs[idx], (numRecs-idx)*sizeof(PBS_AccountingRecord *));
				recs[idx] = rec;
				numRecs++;
				}
			}
		else oldRec = rec;
		}

	return oldRec;
	}
	

// *****************************************************************************
// * PBS_Accounting::detachRecord :
// *	This method will locate teh specified record in the array and will 
// *	remove it from the list.  If the record was successfully located and
// *	removed, then the method will return a pointer to it for deletion.  If
// *	the record does not exist in this PBS_Accounting object, then NULL
// *	will be returned.
// *****************************************************************************
PBS_AccountingRecord * PBS_Accounting::detachRecord ( PBS_AccountingRecord * rec )
	{
	PBS_AccountingRecord * oldRec = NULL;
	if (rec)
		{
		ssize_t idx = getIndex(rec->getJob());
		if(idx>=0 && idx<numRecs && rec==recs[idx])
			{
			oldRec = recs[idx];
			memmove(&recs[idx], &recs[idx+1], (--numRecs-idx)*sizeof(PBS_AccountingRecord *));
			recs[numRecs] = NULL;
			}
		}
	return oldRec;
	}
	
// *****************************************************************************
// * PBS_Accounting::asciiDump :
// *	This method will generate a comprehensive dump of all of the 
// *	PBS_AccountingRecord objects that are stored in this class.
// *****************************************************************************
void PBS_Accounting::asciiDump (FILE * fp, char * prefix )
	{
	char c = 0;
	if(!prefix) prefix = &c;
	fprintf(fp, "%s\n%sJob Report for %s:\n", prefix, prefix, name?name:"Unknown");
	fprintf(fp, "%s---------------------------------------------------\n", prefix);
	
	char startBuf[24];
	char endBuf  [24];
	time_t start;
	time_t end;
	time_t cpu;
	time_t wall;  
	
	for(size_t idx=0; idx<numRecs; idx++)
		{
		start = recs[idx]->getStartTime();
		end   = recs[idx]->getEndTime();
		cpu   = recs[idx]->getCpuTime();
		wall  = recs[idx]->getWallTime();
		strftime(startBuf, 24, "%m/%d/%Y %H:%M", localtime(&start));
		if(end>=start) strftime(endBuf, 24, "%m/%d/%Y %H:%M", localtime(&end));
		else strcpy(endBuf, "Unknown");
		
		fprintf(fp, "%s%16s - Procs: %-3d  Start: %-16s  End: %-16s  CPU: [%3d:%02d:%02d]  Wall: [%3d:%02d:%02d]\n", 
			prefix, recs[idx]->getJob(), recs[idx]->getProcCnt(),
			startBuf, endBuf,
			cpu/3600, (cpu%3600)/60, cpu%60,
			wall/3600, (wall%3600)/60, wall%60);
		}
	}

// *****************************************************************************
// * PBS_AccountingDatabase::summaryDump :
// *	This method writes summary information about the PBS_Accounting object
// *	to the specified FILE.  The recType provided to this method identifies
// *	this objects type: QUEUE or USER.
// *****************************************************************************
void PBS_Accounting::summaryDump ( const char * recType, FILE * fp, char * prefix )
	{
	PBS_AccountingRecord * entry;
	size_t total         = 0;
	time_t totalCpu      = 0;
	time_t totalWall     = 0;
	time_t totalQueue    = 0;
	size_t serial        = 0;
	time_t serialCpu     = 0;
	time_t serialWall    = 0;
	time_t serialQueue   = 0;
	size_t parallel      = 0;
	time_t parallelCpu   = 0;
	time_t parallelWall  = 0;
	size_t parallelProcs = 0;
	time_t parallelQueue = 0;

	for(size_t idx=0; idx<numRecs && (entry = recs[idx])!=NULL; idx++)
		{
		if(entry->getEligibleTime()>0 && 
		   entry->getStartTime()>entry->getEligibleTime())
		   	{
			time_t qTime = entry->getStartTime()-entry->getEligibleTime();
			
			total++;
			totalCpu   += entry->getCpuTime();
			totalWall  += entry->getWallTime();
			totalQueue += qTime; 
			if(entry->getProcCnt()>1)
				{
				parallel++;
				parallelProcs += entry->getProcCnt();
				parallelCpu   += entry->getCpuTime();
				parallelWall  += entry->getWallTime();
				parallelQueue += qTime;
				}
			else	{
				serial++;
				serialCpu   += entry->getCpuTime();
				serialWall  += entry->getWallTime();
				serialQueue += qTime;
				}
			}
		}
			
	fprintf(fp, "\n%s  %s : %s\n", prefix, recType, name);
	fprintf(fp, "%s---------------------------------------------------\n", prefix);
	fprintf(fp, "%s          TOTAL               : % 8d jobs\n", prefix, total);
	fprintf(fp, "%s                CPU  Time     : %8.2f hours\n", prefix, (double)totalCpu/3600.0);
	fprintf(fp, "%s                CPU  Time/Job : %8.2f hours\n", prefix, (double)totalCpu/(3600.0*(double)(total?total:1)));
	fprintf(fp, "%s                Wall Time     : %8.2f hours\n", prefix, (double)totalWall/3600.0);
	fprintf(fp, "%s                Wall Time/Job : %8.2f hours\n", prefix, (double)totalWall/(3600.0*(double)(total?total:1)));
	fprintf(fp, "%s                Wait Time     : %8.2f hours\n", prefix, (double)totalQueue/3600.0);
	fprintf(fp, "%s                Wait Time/Job : %8.2f hours\n", prefix, (double)totalQueue/(3600.0*(double)(total?total:1)));
	fprintf(fp, "%s          PARALLEL            : % 8d jobs\n", prefix, parallel);       
	fprintf(fp, "%s                Average Procs : %8.2f procesors\n", prefix, (double)parallelProcs/(double)(parallel?parallel:1));
	fprintf(fp, "%s                CPU  Time     : %8.2f hours\n", prefix, (double)parallelCpu/3600.0);
	fprintf(fp, "%s                CPU  Time/Job : %8.2f hours\n", prefix, (double)parallelCpu/(3600.0*(double)(parallel?parallel:1)));
	fprintf(fp, "%s                Wall Time     : %8.2f hours\n", prefix, (double)parallelWall/3600.0);
	fprintf(fp, "%s                Wall Time/Job : %8.2f hours\n", prefix, (double)parallelWall/(3600.0*(double)(parallel?parallel:1)));
	fprintf(fp, "%s                Wait Time     : %8.2f hours\n", prefix, (double)parallelQueue/3600.0);
	fprintf(fp, "%s                Wait Time/Job : %8.2f hours\n", prefix, (double)parallelQueue/(3600.0*(double)(parallel?parallel:1)));
	fprintf(fp, "%s          SEQUENTIAL          : % 8d jobs\n", prefix, serial);
	fprintf(fp, "%s                CPU  Time     : %8.2f hours\n", prefix, (double)serialCpu/3600.0);
	fprintf(fp, "%s                CPU  Time/Job : %8.2f hours\n", prefix, (double)serialCpu/(3600.0*(double)(serial?serial:1)));
	fprintf(fp, "%s                Wall Time     : %8.2f hours\n", prefix, (double)serialWall/3600.0);
	fprintf(fp, "%s                Wall Time/Job : %8.2f hours\n", prefix, (double)serialWall/(3600.0*(double)(serial?serial:1)));
	fprintf(fp, "%s                Wait Time     : %8.2f hours\n", prefix, (double)serialQueue/3600.0);
	fprintf(fp, "%s                Wait Time/Job : %8.2f hours\n", prefix, (double)serialQueue/(3600.0*(double)(serial?serial:1)));
	}

	
// *****************************************************************************
// * PBS_NodeAccounting::asciiDump :
// *	This method will generate a comprehensive dump of all of the 
// *	PBS_AccountingRecord objects that are stored in this node's class.
// *****************************************************************************
void PBS_NodeAccounting::asciiDump (FILE * fp, char * prefix )
	{
	char c = 0;
	if(!prefix) prefix = &c;
	fprintf(fp, "%s\n%sJob Accounting for Node: %s\n", prefix, prefix, name?name:"Unknown");
	fprintf(fp, "%s---------------------------------------------------\n", prefix);
	
	char startBuf[24];
	char endBuf  [24];
	time_t start;
	time_t end;
	time_t cpu;
	time_t wall;  
	
	for(size_t idx=0; idx<numRecs; idx++)
		{
		start = recs[idx]->getStartTime();
		end   = recs[idx]->getEndTime();
		wall  = recs[idx]->getWallTime();
		strftime(startBuf, 24, "%m/%d/%Y %H:%M", localtime(&start));
		if(end>=start) strftime(endBuf, 24, "%m/%d/%Y %H:%M", localtime(&end));
		else strcpy(endBuf, "Unknown");
		
		// *************************************************************
		// * Count the number of processors on this system that are
		// * actually used by the job.
		// *************************************************************
		int   procCnt   = 0;
		const char *ptr = recs[idx]->getProcs();
		char  procStr[128];
		sprintf(procStr, "%s/", name);
		
		while((ptr = strstr(ptr, procStr))!=NULL)
			{
			ptr++;
			procCnt++;
			} 

		fprintf(fp, "%s%16s - Procs: %-3d  Start: %-16s  End: %-16s  Wall: [%3d:%02d:%02d]\n", 
			prefix, recs[idx]->getJob(), procCnt,
			startBuf, endBuf,
			wall/3600, (wall%3600)/60, wall%60);
		}
	}

// *****************************************************************************
// * PBS_UserAccounting::asciiDump :
// *	This method will generate a comprehensive dump of all of the 
// *	PBS_AccountingRecord objects that are stored in this class.
// *****************************************************************************
void PBS_UserAccounting::asciiDump (FILE * fp, char * prefix )
	{
	char c = 0;
	if(!prefix) prefix = &c;
	fprintf(fp, "%s\n%sJob Accounting for User: %s\n", prefix, prefix, name?name:"Unknown");
	fprintf(fp, "%s---------------------------------------------------\n", prefix);
	
	char startBuf[24];
	char endBuf  [24];
	time_t start;
	time_t end;
	time_t cpu;
	time_t wall;  
	
	for(size_t idx=0; idx<numRecs; idx++)
		{
		start = recs[idx]->getStartTime();
		end   = recs[idx]->getEndTime();
		cpu   = recs[idx]->getCpuTime();
		wall  = recs[idx]->getWallTime();
		strftime(startBuf, 24, "%m/%d/%Y %H:%M", localtime(&start));
		if(end>=start) strftime(endBuf, 24, "%m/%d/%Y %H:%M", localtime(&end));
		else strcpy(endBuf, "Unknown");
		
		fprintf(fp, "%s%16s - Procs: %-3d  Start: %-16s  End: %-16s  CPU: [%3d:%02d:%02d]  Wall: [%3d:%02d:%02d]\n", 
			prefix, recs[idx]->getJob(), recs[idx]->getProcCnt(),
			startBuf, endBuf,
			cpu/3600, (cpu%3600)/60, cpu%60,
			wall/3600, (wall%3600)/60, wall%60);
		}
	}

// *****************************************************************************
// * PBS_QueueAccounting::asciiDump :
// *	This method will generate a comprehensive dump of all of the 
// *	PBS_AccountingRecord objects that are stored in this class.
// *****************************************************************************
void PBS_QueueAccounting::asciiDump (FILE * fp, char * prefix )
	{
	char c = 0;
	if(!prefix) prefix = &c;
	fprintf(fp, "%s\n%sJob Accounting for Queue: %s\n", prefix, prefix, name?name:"Unknown");
	fprintf(fp, "%s---------------------------------------------------\n", prefix);
	
	char startBuf[24];
	char endBuf  [24];
	time_t start;
	time_t end;
	time_t cpu;
	time_t wall;  
	
	for(size_t idx=0; idx<numRecs; idx++)
		{
		start = recs[idx]->getStartTime();
		end   = recs[idx]->getEndTime();
		cpu   = recs[idx]->getCpuTime();
		wall  = recs[idx]->getWallTime();
		strftime(startBuf, 24, "%m/%d/%Y %H:%M", localtime(&start));
		if(end>=start) strftime(endBuf, 24, "%m/%d/%Y %H:%M", localtime(&end));
		else strcpy(endBuf, "Unknown");
		
		fprintf(fp, "%s%16s - Procs: %-3d  Start: %-16s  End: %-16s  CPU: [%3d:%02d:%02d]  Wall: [%3d:%02d:%02d]\n", 
			prefix, recs[idx]->getJob(), recs[idx]->getProcCnt(),
			startBuf, endBuf,
			cpu/3600, (cpu%3600)/60, cpu%60,
			wall/3600, (wall%3600)/60, wall%60);
		}
	}


// #############################################################################
// # START COLLECTION OF METHODS ASSOCIATED WITH PBS_AccountingDatabase CLASS.
// #############################################################################

// *****************************************************************************
// * PBS_AccountingDatabase::PBS_AccountingDatabase :
// *	This is the constructor for the PBS_AccountingDatabase object.  It will
// *	initialize all of the internal values and then will process all of the
// *	accounting files from the specified DataPath.  Record that are older
// *	than the daysOld parameter will not be processed.
// *****************************************************************************
PBS_AccountingDatabase::PBS_AccountingDatabase ( 
	const char * Name, 
	time_t       DaysOld, 
	time_t       MaxWallHours, 
	const char * DataPath )
	: PBS_Accounting (Name), 
	  daysOld        (DaysOld), 
	  maxWallTime    (MaxWallHours*60*60),
	  dataPath       (NULL), 
	  dataFileOffset (0)
	{
	*dataFile = 0;
	if(DataPath && *DataPath) dataPath = strdup(DataPath);
	else dataPath = strdup("/usr/spool/PBS/server_priv/accounting");
	initialize();
	}

// *****************************************************************************
// * PBS_AccountingDatabase::~PBS_AccountingDatabase :
// *	This is the destructor for the PBS_AccountingDatabase object.  It will
// *	call clear to remove all PBS_AccountingRecord objects and clear their
// *	references in the queue and user lists.  Additionally, the memory
// *	allocated to the name string, dataPath string and recs array will also
// *	be deleted.
// *****************************************************************************
PBS_AccountingDatabase::~PBS_AccountingDatabase ( void )
	{
	clear();	
	
	if(name)
		{
		free(name);
		name = NULL;
		}
	if(recs)
		{
		free(recs);
		recs = NULL;
		}
	if(dataPath)
		{
		free(dataPath);
		dataPath = NULL;
		}
	}

// *****************************************************************************
// * PBS_AccountingDatabase::clear :
// *	This method will remove all previously stored PBS_AccountingRecord
// *	objects from the database and restore the PBS_AccountingDatabase object
// *	toa pristine condition - ready to reload data.
// *****************************************************************************
void PBS_AccountingDatabase::clear ( void )
	{
	PBS_Accounting  * rec = NULL;
	PBSStringHashIter nIter(nodes);
	PBSStringHashIter uIter(users);
	PBSStringHashIter qIter(queues);
	
	while(numRecs>0)
		{
		if(recs[--numRecs]!=NULL) delete recs[numRecs];
		}
	
	while((rec=(PBS_Accounting *)nIter.first())!=NULL)
		{
		nodes.remove(nIter.key());
		delete rec;
		}
	while((rec=(PBS_Accounting *)uIter.first())!=NULL)
		{
		users.remove(uIter.key());
		delete rec;
		}
	while((rec=(PBS_Accounting *)qIter.first())!=NULL) 
		{
		queues.remove(qIter.key());
		delete rec;
		}
	}


// *****************************************************************************
// * PBS_AccountingDatabase::initialize :
// *	This method is called to populate the PBS_AccountingDatabase using the
// *	accounting files that are stored on disk.  It is a protected member 
// *	function and is only called by the constructor and the update method.
// *****************************************************************************
void PBS_AccountingDatabase::initialize ( void )
	{
	// *********************************************************************
	// * Call PBS_AccountingRecord::startCycle to obtain the current time.
	// *********************************************************************
	PBS_AccountingRecord::startCycle();
	
	// *********************************************************************
	// * If there are currently records stored in this database, ensure
	// * that they are cleared before reloading.
	// *********************************************************************
	if(numRecs>0) clear();
	
	// *********************************************************************
	// * Set the scanTime variable to specify the number of days of history
	// * that should be loaded into the database.
	// *********************************************************************
	scanTime = time(NULL)-(daysOld*24*60*60);
	
	// *********************************************************************
	// * This is an array of directory entries for the matching files in
	// * the dataPath directory.
	// *********************************************************************
	struct dirent **dataFiles = NULL;
	
	// *********************************************************************
	// * A four kilobyte buffer will be used to read the record from the
	// * data files.  In the future, this size may not be sufficient so
	// * it is possible that this should be a compiled definition in the
	// * future.
	// *********************************************************************
	char buf[4096];
	
	// *********************************************************************
	// * First collect the names of all of the files that are in the 
	// * dataPath directory. Files are included or excluded by the 
	// * scandirSelect function.
	// *********************************************************************
	int filecnt = scandir(dataPath, &dataFiles, scandirSelect, scandirCompare);

	// *********************************************************************
	// * Walk through each of the selected files and call the readFile 
	// * method to assimilate their contents.
	// *********************************************************************
	while(filecnt>0)
		{
		FILE * fp = NULL;
		
		sprintf(dataFile, "%s/%s", dataPath, dataFiles[--filecnt]->d_name);
		dataFileOffset = 0;
				
		if((fp = fopen(dataFile, "r"))!=NULL)
			{
			while(fgets(buf, 4096, fp)!=NULL) scanRecord(buf);
			dataFileOffset = ftell(fp);
			fclose  (fp);
			}
		else LogError("PBS_Accounting", "Failed to open %s", dataFile);

		// *************************************************************
		// * Call free here to release the dirent structure that was
		// * allocated by the scandir function...
		// *************************************************************
		free(dataFiles[filecnt]);
		}
	// *********************************************************************
	// * Call free here to release the array of dirent pointers that was
	// * allocated by the scandir function.
	// *********************************************************************
	free(dataFiles);
	
	// *********************************************************************
	// * Call purge here to remove any jobs with invalid start/end time
	// * values that will skew the calculations.
	// *********************************************************************
	purge();
	
	// *********************************************************************
	// * Call PBS_AccountingRecord::endCycle to clear the current time.
	// *********************************************************************
	PBS_AccountingRecord::endCycle();
	}

// *****************************************************************************
// * PBS_AccountingDatabase::update :
// *	This method is called to read any records that have been logged since
// *	initialization or since the last update.  If the date has changed, then
// *	the current accounting file will be closed and the next file will be
// *	opened and processed.
// *****************************************************************************
void PBS_AccountingDatabase::update ( void )
	{
	FILE * fp  = NULL;
	
	if(*dataFile)
		{
		char   buf      [4096];
		char   nextFile [MAXPATHLEN];
		time_t now    = time(NULL);
		time_t then   = 0;
		
		PBS_AccountingRecord::startCycle();
	
		do	{
			if((fp=fopen(dataFile, "r"))!=NULL)
				{
				if(dataFileOffset>0) fseek(fp, dataFileOffset, SEEK_SET);
				while(fgets(buf, 4096, fp)!=NULL) scanRecord(buf);
				dataFileOffset = (int)ftell(fp);
				fclose(fp);
				}
			else if(!access(dataFile, F_OK))
				{
				LogError("PBS_Accounting", "Failed to open %s", dataFile);
				}
				
			if((then=getNextLogFile(dataFile, nextFile))>0 && then<=now)
				{
				strcpy(dataFile, nextFile);
				dataFileOffset = 0;
				LogRecord("PBS_Accounting", "Switching to account file %s", dataFile);
				}
			else if(!then)	
				{
				LogError("PBS_Accounting", "Bad file name format %s", dataFile);
				*dataFile      = 0;
				dataFileOffset = 0;
				}
			} while(then>0 && then<=now);
		purge();
		}
	else	{
		clear();
		initialize();
		}
	}

// *****************************************************************************
// * PBS_AccountingDatabase::update :
// *	In addition to populating the database with information stored in the
// *	disk files, this method will also update the database using the current
// *	CPU and wall time utilization from the PBS_Server object.
// *****************************************************************************
void PBS_AccountingDatabase::update ( PBS_Server & server )
	{
	PBS_AccountingRecord * rec;
	
	// *********************************************************************
	// * First call the standard update to refresh the database from the
	// * records on disk.
	// *********************************************************************
	update();
	
	// *********************************************************************
	// * Get a list of running jobs from the server.
	// *********************************************************************
	PBS_Job ** running = server.getJobsByState(PBS_Job::RUNNING);
	size_t     idx     = 0; 
	
	// *********************************************************************
	// * Walk through the list of records and delete any record that does
	// * not have an end time and is not registered with the PBS_Server ---
	// * These jobs are obviously bogus and will only serve to skew the 
	// * computation.
	// *********************************************************************
	for(idx=0; idx<numRecs; )
		{
		if(recs[idx]->getEndTime()==0 &&
		   server.getJob((char *)recs[idx]->getJob())==NULL)
		   	{
			rec = detachRecord(recs[idx]);
			delete rec;
			}
		else idx++;
		}
	
	// *********************************************************************
	// * Walk through the list of running jobs and update the PBS_JobRecord
	// * object using the current wall and cpu times.
	// *********************************************************************
	for(idx=0; running && running[idx]; idx++)
		{
		PBS_AccountingRecord * rec = find(running[idx]->getName());
		if(rec)
			{
			time_t cpuTime  = stringToDuration(running[idx]->getUsedResources()->find("cput"),0);
			time_t wallTime = stringToDuration(running[idx]->getUsedResources()->find("walltime"),0);
			rec->updateRecord(0, cpuTime, wallTime); 
			}
		}
	// *********************************************************************
	// * Delete the temporary 'running' array that was returned by the
	// * PBS_Server object --- don't delete the actual objects.
	// *********************************************************************
	delete[] running;
	}

// *****************************************************************************
// * PBS_AccountingDatabase::purge :
// *	This method will remove all database entries for jobs that completed
// *	prior to 'daysOld' days in the past. Also, if a job has no specified
// *	endTime and it has been 'running' longer than the maxWallTime value - 
// *	it will be purged under the assumption that it is a job that was deleted
// *	during a server outage.
// *****************************************************************************
void PBS_AccountingDatabase::purge ( void )
	{
	PBS_AccountingRecord * rec;
	time_t now = time(NULL);
	time_t old = now-(daysOld*24*60*60);

	for(size_t idx=0; idx<numRecs; )
		{
		time_t startTime   = recs[idx]->getStartTime();
		time_t endTime     = recs[idx]->getEndTime();
		
		if((endTime && endTime<old) || (!endTime && (now-startTime)>maxWallTime))
			{
			rec = detachRecord(recs[idx]);
			delete rec;
			}
		else idx++;
		}
	}		


// *****************************************************************************
// * PBS_AccountingDatabase::findNode :
// *	This method is called to find the PBS_Accounting object associated with
// *	a specific node.  If an associated node cannot be found, then a 
// *	PBS_Accounting object will be created for the node and added to the 
// *	nodes list. In the event that the specified node name is invalid (NULL
// *	or an empty string), a NULL pointer will be returned.
// *****************************************************************************
PBS_NodeAccounting * PBS_AccountingDatabase::findNode ( const char * node )
	{
	PBS_NodeAccounting * entry = NULL;
	if(node && *node && (entry = (PBS_NodeAccounting *)nodes.find(node))==NULL)
		{
		entry = new PBS_NodeAccounting(node);
		nodes.insert(entry->getName(), entry);
		}
	return entry;
	}

// *****************************************************************************
// * PBS_AccountingDatabase::findUser :
// *	This method is called to find the PBS_Accounting object associated with
// *	a specific user.  If an associated user cannot be found, then a 
// *	PBS_Accounting object will be created for the user and added to the 
// *	users list. In the event that the specified user name is invalid (NULL
// *	or an empty string), a NULL pointer will be returned.
// *****************************************************************************
PBS_UserAccounting * PBS_AccountingDatabase::findUser ( const char * user )
	{
	PBS_UserAccounting * entry = NULL;

	// *********************************************************************
	// * Remember to elliminate any domain identification from the user's
	// * name before processing.
	// *********************************************************************
	char localUser[256];
	substrcpy(localUser, user, "@");
	
	if(*localUser && (entry = (PBS_UserAccounting *)users.find(localUser))==NULL)
		{
		entry = new PBS_UserAccounting(localUser);
		users.insert(entry->getName(), entry);
		}
	return entry;
	}

// *****************************************************************************
// * PBS_AccountingDatabase::findQueue :
// *	This method will locate and return the PBS_Accounting object associated 
// *	with the specified queue name. In the event that no associated queue
// *	exists, then a new PBS_Accounting object will be craeted and added to 
// *	the list.  If the specified queue name is invalid (NULL or empty) then
// *	a NULL pointer will be returned.
// *****************************************************************************
PBS_QueueAccounting * PBS_AccountingDatabase::findQueue ( const char * queue )
	{
	PBS_QueueAccounting * entry = NULL;
	if(queue && *queue && (entry = (PBS_QueueAccounting *)queues.find(queue))==NULL)
		{
		entry = new PBS_QueueAccounting(queue);
		queues.insert(entry->getName(), entry);
		}
	return entry;
	}
	
// *****************************************************************************
// * PBS_AccountingDatabase::scanRecord :
// *	This method is called during the initialize and update functions to 
// *	scan the contents of the text line and generate a new 
// *	PBS_AccountingRecord object. The new object will be added to the primary
// *	array of records. If an old record exists with the same job name, then
// *	it will be deleted.
// *****************************************************************************
int PBS_AccountingDatabase::scanRecord ( const char * buf )
	{
	int                    result = -1;
	PBS_AccountingRecord * rec    = PBS_AccountingRecord::create(*this, buf);
	PBS_AccountingRecord * oldRec = NULL;
	if(rec)
		{
		result = 0;
		if((oldRec=attachRecord(rec))!=NULL)
			{
			if(rec==oldRec) result = -1;
			delete oldRec;
			}
		}
	return result;
	}	

// *****************************************************************************
// * PBS_AccountingDatabase::asciiDump :
// *	This method is called to create an asciiDump of the contents of the 
// *	PBS_AccountingDatabase.  Because this class contains comprehensive
// *	data, the asciiDump method will generate a summary of the objects 
// *	contents.
// *****************************************************************************
void PBS_AccountingDatabase::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	if(!prefix) prefix = &c;
	
	size_t  strIdx  = 0;
	size_t  strCnt  = 0;
	size_t  strMax  = 128; 
	char ** str     = (char **)malloc(128*sizeof(char *));
	char *  sections[3];
	
	PBSStringHash * hash[3];
	hash [0] = &queues;
	hash [1] = &users;
	hash [2] = &nodes;
	sections[0] = "Queue";
	sections[1] = "User";
	sections[2] = "Node";
	
	fprintf(fp, "%s========================================================\n", prefix);
	fprintf(fp, "%sPBS Accounting Summary For Server: %s\n", prefix, name);
	fprintf(fp, "%s========================================================\n", prefix);

	char * subPrefix = new char[strlen(prefix)+5];
	sprintf(subPrefix, "%s    ", prefix);
	
	for(int idx=0; idx<3; idx++)
		{
		PBS_Accounting  * rec;
		PBSStringHashIter iter(*(hash[idx]));
		
		fprintf(fp, "%s\n%s  ============================\n", prefix, prefix);
		fprintf(fp, "%s  Begin %s Summary:\n", prefix, sections[idx]);
		fprintf(fp, "%s  ============================\n", prefix);
			
		for(iter.first(); (rec=(PBS_Accounting *)iter.data())!=NULL; iter++)
			{
			if(strCnt>=strMax) str = (char **)realloc(str, (strMax*=2)*sizeof(char *));
			str[strCnt++] = iter.key();
			}
		qsort(str, strCnt, sizeof(char *), stringPointerCompare);
		for(strIdx=0; strIdx<strCnt; strIdx++)
			{
			if((rec = (PBS_Accounting *)hash[idx]->find(str[strIdx]))!=NULL)
				rec->asciiDump(fp, subPrefix);
			}
		}	

	free(str);
	}
