// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Base.h
// * Author:      Walt Akers
// * Description: This file contains the class definition for the PBS_Base
// *              class.  This class provides mechanisms that are generally 
// *              useful to the other PBS classes.
// *****************************************************************************
#ifndef _PBS_BASE_H_
#define _PBS_BASE_H_ 1

#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/param.h>
#include <PBS_Hash.h>
#include <PBS_Resource.h>

extern "C" {
#include <pbs_error.h>
#include <pbs_ifl.h>
}

// *****************************************************************************
// * class PBS_Base :
// *	This is the base class for the PBS loader classes.  The child classes
// *	will use the methods defined here to load data that is associated with
// *	an individual PBS component.
// *****************************************************************************
class PBS_Base
{
private:
	// *********************************************************************
	// * PBSStringIndex queueHash :
	// *	This hash table is used to associate queueName with integer 
	// *	identifiers.  The next available queue ID is maintained in the
	// *	nextQueueID variable.
	// *********************************************************************
	static PBSStringIndex queueHash;
	static unsigned       nextQueueID;
	
	// *********************************************************************
	// * PBSStringIndex userHash :
	// *	This has table is used to associate a userName with an integer
	// *	identifier. The next available user ID is maintained in the
	// *	nextUserID variable.
	// *********************************************************************
	static PBSStringIndex userHash;
	static unsigned       nextUserID;

public:
	// *********************************************************************
	// * These two methods allow the caller to obtain a locally defined
	// * identifier associated with either a queue or a username.
	// *********************************************************************
	static unsigned      getUserIDFromName  ( const char * userName );
	static unsigned      getQueueIDFromName ( const char * queueName );

	// *********************************************************************
	// * Enumerated BOOLEAN class.
	// *********************************************************************
	typedef enum
		{
		TRUE=-1,
		FALSE
		} Boolean;
	
	// *********************************************************************
	// * An ordered enumeration of queue types.
	// *********************************************************************
	typedef enum
		{
		EXECUTION=0,
		ROUTING
		} QueueType;
	
	// *********************************************************************
	// * An ordered enumeration of job states.
	// *********************************************************************
	typedef enum
		{
		TRANSIT=0,
		QUEUED,
		HELD,
		WAITING,
		RUNNING,
		EXITING
		} StateCountIndex;    

	// *********************************************************************
	// * Type definition for the variable used to hold various job states.
	// *********************************************************************
	typedef size_t StateCount[6];
	
	inline void nullify             (char * & ptr);
	inline void nullify             (char ** & ptr);
	inline void readStringValue     (char * & val, struct attrl *attrib );
	inline void readListValue       (char ** & array, struct attrl *attrib, char delimiter=',');
	inline void readIntegerValue    (int & val, struct attrl *attrib);
	inline void readIntegerValue    (unsigned & val, struct attrl *attrib);
	inline void readIntegerValue    (long & val, struct attrl *attrib);
	inline void readIntegerValue    (unsigned long & val, struct attrl *attrib);
	inline void readBooleanValue    (Boolean & val, struct attrl *attrib);
	inline void readStateCountValue (StateCount & state_count, struct attrl *attrib );
	inline void readResourceValue   (PBS_Resource & rsc, struct attrl *attrib);
	inline void readQueueTypeValue  (QueueType & val, struct attrl * attrib);
	inline time_t stringToDuration  (const char * str, time_t defaultValue);
};


// *********************************************************************
// * PBS_Base::nullify :
// *	This method will free the memory pointer and then set it to 
// *	NULL.  This function should only be called with pointers that
// *	were created by malloc, alloc, calloc or realloc.  Memory
// *	blocks that were created using new must be destroyed using 
// *	delete.
// *********************************************************************
inline void PBS_Base::nullify ( char * & ptr )
	{
	if(ptr!=NULL)
		{
		free(ptr);
		ptr = NULL;
		}
	}

// *********************************************************************
// * PBS_Base::nullify :
// *	This method will free each memory pointer within the array
// *    and then will delete the array itself.  The internal 
// *	pointers should have been created using strdup and the overall
// *	array should have been allocated using new.
// *********************************************************************
inline void PBS_Base::nullify (char ** & ptr)
	{
	int idx = 0;
	
	if(ptr!=NULL)
		{
		while(ptr[idx]!=NULL) nullify(ptr[idx++]);
		delete[] ptr;
		ptr = NULL;
		}
	} 


// *********************************************************************
// * PBS_Base::readStringValue :
// *	This method is called by the various loadValue methods in the
// *	child classes whenever a string attribute needs to be read and
// *	stored.
// *********************************************************************
inline void PBS_Base::readStringValue ( char * & val, struct attrl * attrib )
	{
	val = (attrib->value && *attrib->value)?strdup(attrib->value):NULL;
	}

// *********************************************************************
// * PBS_Base::readListValue :
// *	This method will create an array of strings from a comma
// *	separated list.  The array will be terminated by a NULL pointer.
// *	The array (and its contents) that is created by this method 
// *	can be destroyed using the nullify method defined in this class.
// *********************************************************************
inline void PBS_Base::readListValue ( char ** & array, struct attrl * attrib, char delimiter )
	{
	char * list = attrib->value;
	char * ptr;
	int    cnt = 0;

	array = NULL;

	if(list && *list)
		{
		for(cnt=1, ptr=list; *ptr; ptr++) if(*ptr==delimiter) cnt++;

		array = new char * [cnt+1];
		cnt   = 0;
		while(isspace(*list) && *list) list++;
		while(*list)
			{
			ptr=list;
			while(*ptr && *ptr!=delimiter) ptr++;
			char c = *ptr;
			*ptr   =  0;
			array[cnt++]=strdup(list);
			*ptr   =  c;
			list   =  c?ptr+1:ptr; 
			while(isspace(*list) && *list) list++;
			}
		array[cnt] = NULL;
		}
	}

// *********************************************************************
// * PBS_Base::readIntegerValue :
// *	This method is called by the loadValue method of the various
// *	child classes whenever an integer attribute needs to be 
// *	interpretted and stored.
// *********************************************************************
inline void PBS_Base::readIntegerValue (int & val, struct attrl * attrib)
	{
	val = attrib->value?atoi(attrib->value):0;
	}

// *********************************************************************
// * PBS_Base::readIntegerValue :
// *	This method is called by the loadValue method of the various
// *	child classes whenever an integer attribute needs to be 
// *	interpretted and stored.
// *********************************************************************
inline void PBS_Base::readIntegerValue (unsigned & val, struct attrl * attrib)
	{
	val = attrib->value?atoi(attrib->value):0;
	}

// *********************************************************************
// * PBS_Base::readIntegerValue :
// *	This method is called by the loadValue method of the various
// *	child classes whenever an integer attribute needs to be 
// *	interpretted and stored.
// *********************************************************************
inline void PBS_Base::readIntegerValue (long & val, struct attrl * attrib)
	{
	val = attrib->value?atoi(attrib->value):0;
	}

// *********************************************************************
// * PBS_Base::readIntegerValue :
// *	This method is called by the loadValue method of the various
// *	child classes whenever an integer attribute needs to be 
// *	interpretted and stored.
// *********************************************************************
inline void PBS_Base::readIntegerValue (unsigned long & val, struct attrl * attrib)
	{
	val = attrib->value?atoi(attrib->value):0;
	}

// *********************************************************************
// * PBS_Base::readBooleanValue :
// *	This method is called by the loadValue method of the various
// *	child classes whenever a boolean attribute needs to be 
// *	interpretted and stored.
// *********************************************************************
inline void PBS_Base::readBooleanValue (Boolean & val, struct attrl * attrib)
	{
	val = TRUE;
	if(attrib->value)
		{
		int  i;
		char uprStr[6];

		for(i=0; i<5 && attrib->value[i]; i++)
			{
			uprStr[i] = toupper(attrib->value[i]);
			}
		uprStr[i] = 0;
	
		if(!strcmp(uprStr, "FALSE") || 
		   !strcmp(uprStr, "NO")    ||
		   !strcmp(uprStr, "0")     ||
		   !strcmp(uprStr, "F")     ||
		   !strcmp(uprStr, "N"))  val = FALSE;
		}		
	}


// *****************************************************************************
// * PBS_Base::readStateCountValue :
// *	This method is called by the loadValue method when the 
// *	"state_count" attribute must be processed.  The state_count 
// *	value contains 6 seperate entries that identify how many of 
// *	the server's jobs are in each different job state.
// *****************************************************************************
inline void PBS_Base::readStateCountValue ( StateCount & state_count, struct attrl * attrib )
	{
	char * ptr;

	if((ptr = strstr(attrib->value, "Transit:"))!=NULL)
		state_count[TRANSIT] = atoi(ptr+8);
	else    state_count[TRANSIT] = 0;
	
	if((ptr = strstr(attrib->value, "Queued:"))!=NULL)
		state_count[QUEUED] = atoi(ptr+7);
	else	state_count[QUEUED] = 0;
	
	if((ptr = strstr(attrib->value, "Held:"))!=NULL)
		state_count[HELD] = atoi(ptr+5);
	else	state_count[HELD] = 0;
	
	if((ptr = strstr(attrib->value, "Waiting:"))!=NULL)
		state_count[WAITING] = atoi(ptr+8);
	else	state_count[WAITING] = 0;
	
	if((ptr = strstr(attrib->value, "Running:"))!=NULL)
		state_count[RUNNING] = atoi(ptr+8);
	else	state_count[RUNNING] = 0;
	
	if((ptr = strstr(attrib->value, "Exiting:"))!=NULL)
		state_count[EXITING] = atoi(ptr+8);
	else	state_count[EXITING] = 0;
	}


// *****************************************************************************
// * PBS_Base::readResourceValue   :
// *	This method is called by the loadValue method whenever an attribute
// *	of type PBS_Resource is received.  The attrib->resource field
// *	contains the name of the resource that is being defined and the 
// *	attrib->value field contains the value of the resource.  
// *
// *	The PBS_Resource class maintains a list of names and values for all 
// *	resources that are assigned through this interface.  The user must
// *	convert the value string to the appropriate type when he extracts
// *	the value from the PBS_Resource object.
// *****************************************************************************
inline void PBS_Base::readResourceValue   (PBS_Resource & rsc, struct attrl * attrib)
	{
	rsc.insert(attrib->resource, attrib->value);
	}

// *****************************************************************************
// * PBS_Base::readQueueTypeValue :
// *	This method will translate the queue_type character string attribute
// *	into an integer value.
// *****************************************************************************
void PBS_Base::readQueueTypeValue (QueueType & val, struct attrl * attrib)
	{
	val = EXECUTION;
	
	switch(tolower(*attrib->value))
		{
		case 'r':
		val = ROUTING;
		break;
		}
	}
	
// *****************************************************************************
// * PBS_Base::stringToDuration :
// *    This method converts a time string to a time_t. The defaultValue 
// *    provided by the caller will be used in the event that the str is NULL
// *    or blank.
// *****************************************************************************
inline time_t PBS_Base::stringToDuration ( const char * str, time_t defaultValue )
        {
        time_t result=defaultValue;

        if(str && *str)
                {
                int multx = 1;
                for(const char *ptr=str; multx<3600 && (ptr=strchr(ptr, ':'))!=NULL; ptr++)
                        {
                        multx*=60;
                        }

                result    = 0;
                while(*str && multx)
                        {
                        result += atoi(str)*multx;
                        if(*(str+=strcspn(str, ":"))) str++;
                        multx/=60;
                        }
                }

        return result;
        }

#endif /* _PBS_BASE_H_ 1 */
