// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	This header defines a local implementation of a string and integer 
//	hash table and associated iterators.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _PBS_HASH_H_
#define _PBS_HASH_H_ 1

#include <stdlib.h>
#include <string.h>

class PBSIntHash
{
friend class PBSIntHashIter;

private:
	class Node
		{
		public:
			int    hashInt;
			void * hashData;
			Node * next;
		};

	size_t    nodeCnt;
	size_t    nodeMax;
	Node   ** nodes;
	
	inline Node * extract ( void );
	inline void   rehash  ( void );

public:
	inline PBSIntHash  ( int maxNodes = 256 );
	inline ~PBSIntHash ( void );	
	
	inline void   insert ( int hashInt, void * hashData );
	inline void * remove ( int hashInt );
	inline void *   find ( int hashInt );
};


class PBSStringHash
{
friend class PBSStringHashIter;

private:
	class Node
		{
		public:
			char * hashStr;
			void * hashData;
			Node * next;
		};
	
	int       copyKeyFlag;
	size_t    nodeCnt;
	size_t    nodeMax;
	Node   ** nodes;
	
	inline Node * extract ( void );
	inline void   rehash  ( void );

public:
	inline PBSStringHash  ( int maxNodes = 256, int copyFlag = 0);
	inline ~PBSStringHash ( void );	
	
	inline unsigned int hash ( const char * hashStr );
	inline void       insert ( const char * hashStr, void * hashData );
	inline void *     remove ( const char * hashStr );
	inline void *       find ( const char * hashStr );
};

class PBSStringIndex
{
private:
	class Node
		{
		public:
			char   * hashStr;
			size_t   hashData;
			Node   * next;
		};
	
	int       copyKeyFlag;
	size_t    nodeCnt;
	size_t    nodeMax;
	Node   ** nodes;
	
	inline Node * extract ( void );
	inline void   rehash  ( void );

public:
	inline PBSStringIndex  ( int maxNodes = 256, int copyFlag = 0);
	inline ~PBSStringIndex ( void );	
	
	inline unsigned int hash ( const char * hashStr );
	inline void       insert ( const char * hashStr, size_t hashData );
	inline size_t     remove ( const char * hashStr );
	inline size_t       find ( const char * hashStr );
};

class PBSIntHashIter
{
private:
	PBSIntHash       & hash;
	PBSIntHash::Node * node;
	size_t             idx;
	
public:
	PBSIntHashIter ( PBSIntHash & Hash )
		: hash(Hash),
		  node(NULL),
		  idx (0)
		{
		first();
		}
	void * first ( void )
		{
		for(idx=0; idx<hash.nodeMax && (node=hash.nodes[idx])==NULL; idx++);
		return node?node->hashData:NULL;
		}
	void * operator ++ ( int x )
		{
		PBSIntHash::Node * result = node;

		if(x==0)
			{
			if(node) node = node->next;
			if(!node) for( idx++; idx<hash.nodeMax && (node=hash.nodes[idx])==NULL; idx++);
			}
		return result?result->hashData:NULL;
		}
	void * operator ++ ( void )
		{
		operator ++ ( 0 );
		return data();
		}
	void * data ( void )
		{
		return node?node->hashData:NULL;
		}
	int    key  ( void )
		{
		return node?node->hashInt:0;
		}
};


class PBSStringHashIter
{
private:
	PBSStringHash       & hash;
	PBSStringHash::Node * node;
	size_t             idx;
	
public:
	PBSStringHashIter ( PBSStringHash & Hash )
		: hash(Hash),
		  node(NULL),
		  idx (0)
		{
		first();
		}
	void * first ( void )
		{
		for(idx=0; idx<hash.nodeMax && (node=hash.nodes[idx])==NULL; idx++);
		return node?node->hashData:NULL;
		}
	void * operator ++ ( int x )
		{
		PBSStringHash::Node * result = node;
		
		if(x==0)
			{
			if(node) node = node->next;
			if(!node) for( idx++; idx<hash.nodeMax && (node=hash.nodes[idx])==NULL; idx++);
			}
			
		return result?result->hashData:NULL;
		}
	void * operator ++ ( void )
		{
		operator ++ ( 0 );
		return data();
		}
	void * data ( void )
		{
		return node?node->hashData:NULL;
		}
	char * key  ( void )
		{
		return node?node->hashStr:0;
		}
};

inline PBSIntHash::PBSIntHash ( int maxNodes )
	: nodeCnt(0), 
	  nodeMax(maxNodes), 
	  nodes  (NULL)
	{
	nodes = new Node * [nodeMax];
	memset(nodes, 0, nodeMax*sizeof(Node *));
	}

inline PBSIntHash::~PBSIntHash ( void )
	{
	Node * node = extract();
	delete[] nodes;
	
	while(node!=NULL)
		{
		Node * curr = node;
		node = node->next;
		delete curr;
		}
	}
	
inline PBSIntHash::Node * PBSIntHash::extract ( void )
	{
	Node * head = NULL;
	Node * tail = NULL;
	
	for(int idx=0; idx<nodeMax; idx++)
		{
		if(nodes[idx])
			{
			if(!head) head       = nodes[idx];
			if(tail)  tail->next = nodes[idx];
			else      tail       = nodes[idx];
			while(tail->next!=NULL) tail = tail->next;
			}
		}
	memset(nodes, 0, nodeMax*sizeof(Node *));
	nodeCnt = 0;
	return head;
	}
	
inline void PBSIntHash::rehash ( void )
	{
	Node * old = extract();

	nodeMax*=2;
	if(nodes) delete nodes;
	nodes = new Node *[nodeMax];
	memset(nodes, 0, nodeMax*sizeof(Node *));
	
	while(old)
		{
		Node * node = old;
		size_t idx  = node->hashInt%nodeMax;
		Node * curr = nodes[idx];
		Node * prev = NULL;
		
		old         = node->next;
		node->next  = NULL;
	
		while(curr && curr->hashInt!=node->hashInt)
			{
			prev = curr;
			curr = curr->next;
			}
		if(curr)
			{
			node->next = curr->next;
			delete curr;
			}
		if(prev) prev->next = node;
		else     nodes[idx] = node;	
		}
	}

inline void PBSIntHash::insert ( int hashInt, void * hashData )
	{
	size_t idx  = hashInt%nodeMax;
	Node * prev = NULL;
	Node * curr = nodes[idx];
	
	while(curr && curr->hashInt!=hashInt)
		{
		prev = curr;
		curr = curr->next;
		}
	if(curr) curr->hashData = hashData;
	else	{
		Node * node   = new Node;
		node->hashInt   = hashInt;
		node->hashData = hashData;
		node->next    = NULL;
		
		if(prev) prev->next = node;
		else nodes[idx] = node;
		
		nodeCnt++;
		}
	
	if(nodeCnt>nodeMax) rehash();
	}
	
inline void * PBSIntHash::remove ( int hashInt )
	{
	size_t idx  = hashInt%nodeMax;
	Node * prev = NULL;
	Node * curr = nodes[idx];
	void * data = NULL;
	
	while(curr && curr->hashInt!=hashInt)
		{
		prev = curr;
		curr = curr->next;
		}
	if(curr)
		{
		if(prev) prev->next = curr->next;
		else     nodes[idx] = curr->next;
		data = curr->hashData;
		delete curr;
		
		nodeCnt--;
		}
	return data;
	}

inline void * PBSIntHash::find ( int hashInt )
	{
	Node * curr = nodes[hashInt%nodeMax];

	while(curr && curr->hashInt!=hashInt) curr = curr->next;
	return (curr?curr->hashData:NULL);
	}

inline PBSStringHash::PBSStringHash  ( int maxNodes, int copyKey ) 
	: copyKeyFlag(copyKey),
	  nodeCnt(0), 
	  nodeMax(maxNodes), 
	  nodes  (NULL)
	{
	nodes = new Node * [nodeMax];
	memset(nodes, 0, nodeMax*sizeof(Node *));
	}

inline PBSStringHash::~PBSStringHash ( void )
	{
	Node * node = extract();
	delete[] nodes;
	
	while(node!=NULL)
		{
		Node * curr = node;
		node = node->next;

		if(copyKeyFlag) free(curr->hashStr);
		delete curr;
		}
	}

inline PBSStringHash::Node * PBSStringHash::extract ( void )
	{
	Node * head = NULL;
	Node * tail = NULL;
	
	for(int idx=0; idx<nodeMax; idx++)
		{
		if(nodes[idx])
			{
			if(!head) head       = nodes[idx];
			if(tail)  tail->next = nodes[idx];
			else      tail       = nodes[idx];
			while(tail->next!=NULL) tail = tail->next;
			}
		}
	memset(nodes, 0, nodeMax*sizeof(Node *));
	nodeCnt = 0;
	return head;
	}
	
inline void PBSStringHash::rehash ( void )
	{
	Node * old = extract();

	nodeMax*=2;
	if(nodes) delete[] nodes;
	nodes = new Node *[nodeMax];
	memset(nodes, 0, nodeMax*sizeof(Node *));
	
	while(old)
		{
		Node * node = old;
		size_t idx  = hash(node->hashStr);
		Node * curr = nodes[idx];
		Node * prev = NULL;
		
		old         = node->next;
		node->next  = NULL;
	
		while(curr && strcmp(curr->hashStr,node->hashStr))
			{
			prev = curr;
			curr = curr->next;
			}
		if(curr)
			{
			node->next = curr->next;

			if(copyKeyFlag) free(curr->hashStr);
			delete curr;
			}
		if(prev) prev->next = node;
		else     nodes[idx] = node;	
		}
	}

inline unsigned int PBSStringHash::hash ( const char * hashStr )
	{
	unsigned int hash = 0, g;

	for (int i = 0; hashStr && hashStr[i]; i++)
		{
		hash = (hash << 4) + hashStr[i];
		if (g = hash & 0xf0000000)
			{
			hash ^= g >> 24;
			hash ^= g;
			}
		}
	return (hash % nodeMax);
	}

inline void PBSStringHash::insert ( const char * hashStr, void * hashData )
	{
	unsigned int idx  = hash(hashStr);
	Node *       prev = NULL;
	Node *       curr = nodes[idx];
	
	while(curr && strcmp(curr->hashStr, hashStr))
		{
		prev = curr;
		curr = curr->next;
		}
	if(curr) curr->hashData = hashData;
	else	{
		Node * node    = new Node;
		node->hashStr  = copyKeyFlag?strdup(hashStr):(char *)hashStr;
		node->hashData = hashData;
		node->next     = NULL;
		
		if(prev) prev->next = node;
		else     nodes[idx] = node;
		
		nodeCnt++;
		}
	if(nodeCnt>nodeMax) rehash();
	}

inline void * PBSStringHash::remove ( const char * hashStr )
	{
	unsigned int idx = hash(hashStr);
	Node * prev = NULL;
	Node * curr = nodes[idx];
	void * data = NULL;
	
	while(curr && strcmp(curr->hashStr, hashStr))
		{
		prev = curr;
		curr = curr->next;
		}
	if(curr)
		{
		if(prev) prev->next = curr->next;
		else     nodes[idx] = curr->next;
		
		if(copyKeyFlag) free(curr->hashStr);
		data = curr->hashData;
		delete curr;
		
		nodeCnt--;
		}

	return data;
	}

inline void * PBSStringHash::find ( const char * hashStr )
	{
	Node * curr = nodes[hash(hashStr)];
	while(curr && strcmp(curr->hashStr, hashStr)) curr = curr->next;
	return curr?curr->hashData:NULL;
	}

inline PBSStringIndex::PBSStringIndex  ( int maxNodes, int copyKey ) 
	: copyKeyFlag(copyKey),
	  nodeCnt(0), 
	  nodeMax(maxNodes), 
	  nodes  (NULL)
	{
	nodes = new Node * [nodeMax];
	memset(nodes, 0, nodeMax*sizeof(Node *));
	}

inline PBSStringIndex::~PBSStringIndex ( void )
	{
	Node * node = extract();
	delete[] nodes;
	
	while(node!=NULL)
		{
		Node * curr = node;
		node = node->next;

		if(copyKeyFlag) free(curr->hashStr);
		delete curr;
		}
	}

inline PBSStringIndex::Node * PBSStringIndex::extract ( void )
	{
	Node * head = NULL;
	Node * tail = NULL;
	
	for(int idx=0; idx<nodeMax; idx++)
		{
		if(nodes[idx])
			{
			if(!head) head       = nodes[idx];
			if(tail)  tail->next = nodes[idx];
			else      tail       = nodes[idx];
			while(tail->next!=NULL) tail = tail->next;
			}
		}
	memset(nodes, 0, nodeMax*sizeof(Node *));
	nodeCnt = 0;
	return head;
	}
	
inline void PBSStringIndex::rehash ( void )
	{
	Node * old = extract();

	nodeMax*=2;
	if(nodes) delete[] nodes;
	nodes = new Node *[nodeMax];
	memset(nodes, 0, nodeMax*sizeof(Node *));
	
	while(old)
		{
		Node * node = old;
		size_t idx  = hash(node->hashStr);
		Node * curr = nodes[idx];
		Node * prev = NULL;
		
		old         = node->next;
		node->next  = NULL;
	
		while(curr && strcmp(curr->hashStr,node->hashStr))
			{
			prev = curr;
			curr = curr->next;
			}
		if(curr)
			{
			node->next = curr->next;

			if(copyKeyFlag) free(curr->hashStr);
			delete curr;
			}
		if(prev) prev->next = node;
		else     nodes[idx] = node;	
		}
	}

inline unsigned int PBSStringIndex::hash ( const char * hashStr )
	{
	unsigned int hash = 0, g;

	for (int i = 0; hashStr && hashStr[i]; i++)
		{
		hash = (hash << 4) + hashStr[i];
		if (g = hash & 0xf0000000)
			{
			hash ^= g >> 24;
			hash ^= g;
			}
		}
	return (hash % nodeMax);
	}

inline void PBSStringIndex::insert ( const char * hashStr, size_t hashData )
	{
	unsigned int idx  = hash(hashStr);
	Node *       prev = NULL;
	Node *       curr = nodes[idx];
	
	while(curr && strcmp(curr->hashStr, hashStr))
		{
		prev = curr;
		curr = curr->next;
		}
	if(curr) curr->hashData = hashData;
	else	{
		Node * node    = new Node;
		node->hashStr  = copyKeyFlag?strdup(hashStr):(char *)hashStr;
		node->hashData = hashData;
		node->next     = NULL;
		
		if(prev) prev->next = node;
		else     nodes[idx] = node;
		
		nodeCnt++;
		}
	if(nodeCnt>nodeMax) rehash();
	}

inline size_t PBSStringIndex::remove ( const char * hashStr )
	{
	unsigned int idx = hash(hashStr);
	Node * prev = NULL;
	Node * curr = nodes[idx];
	size_t data = 0;
	
	while(curr && strcmp(curr->hashStr, hashStr))
		{
		prev = curr;
		curr = curr->next;
		}
	if(curr)
		{
		if(prev) prev->next = curr->next;
		else     nodes[idx] = curr->next;
		
		if(copyKeyFlag) free(curr->hashStr);
		data = curr->hashData;
		delete curr;
		
		nodeCnt--;
		}

	return data;
	}

inline size_t PBSStringIndex::find ( const char * hashStr )
	{
	Node * curr = nodes[hash(hashStr)];
	while(curr && strcmp(curr->hashStr, hashStr)) curr = curr->next;
	return curr?curr->hashData:0;
	}

#endif /* _PBS_HASH_H_ */
