#ifndef _PBS_KEYWORDTABLE_H_
#define _PBS_KEYWORDTABLE_H_ 

#include <stdio.h>
#include <PBS_Hash.h>

// *****************************************************************************
// * union PBS_Key :
// *	This is a union that is used to express a key.  The first eight bytes of 
// *	the union is the index into a table of bit masks (0 to 255).  The next
// *	24 bits is the bitsmask itself.  Using this technique, a table of 
// *	24 * 255 (6120) keywords can be represented in the table.
// *****************************************************************************
typedef union
	{
	struct	{		
		unsigned idx  : 8;
		unsigned bits : 24;
		} mask;
	unsigned value;
	} PBS_Key;
	

// *****************************************************************************
// * class PBS_KeywordTable :
// *	This object is used to maintain a list of keywords that are used to
// *	match a job's requirements with a node's resources.
// *****************************************************************************
class PBS_KeywordTable
{
friend class PBS_KeySet;

protected:
	size_t         idxMax;
	int            next;
	PBSStringIndex strHash;
	PBSIntHash     keyHash;

public:
	PBS_KeywordTable ( void )
		: strHash(16), 
		  keyHash(16), 
		  next   (0), 
		  idxMax (1)
		{
		}	
	
	~PBS_KeywordTable ( void )
		{
		PBSIntHashIter iter(keyHash);
		char         * str;

		for(iter.first(); (str=(char *)iter.data())!=NULL; iter++)
			{
			strHash.remove(str);
			free(str);
			}
		}
		
	unsigned addKeyword ( char * keyword )
		{
		unsigned keyValue = 0;
		
		if(keyword && 
		  *keyword && 
		  (keyValue=(unsigned)strHash.find(keyword))==0)
			{
			int bit = next%24;
			int idx = next/24;
			if(idx<256)
				{
				PBS_Key key;
				char  * name = strdup(keyword);
				key.mask.idx  = idx;
				key.mask.bits = 1<<bit;
				keyValue      = key.value;
				strHash.insert(name, keyValue);
				keyHash.insert(keyValue, name);
				
				idxMax = idx+1;
				next++;
				}
			}
		return keyValue;
		}

	unsigned getMaxIndex ( void )
		{
		return idxMax;
		}
		
	unsigned getKey ( char * keyword )
		{
		return addKeyword(keyword);
		}

	const char * getKeyword ( unsigned key )
		{
		return (const char *)keyHash.find(key);
		}
		
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL )
		{
		char   blank = 0;
		size_t prefixLen;
		char   headLine[82];
	
		if(prefix==NULL) prefix = &blank;
		prefixLen = strlen(prefix);
	
		memset(headLine, '-', 80);
		headLine[80]='\n';
		headLine[81]='\0';
		memcpy(headLine, prefix, (prefixLen<=80)?prefixLen:80);
		
		fprintf(fp, "%s        Key    Idx           Bit    Keyword\n", prefix);
		fprintf(fp, headLine);
		for(int idx=0; idx<24*256; idx++)
			{
			PBS_Key      key;
			const char * str;

			key.mask.idx  = idx/24;
			key.mask.bits = 1<<(idx%24);
			
			if((str = (char *)keyHash.find(key.value))!=NULL)
				{
				fprintf(fp, "%s %10u    %3u    %10u    %s\n", prefix, key.value, key.mask.idx, key.mask.bits, str); 
				}
			}
		}
};

// *****************************************************************************
// * class PBS_KeySet :
// *	This table is a colleciton of keys that represent all of the keywords
// *	supported by a node or requested by a job.
// *****************************************************************************
class PBS_KeySet
{
protected:
	PBS_KeywordTable & table;
	size_t             idxCnt;
 
	PBS_Key          * keys;
	size_t             keyCnt;
	
public:
	// *********************************************************************
	// * PBS_KeySet :
	// *	Initializes internal variables and prepares the object for 
	// *	operation.  This object must be created through the 
	// *	PBS_KeywordTable object and then deleted by the user.
	// *********************************************************************
	PBS_KeySet ( PBS_KeywordTable & Table )
		: table  (Table),
		  idxCnt (0),
		  keys   (NULL),
		  keyCnt (0)
		{
		update();
		}
	
	// *********************************************************************
	// * PBS_KeySet :
	// *	This is the copy constructor for the PBS_KeySet class. It
	// *	will extract the contents of the KeySet reference and 
	// *	DUPLICATE them locally...  Making a local copy will prevent
	// *	memory errors whenever either the original or the copy
	// *	PBS_KeySet object is deleted.
	// *********************************************************************
	PBS_KeySet ( PBS_KeySet & KeySet )
		: table  (KeySet.table),
		  idxCnt (0),
		  keyCnt (0),
		  keys   (NULL)
		{
		// *************************************************************
		// * Call the update method of both objects to synchronise the
		// * number of PBS_Key entries with the appropriate number (as
		// * specified in the PBS_KeywordTable).
		// *************************************************************
		update();
		KeySet.update();
		
		// *************************************************************
		// * Copy the keys from the user provided KeySet to the local
		// * keys objects.
		// *************************************************************
		memcpy(keys, KeySet.keys, KeySet.idxCnt*sizeof(PBS_Key));
		keyCnt = KeySet.keyCnt;
		}
	
	// *********************************************************************
	// * ~PBS_KeySet :
	// *	Destructor for the object, deletes any memory that was allocated
	// *	during construction or usage.
	// *********************************************************************
	~PBS_KeySet ( void )
		{
		free (keys);
		}
		
	// *********************************************************************
	// * update:
	// *	This method should be called prior to performing any comparison
	// *	or insert operations to ensure that the PBS_KeySet has been 
	// *	updated with the most recent size information - and expanded
	// *	if necessary.
	// *********************************************************************
	void update ( void )
		{
		if(idxCnt<table.idxMax)
			{
			keys = (PBS_Key *)(keys?realloc(keys, table.idxMax*sizeof(PBS_Key)):malloc(table.idxMax*sizeof(PBS_Key)));
			while(idxCnt<table.idxMax)
				{
				keys[idxCnt].mask.idx    = idxCnt;
				keys[idxCnt++].mask.bits = 0;
				}
			}
		}
	
	// *********************************************************************
	// * clear :
	// *	This method clears the value of any keys that may have 
	// *	previously been stored in the PBS_KeySet - however, the
	// *	keys unions are not deleted. This facilitates later use and  
	// *	reduces the need for reallocations.
	// *********************************************************************
	void clear ( void )
		{
		for(int idx=0; idx<idxCnt; idx++) keys[idx].mask.bits = 0;
		}
	
	// *********************************************************************
	// * addKey :
	// *	This method adds a key by name.  First the keyname is added to
	// *	the PBS_KeywordTable (if necessary), then the resulting key
	// *	value is added to the keyset.
	// *********************************************************************
	void addKey ( char * keyValue )
		{
		addKey(table.addKeyword(keyValue));
		}
		
		
	// *********************************************************************
	// * addKey :
	// *	This method adds a key - or a collection of keys to a 
	// *	PBS_KeySet. Now, I know that simply performing an |= with the
	// *	bits from the key will add them, however, I take the longer 
	// *	approach because I need to maintain a count of the number of 
	// *	keys in the PBS_KeySet.
	// *********************************************************************
	void addKey ( unsigned keyValue )
		{
		update();

		PBS_Key key;
		key.value = keyValue;
		if(key.mask.idx<idxCnt)
			{
			for(int i=0; i<24; i++)
				{
				int offset = (1<<i);
				
				if((key.mask.bits & offset) && !(keys[key.mask.idx].mask.bits & offset))
					{
					keys[key.mask.idx].mask.bits |= offset;
					keyCnt++;
					}
				}
			}
		}
	
	// *********************************************************************
	// * delKey :
	// *	This method removes a key - or a collection of keys from a 
	// *	PBS_KeySet. The added complication exists to maintain a running
	// *	count of the number of keys that are currently registered in 
	// *	this PBS_KeySet.
	// *********************************************************************
	void delKey ( unsigned keyValue )
		{
		update();

		PBS_Key key;
		key.value = keyValue;
		
		if(key.mask.idx<idxCnt)
			{
			for(int i=0; i<24; i++)
				{
				int offset = (1<<i);
				
				if((key.mask.bits & offset) && (keys[key.mask.idx].mask.bits & offset))
					{
					keys[key.mask.idx].mask.bits ^= offset;
					keyCnt--;
					}
				}
			}
		}

	// *********************************************************************
	// * Considering that the comparison is between a job and a node, 
	// * the following is the basis of the comparison.
	// *
	// * job == node : The node has exactly the same keys that are requested 
	// *               by the job - no more and no less - return 0.
	// * job < node  : The node has all of the keys required by the node,
	// *               and at least one key that was not requested - return 
	// *               a positive value indicating the number of keys in 
	// *               addition to the number requested.
	// * job > node  : The job has requested at least one key that was not
	// *               provided by the node - return the number of keys
	// *               that are not reporesented as a negative value.
	// *********************************************************************
	static int compare ( PBS_KeySet & job, PBS_KeySet & node )
		{
		// *************************************************************
		// * Call the update function of both PBS_KeySets to ensure that
		// * they have been allocated to the appropriate sizes.
		// *************************************************************
		node.update();
		job.update();
		
		// *************************************************************
		// * Get the difference between the number of keys in the job
		// * and the number of keys in the node - if there are fewer
		// * keys in thenode than in the job - we already know it's a
		// * mismatch and we don't need to look any farther.
		// *************************************************************
		int result = node.keyCnt - job.keyCnt;
		
		for(int idx=0; result>=0 && idx<job.table.idxMax; idx++)
			{
			if((node.keys[idx].mask.bits & job.keys[idx].mask.bits)!=job.keys[idx].mask.bits)
				{
				result = -1;
				}
			}
		
		return result;
		}

	// *********************************************************************
	// * asciiDump :
	// *	Produce a diagnostic dump of the PBS_KeySet object to the 
	// *	specified file pointer.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL )
		{
		char   blank = 0;
		size_t prefixLen;
		char   headLine[82];
	
		if(prefix==NULL) prefix = &blank;
		prefixLen = strlen(prefix);
	
		memset(headLine, '-', 80);
		headLine[80]='\n';
		headLine[81]='\0';
		memcpy(headLine, prefix, (prefixLen<=80)?prefixLen:80);
		
		fprintf(fp, "%s        Key    Idx           Bit    Keyword\n", prefix);
		fprintf(fp, headLine);
		for(int idx=0; idx<idxCnt; idx++)
			{
			if(keys[idx].mask.bits!=0)
				{
				for(int bit=0; bit<24; bit++)
					{
					PBS_Key key;
					if((keys[idx].mask.bits & (1<<bit))!=0)
						{
						key.mask.idx = idx;
						key.mask.bits = (1<<bit);
						const char * str = table.getKeyword(key.value);
						if(str) fprintf(fp, "%s %10u    %3u    %10u    %s\n", prefix, key.value, key.mask.idx, key.mask.bits, str); 
						}
					}
				}
			}
		}
};

#endif /* _PBS_KEYWORDTABLE_H_ */
