// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_MomNode.cc
// * Author:      Walt Akers
// * Description: This file contains the source code for the functions defined
// *              in the PBS_MomNode class.
// *****************************************************************************
#include <string.h>
#include <stdlib.h>
#include <sys/param.h>
#include <PBS_MomNode.h>


extern "C" {
#include <pbs_error.h>
#include <pbs_ifl.h>
#include <rm.h>
#include <net_connect.h>
}


// *****************************************************************************
// * class PBS_MomSocketTable :
// *	This is a class used to maintain a list of open connections to 
// *	Machine Oriented Mini-Servers. 
// *****************************************************************************
class PBS_MomSocketTable 
{
protected:
	size_t   entryMax;
	size_t   entryCnt;
	int   *  port;
	char  ** name;

	void insert         ( int idx, char * mom );
	void remove         ( char * mom );
	void remove         ( int idx );
	int  find           ( char * mom );

public:
	PBS_MomSocketTable  ( void );
	~PBS_MomSocketTable ( void );
	int  connect        ( char * mom );
	void disconnect     ( int idx );
} GlobalMomSockets;

// *****************************************************************************
// * PBS_MomSocketTable::PBS_MomSocketTable :
// *	Constructor for the PBS_MomSocketTable.  This method initializes the internal
// *	variables to their default values.
// *****************************************************************************
PBS_MomSocketTable::PBS_MomSocketTable ( void ) 
	: entryMax(0),
	  entryCnt(0),
	  port    (NULL),
	  name    (NULL)
	{
	}
	
// *****************************************************************************
// * PBS_MomSocketTable::~PBS_MomSocketTable :
// *	Destructor for the class.  This method deletes all memory that was
// *	previously allocated by the class.
// *****************************************************************************
PBS_MomSocketTable::~PBS_MomSocketTable ( void )
	{
	if(port) free (port);
	if(name)
		{
		while(entryCnt>0) free (name[--entryCnt]);
		free (name);
		}
	}

// *****************************************************************************
// * PBS_MomSocketTable::insert :
// *	This method allows the caller to insert a new machine oriented
// *	mini-server into the class.  The port and the name must be specified.
// *
// *	If a matching name already exists, then a new record will not be 
// *	inserted.
// *****************************************************************************
void PBS_MomSocketTable::insert ( int idx, char * mom )
	{
	if(idx>=0 && mom && *mom)
		{
		int i;
		for(i=0; i<entryCnt && strcmp(mom, name[i]); i++);
		if(i==entryCnt)
			{
			if((entryCnt+1)>=entryMax)
				{
				entryMax = entryMax?(entryMax*2):16;
				port = (int   *)(port?realloc(port, entryMax*sizeof(int)):malloc(entryMax*sizeof(int)));
				name = (char **)(name?realloc(name, entryMax*sizeof(char *)):malloc(entryMax*sizeof(char *)));
				}
			name[entryCnt]   = strdup(mom);
			port[entryCnt++] = idx;
			}
		}
	}

// *****************************************************************************
// * PBS_MomSocketTable::remove :
// *	This method locates a record in the class that has a matching name and
// *	deletes it.
// *****************************************************************************
void PBS_MomSocketTable::remove ( char * mom )
	{
	if(mom && *mom)
		{
		int i;
		for(i=0; i<entryCnt && strcmp(mom, name[i]); i++);
		if(i<entryCnt)
			{
			free(name[i]);
			if((--entryCnt - i)>0)
				{
				memmove(&port[i], &port[i+1], (entryCnt-i)*sizeof(int));
				memmove(&name[i], &name[i+1], (entryCnt-i)*sizeof(char *));
				}
			}
		}
	} 

// *****************************************************************************
// * PBS_MomSocketTable::remove :
// *	This method locates a record in the class that has a matching name and
// *	deletes it.
// *****************************************************************************
void PBS_MomSocketTable::remove ( int idx )
	{
	int i;
	for(i=0; i<entryCnt && idx!=port[i]; i++);
	if(i<entryCnt)
		{
		free(name[i]);
		if((--entryCnt - i)>0)
			{
			memmove(&port[i], &port[i+1], (entryCnt-i)*sizeof(int));
			memmove(&name[i], &name[i+1], (entryCnt-i)*sizeof(char *));
			}
		}
	}


// *****************************************************************************
// * PBS_MomSocketTable::find :
// *	This method locates the record in the class with a matching name and 
// *	returns the associated file descriptor.  If no matching record is 
// *	found, then -1 is returned.
// *****************************************************************************
int PBS_MomSocketTable::find ( char * mom )
	{
	int i;
	
	for(i=0; i<entryCnt && strcmp(mom, name[i]); i++);
	return (i<entryCnt)?port[i]:-1;
	}

// *****************************************************************************
// * PBS_MomSocketTable::connect :
// *	This method generates a connection to a machine oriented mini-server
// *	(if it doesn't already exist) and returns the file descriptor or 
// *	-1 if a connection could not be established.
// *****************************************************************************
int PBS_MomSocketTable::connect ( char * mom )
	{
	int idx = -1;
	
	if(mom && *mom)
		{
		if((idx=find(mom))<0 && (idx = openrm(mom, 0))>=0)
			{
			insert(idx, mom);
			}
		}
	return idx;
	}


// *****************************************************************************
// * PBS_MomSocketTable::disconnect :
// *	This method removes a connection to a machine oriented mini-server and
// *	closes the socket.
// *****************************************************************************
void PBS_MomSocketTable::disconnect ( int idx )
	{
	int i;
	for(i=0; i<entryCnt && port[i]!=idx; i++);
	if(i<entryCnt)
		{
		closerm(idx);
		remove (idx);
		}
	}


// *****************************************************************************
// * PBS_MomNode::defaultNameList :
// *	This is the list of default property names that can be requested from
// *	any of the Machine Oriented Mini-Servers (MOMs) on the PBS Node. 
// *****************************************************************************
const char * PBS_MomNode::defaultNameList[PBS_MomNode::DEFAULT_NAME_CNT] = 
	{
	"arch",
	"cput",
	"idletime",
	"loadave",
	"mem",
	"ncpus",
	"nsessions",
	"nusers",
	"physmem",
	"sessions",
	"size",
	"uname",
	"walltime"
	};

// *****************************************************************************
// * PBS_MomNode::PBS_MomNode :
// *	This is the constructor for the PBS_MomNode class.  It initializes
// *	internal variables.  The values are not populated until an explicit 
// *	call to loadData is made.
// *****************************************************************************
PBS_MomNode::PBS_MomNode ( char * nodeName )
	: name       (NULL),
	  entryCnt   (0),
	  nameLst    (NULL),
	  valueLst   (NULL),
	  errorState (0)
	{
	name     = strdup(nodeName);
	nameLst  = (char **)malloc(DEFAULT_NAME_CNT*sizeof(char *));
	valueLst = (char **)malloc(DEFAULT_NAME_CNT*sizeof(char *));
	for(entryCnt=0; entryCnt<DEFAULT_NAME_CNT; entryCnt++)
		{
		nameLst[entryCnt]  = strdup(defaultNameList[entryCnt]);
		valueLst[entryCnt] = NULL;
		}
	}

// *****************************************************************************
// * PBS_MomNode::PBS_MomNode :
// *	This is the destructor for the PBS_MomNode class.  It frees
// *	any memory previously allocated to the classes objects.
// *****************************************************************************
PBS_MomNode::~PBS_MomNode ( void )
	{
	while(entryCnt>0)
		{
		if(nameLst[--entryCnt]!=NULL) free(nameLst[entryCnt]);
		if(valueLst[entryCnt]!=NULL)  free(valueLst[entryCnt]);
		}
	if(name)     free(name);
	if(nameLst)  free(nameLst);
	if(valueLst) free(valueLst);
	}


// *****************************************************************************
// * PBS_MomNode::loadData :
// *	This method connects to the Machine Oriented Mini-Server on the PBS
// *	node and requests all of the property names that are listed in its
// *	namelst array.  The corresponding values are stored in the valueLst 
// *	array.
// *****************************************************************************
void PBS_MomNode::loadData ( void )
	{
	int mom_sd = GlobalMomSockets.connect(name);

	if(mom_sd>=0)
		{
		int i;
		
		clear();
		fullresp(0);
		errorState = 0;
		
		for(i=0; i<entryCnt; i++)
			{
			addreq(mom_sd, (char *)nameLst[i]);
			}

		for(i=0; i<entryCnt; i++)
			{
			char * mom_ans = getreq(mom_sd);
			if(mom_ans!=NULL)
				{
				if(*mom_ans && *mom_ans!='?')
					{
					valueLst[i] = mom_ans;
					}
				else free(mom_ans);
				}
			else if(pbs_errno)
				{
				clear();
				errorState = pbs_errno;
				GlobalMomSockets.disconnect(mom_sd);
				return;
				}
			}
		}
	}


// *****************************************************************************
// * PBS_MomNode::clear :
// *	This method clears the values that are currently stored in the object 
// *	so that new values can be requested from the node.
// *****************************************************************************
void PBS_MomNode::clear ( void )
	{
	for(int i=0; i<entryCnt; i++)
		{
		if(valueLst[i])
			{
			free(valueLst[i]);
			valueLst[i] = NULL;
			}
		}
	}


// *****************************************************************************
// * PBS_MomNode::findProperty :
// *	This method locates a property name from the nameLst and returns the
// *	value associated with it.  If the property name does not exist, then
// *	it is added to the list and the data is reloaded.
// *****************************************************************************
char * PBS_MomNode::findProperty ( char * property )
	{
	for(int i=0; i<entryCnt; i++)
		{
		if(nameLst[i] && !strcmp(nameLst[i], property))
			return valueLst[i];
		}

	nameLst = (char **) realloc(nameLst, (entryCnt+1)*sizeof(char *));
	nameLst[entryCnt] = strdup(property);
	valueLst = (char **) realloc(valueLst, (entryCnt+1)*sizeof(char *));
	valueLst[entryCnt++] = NULL;
	
	loadData();
	
	return valueLst[entryCnt-1];
	}

// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned. 
// *
// *	If the string retrieved from the node is too long to fit in the user
// *	provided buffer, then the value is truncated.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, char * value, int len )
	{
	char * stringval = findProperty(prop);

	if(value && len>0)
		{
		*value=0;
		if(stringval && *stringval)
			{
			strncpy(value, stringval, len);
			value[len-1] = 0;
			}
		}
	
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *
// *	If successful, this method will allocate sufficnet space to the value
// *	pointer to store the result.  The caller is responsible for freeing
// *	this memory when it is no longer needed.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, char ** value )
	{
	char * stringval = findProperty(prop);
	*value           = NULL;

	if(stringval && *stringval)
		{
		*value = strdup(stringval);
		}
	
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, int & value )
	{
	char * stringval = findProperty(prop);
	value            = 0;
	
	if(stringval && *stringval)
		{
		value = atoi(stringval);
		}
		
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, unsigned & value )
	{
	char * stringval = findProperty(prop);
	value            = 0;
	
	if(stringval && *stringval)
		{
		value = atoi(stringval);
		}
		
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, long & value )
	{
	char * stringval = findProperty(prop);
	value            = 0;
	
	if(stringval && *stringval)
		{
		value = atoi(stringval);
		}
		
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, unsigned long & value )
	{
	char * stringval = findProperty(prop);
	value            = 0;
	
	if(stringval && *stringval)
		{
		value = atoi(stringval);
		}
		
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, float & value )
	{
	char * stringval = findProperty(prop);
	value            = 0;
	
	if(stringval && *stringval)
		{
		value = atof(stringval);
		}
		
	return stringval?0:-1;
	}
	
// *****************************************************************************
// * PBS_MomNode::getProperty :
// *	Retrieves the specific property value in the user specified type. A zero
// *	is returned if the property has been read from the node and can be 
// *	retrieved.  Otherwise, a non-zero is returned.
// *****************************************************************************
int PBS_MomNode::getProperty ( char * prop, double & value )
	{
	char * stringval = findProperty(prop);
	value            = 0;
	
	if(stringval && *stringval)
		{
		value = atof(stringval);
		}
		
	return stringval?0:-1;
	}

