// *****************************************************************************
// * Copyright (c) 1999      Southeastern Universities Research Association,
// *                         Thomas Jefferson National Accelerator Facility
// * 
// * This software was developed under a United States Government license
// * described in the NOTICE file included as part of this distribution.
// *****************************************************************************
// *
// * File:        PBS_Node.cc
// * Author:      Walt Akers
// * Description: This file contains the source code for the functions defined
// *              in the PBS_Node class.
// *****************************************************************************
#include <PBS_Node.h>
#include <PBS_Server.h>

extern "C" {
#include <rm.h>
#include <net_connect.h>
}

// *****************************************************************************
// * These are the name string sassociated with the NodeType enumeration.
// *****************************************************************************
char * PBS_Node::NodeTypeList[3] = 
	{
	"Unknown",
	"Cluster",
	"Timeshared"
	};

// *****************************************************************************
// * These are the name string sassociated with the NodeState enumeration.
// *****************************************************************************
char * PBS_Node::NodeStateList[7] = 
	{
	"Unknown",
	"Free",
	"Down",
	"Offline",
	"Reserved",
	"Job-Exclusive",
	"Job-Shared"
	};

// *****************************************************************************
// * PBS_Node::readNodeTypeValue :
// *	This method will read a NodeType value from  the attrib structure
// *	that was returned as part of the pbs_statenode function.
// *****************************************************************************
void PBS_Node::readNodeTypeValue ( NodeType & val, struct attrl * attrib )
	{
	if(!strcmp(attrib->value, "cluster")) val = CLUSTER;
	else if(!strcmp(attrib->value, "time-shared")) val = TIMESHARED;
	else val = UNKNOWN_TYPE;
	}
	
// *****************************************************************************
// * PBS_Node::readNodeTypeValue :
// *	This method will read a NodeType value from  the attrib structure
// *	that was returned as part of the pbs_statenode function.
// *****************************************************************************
void PBS_Node::readNodeStateValue ( NodeState & val, struct attrl * attrib )
	{
	if(!strcmp(attrib->value, "free")) val = FREE;
	else if(!strcmp(attrib->value, "down")) val = DOWN;
	else if(!strcmp(attrib->value, "offline")) val = OFFLINE;	
	else if(!strcmp(attrib->value, "reserve")) val = RESERVED;	
	else if(!strcmp(attrib->value, "job-exclusive")) val = JOB_EXCLUSIVE;	
	else if(!strcmp(attrib->value, "job-sharing")) val = JOB_SHARING;	
	else val = UNKNOWN_STATE;
	}

// *****************************************************************************
// * PBS_Node::loadData :
// *	This method can be called by the constructor or independently to cause
// *	the object to be populated with new node data.
// *
// *	The batch_status object is retrieved by calling the pbs_statque function
// *	using the connection ID of the server.  Each attribute of that object
// *	is then examined and the value is loaded into this class.
// *****************************************************************************
int PBS_Node::loadData ( struct batch_status * bstat )
	{
	struct attrl        * attrib;

	// *********************************************************************
	// * Call clear to clear the contents of the object.
	// *********************************************************************
	clear();

	// *********************************************************************
	// * Copy the name of the node into the name variable.
	// *********************************************************************
	name = strdup(bstat->name);

	// *********************************************************************
	// * Process the contents of each of the attributes that was returned
	// * as part of the batch_status object.
	// *********************************************************************
	for(attrib=bstat->attribs; attrib!=NULL; attrib = attrib->next)
		{
		switch(*attrib->name)
			{
			case 'j':
			if(!strcmp(attrib->name, "jobs")) readListValue(jobs, attrib);
			else LogError("PBS_Node::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'n':
			if(!strcmp(attrib->name, "ntype")) readNodeTypeValue(ntype, attrib);
			else if(!strcmp(attrib->name, "np")) readIntegerValue(np, attrib);
			else LogError("PBS_Node::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 'p':
			if(!strcmp(attrib->name, "properties")) readListValue(keywords, attrib);
			else LogError("PBS_Node::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			case 's':
			if(!strcmp(attrib->name, "state")) readNodeStateValue(state, attrib);
			else LogError("PBS_Node::loadData", "Didn't process attribute %s", attrib->name);
			break;
			
			default:
			LogError("PBS_Node::loadData", "Didn't process attribute %s", attrib->name);
			break;
			}
		}

	// *********************************************************************
	// * Features requiring communications with the individual MOMs have 
	// * been removed to improve performance and reliability
	// *********************************************************************
	// properties.loadData();
	
	return 0;
	}

// *****************************************************************************
// * PBS_Node::clear :
// *	This method releases all memory that was allocated to the object and 
// *	then restores all internal variables to their default values.  This 
// *	method is generally called as part of object destruction, or to clear
// *	values so they can be repopulated from the server.
// *****************************************************************************
void PBS_Node::clear ( void )
	{
	nullify(name);
	nullify(jobs);
	nullify(keywords);
	
	// *********************************************************************
	// * Features requiring communications with the individual MOMs have 
	// * been removed to improve performance and reliability
	// *********************************************************************
	// properties.clear();
	
	np    = 1;
	ntype = UNKNOWN_TYPE;
	state = UNKNOWN_STATE;
	}


// *****************************************************************************
// * PBS_Node::asciiDump :
// *	This method generates a plain text report of the contents of the
// *	object.  This method is generally used for debugging, however, the
// *	information is also useful for reporting and verifying changes made
// *	from other sources.  
// * 
// * 	The prefix string will be prepended to each string that is outputted 
// *	by this method.  It is a mechanism to provide cohesive formatting for
// *	objects that contain other objects.
// *****************************************************************************
void PBS_Node::asciiDump ( FILE * fp )
	{
	int i     = 0;
	int first = 0;
	
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "NODE %s\n", name);
	fprintf(fp, "-----------------------------------------------------\n");
	fprintf(fp, "NODE CONFIGURATION\n");
	fprintf(fp, "     Server:        %s\n", server.getName());
	fprintf(fp, "     Name:          %s\n", name);
	fprintf(fp, "     Processors:    %i\n", np);
	fprintf(fp, "     Type:          %s\n", NodeTypeList[ntype]);
	fprintf(fp, "     State:         %s\n", NodeStateList[state]);

	// *********************************************************************
	// * Write the list of keywords - if any.
	// *********************************************************************
	fprintf(fp, "     Keywords:");
	if(keywords && keywords[0])
		{
		for(i=0; keywords[i]!=NULL; i++)
			fprintf(fp, "%*c%s\n", (i?20:6), 0x20, keywords[i]);
		}
	else fprintf(fp, "\n");

	// *********************************************************************
	// * Features requiring communications with the individual MOMs have 
	// * been removed to improve performance and reliability
	// *********************************************************************
	// *********************************************************************
	// * Write the list of properties - if any.
	// *********************************************************************
	// fprintf(fp, "     Properties:");
	// for(i=0, first=0; i<properties.entryCnt; i++)
	// 	{
	// 	if(properties.valueLst[i]!=NULL)
	// 		fprintf(fp, "%*c%s = %s\n", (first++)?20:4, 0x20, 
	// 			properties.nameLst [i],
	// 			properties.valueLst[i]);
	// 	}
	// if(!first) fprintf(fp, "\n");
	
	// *********************************************************************
	// * Write the list of jobs - if any.
	// *********************************************************************
	fprintf(fp, "     Jobs:");
	if(jobs && jobs[0])
		{
		for(i=0; jobs[i]!=NULL; i++)
			fprintf(fp, "%*c%s\n", (i?20:10), 0x20, jobs[i]);
		}
	else fprintf(fp, "\n");

	// *********************************************************************
	// * Insert a carriage return an flush the buffer before finishing.
	// *********************************************************************
	fprintf(fp, "\n");
	fflush(fp);
	}
	
// *****************************************************************************
// * PBS_NodeList::PBS_NodeList :
// *	Default constructor for the class.  It initializes all internal
// *	variables to their default values.
// *****************************************************************************
PBS_NodeList::PBS_NodeList ( void )
	: nodes(NULL),
	  nodeCnt(0)
	{
	}

// *****************************************************************************
// * PBS_NodeList::PBS_NodeList :
// *	Parameterized constructor for the class.  It initializes the
// *	internal variables using an array of nodes that is terminated
// *	by a NULL pointer.
// *
// *	The array becomes the property of the class and should not be
// *	accessed directly by the caller.
// *****************************************************************************
PBS_NodeList::PBS_NodeList ( PBS_Node ** nodelist )
	: nodes  (NULL),
	  nodeCnt(0)
	{
	addNodes ( nodelist );
	}

// *****************************************************************************
// * PBS_NodeList::~PBS_NodeList :
// *	Destructor for the object.  It calls clear to free the array structure
// *	associated with the object. The PBS_Server is responsible for 
// *	deleteing the individual PBS_Nodes.
// *****************************************************************************
PBS_NodeList::~PBS_NodeList ( void )
	{
	if(nodes) free(nodes);
	nodes   = NULL;
	nodeCnt = 0;
	}

// *****************************************************************************
// * PBS_NodeList::addNodes :
// *	This method populates the object using an external array of 
// *	nodes that is termimated by a NULL pointer.  The array becomes 
// *	the property of the class and should not be accessed directly 
// *	by the caller.
// *****************************************************************************
void PBS_NodeList::addNodes ( PBS_Node ** nodelist )
	{
	if(nodes) free(nodes);
	nodes   = nodelist;
	nodeCnt = 0;

	if(nodes!=NULL) while(nodes[nodeCnt]!=NULL) nodeCnt++;
	}
		
// *****************************************************************************
// * PBS_NodeList::getNode :
// *	This method returns a specific node by name.  If the node doesn't
// *	exist in the list, then a NULL will be returned.
// *****************************************************************************
PBS_Node * PBS_NodeList::getNode ( char * name )
	{
	PBS_Node * result = NULL;
	int idx;
		
	if(name && *name)
		{
		for(idx=0; idx<nodeCnt && strcmp(nodes[idx]->getName(), name); idx++);
		if(idx<nodeCnt) result = nodes[idx];
		}
	
	return result;
	}
	
// *****************************************************************************
// * PBS_NodeList::getNodeCount :
// *	Returns a count of all nodes that are currently in the list.
// *****************************************************************************
size_t PBS_NodeList::getNodeCount ( void )
	{
	return nodeCnt;
	}

// *****************************************************************************
// * PBS_NodeList::getNodes :
// *	This method returns an array of all nodes that are available in 
// *	the list.  The user is responsible for deleteing the array 
// *	that is allocated by this call.
// *****************************************************************************
PBS_Node ** PBS_NodeList::getNodes ( void )
	{
	PBS_Node ** allNodes = nodeCnt?(new PBS_Node *[nodeCnt+1]):NULL;
	if(nodeCnt)
		{
		memcpy(allNodes, nodes, nodeCnt*sizeof(PBS_Node *));
		allNodes[nodeCnt] = NULL;
		}
	return allNodes;	
	}

// *****************************************************************************
// * PBS_NodeList::getNodesByState :
// *	This method returns an array of all nodes whose state is the same as 
// *	specified in the NodeState variable.
// *****************************************************************************
PBS_Node ** PBS_NodeList::getNodesByState ( PBS_Node::NodeState state )
	{
	PBS_Node **  selectNodes = NULL;
	int          selectCnt   = 0;
	size_t       selectIdx;
	size_t       allIdx;	
	
	for(allIdx=0; allIdx<nodeCnt; allIdx++)
		{
		if(nodes[allIdx] && nodes[allIdx]->getState()==state) 
			selectCnt++;
		}

	if(selectCnt>0)
		{
		selectNodes = new PBS_Node * [selectCnt+1];

		for(allIdx=0, selectIdx=0; 
		    allIdx<nodeCnt && selectIdx<selectCnt; 
		    allIdx++)
			{
			if(nodes[allIdx] && nodes[allIdx]->getState()==state)
				{
				selectNodes[selectIdx++] = nodes[allIdx];
				}
			}
		selectNodes[selectIdx] = NULL;
		}

	return selectNodes;
	}

// *****************************************************************************
// * PBS_NodeList::getNodesByType :
// *	This method returns an array of all nodes whose type is the same as 
// *	specified in the NodeType variable.
// *****************************************************************************
PBS_Node ** PBS_NodeList::getNodesByType ( PBS_Node::NodeType type )
	{
	PBS_Node **  selectNodes = NULL;
	int          selectCnt   = 0;
	size_t       selectIdx;
	size_t       allIdx;	
	
	for(allIdx=0; allIdx<nodeCnt; allIdx++)
		{
		if(nodes[allIdx] && nodes[allIdx]->getType()==type) 
			selectCnt++;
		}

	if(selectCnt>0)
		{
		selectNodes = new PBS_Node * [selectCnt+1];

		for(allIdx=0, selectIdx=0; 
		    allIdx<nodeCnt && selectIdx<selectCnt; 
		    allIdx++)
			{
			if(nodes[allIdx] && nodes[allIdx]->getType()==type)
				{
				selectNodes[selectIdx++] = nodes[allIdx];
				}
			}
		selectNodes[selectIdx] = NULL;
		}

	return selectNodes;
	}
	
// *****************************************************************************
// * PBS_NodeList::asciiDump :
// *	Sequentially dumps all of the nodes in the list.
// *****************************************************************************
void PBS_NodeList::asciiDump ( FILE * fp )
	{
	for(int i=0; i<nodeCnt; i++)
		{
		if(nodes[i]) nodes[i]->asciiDump(fp);
		}
	}
	
