// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distributioon.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	NodeCommodity.cc realizes the methods defined within the NodeCommodity
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <Brokerage.h>		// Included for NodeCommodity definition
#include <ClusterNode.h>	// Included for ClusterNode class Definition
#include <stdio.h>		// Include for FILE structure and stdio

// *****************************************************************************
// * NodeCommodity::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the NodeCommodity class.
// *****************************************************************************
NodeCommodity * NodeCommodity::freeList;

// *****************************************************************************
// * operator NodeCommodity::new :
// *	This operator is called whenever a new NodeCommodity object is 
// *	requested. The operator will first attempt to extract a NodeCommodity
// *	object from those preallocated on the freeList.  If no NodeCommodity
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * NodeCommodity::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new NodeCommodity[32];
		for(i=0; i<31; i++) freeList[i].setNext(&freeList[i+1]);
		freeList[i].setNext(NULL);
		}
	NodeCommodity * retval = freeList;
	freeList = retval->getNext();
	retval->setNext(NULL);
	
	return (void *)retval;
	}
	
// *****************************************************************************
// * operator NodeCommodity::delete :
// *	This is the delete operator for the NodeCommodity class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void NodeCommodity::operator delete ( void * ptr )
	{
	((NodeCommodity *)ptr)->setNext(freeList);
	freeList = (NodeCommodity *)ptr;
	}
	
	
// *****************************************************************************
// * NodeCommodity::NodeCommodity:
// *	This is a private constructor for the NodeCommodity object.  It will
// *	be called each time the NodeCommodity::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
NodeCommodity::NodeCommodity ( void )
	: next    (NULL),
	  nodeType(0),
	  node    (NULL),
	  holder  (NULL)
	{
	}
	
// *****************************************************************************
// * NodeCommodity::NodeCommodity :
// *	This is the public constructor for the NodeCommodity object. It will
// *	be called when the user needs to allocate a new NodeCommodity object.
// *****************************************************************************
NodeCommodity::NodeCommodity ( ClusterNode & Node, size_t Type, time_t start, time_t duration )
	: next        (NULL),
	  node        (&Node),
	  nodeType    (Type),
	  rangeStart  (start),
	  jobDuration (duration),
	  weight      (0.0),
	  holder      (NULL)
	{
	}
		
// *****************************************************************************
// * NodeCommodity::~NodeCommodity :
// *	This is the destructor for the NodeCommodity object. Since the object
// *	does not allocate any memory, this method is used as a placeholder only.
// *****************************************************************************
NodeCommodity::~NodeCommodity ( void )
	{
	}

// *****************************************************************************
// * NodeCommodity::computeWeight :
// *	This method will used the information provided to compute the weight of 
// *	this node.  The nodes weight will then be used to establish it's 
// *	position in the list of nodes.
// *
// *	You know, could do alot of exotic stuff here... but for the sake of 
// *	simplicity I am simply going to add the gapSize (always an integer>=0),
// *	to the utilization (always a double<=1.0) to create the weight.
// *
// *	Using this approach the amount of space lost beneath a scheduled job
// *	is always the driving term and the utilization is always secondary and
// *	will only be considered when the gap size is equal.  
// *****************************************************************************
double NodeCommodity::computeWeight ( size_t procCnt )
	{
	double gapSize = node->getGapSize (procCnt, rangeStart, jobDuration);
	double utilization;
	
	// *********************************************************************
	// * The utilization flag should be inverted based on the System 
	// * Administrators desire to either (TRUE: spread the jobs evenly 
	// * across all nodes) or (FALSE: Concentrate as many jobs as possible
	// * on each node).
	// *********************************************************************
	if(SchedStatic::getBalanceFlag()) 
		{
		utilization = node->getUtilization(rangeStart, jobDuration);
		}
	else utilization = 1.0 - node->getUtilization(rangeStart, jobDuration);
	
	weight = utilization + gapSize*(double)(procCnt*(rangeStart-SchedStatic::getNow()));
	
	return weight;
	}
	
// *****************************************************************************
// * NodeCommodity::asciiDump :
// *	This method is called to generate an ascii text report of the contents
// *	of the NodeCommodity object.  This data will be reported to the file
// *	descriptor specified in the fp parameter.  Each line of the report will
// *	be prefixed by the string specified in the prefix parameter. The 
// *	compressed flag indicates that all data should be reported to the same
// *	output line.
// *****************************************************************************
void NodeCommodity::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	
	fprintf(fp, "%sNodeCommodity for node %s\n", 
		prefixString, node->getName());
	fprintf(fp, "%s   => Index                : %i\n", 
		prefixString, node->getIndex());
	fprintf(fp, "%s   => Range Start Time     : %s", 
		prefixString, ctime(&rangeStart));
	fprintf(fp, "%s   => Job Duration         : %d:%02d:%02d hours\n",
		prefixString, jobDuration/3600, (jobDuration%3600)/60, jobDuration%60);
	fprintf(fp, "%s   => Node Weight          : %g\n",
		prefixString, weight);
	fprintf(fp, "%s   => Node Types Supported :\n",
		prefixString);
	for(int idx=0; idx<SchedStatic::MAXSUBSETS; idx++)
		{
		if(hasType(idx))
			{
			fprintf(fp, "%s      => %i\n", prefixString, idx);
			}
		}
	}
	
// *****************************************************************************
// * NodeBroker::NodeBroker :
// *	This is the public constructor for the object. It will 
// *	initialize all internal variables.
// *****************************************************************************
NodeBroker::NodeBroker ( void )
	: procCnt(0)
	{
	clear();
	}

// *****************************************************************************
// * NodeBroker::~NodeBroker :
// *	This is the destructor for the NodeBroker object.  It currently
// *	performs no real operations.
// *****************************************************************************
NodeBroker::~NodeBroker ( void )
	{
	clear ();
	}

// *****************************************************************************
// * NodeBroker::issueNode :
// *	This method is used to give a node to the NodeBroker object. The
// *	method will install the NodeCommodity object to the end of the
// *	list, will update the typeMask to represent all capabilities
// *	that are provided by this object and will increment the 
// *	nodeCnt variable.
// *****************************************************************************
void NodeBroker::issueNode ( NodeCommodity & node )
	{
	// *********************************************************************
	// * First logically OR the local typeMask with the capabilities
	// * provided by the node.  This will be used to express all of the 
	// * capabilties that are currently available by NodeCommodities
	// * within this NodeBroker.
	// *
	// * NOTE : The typeMask provided by the NodeCommodity has already
	// *        been converted from subset indices to a bitmask... Direct
	// *	    ORing is allowed.
	// *********************************************************************
	typeMask |= node.getType();
	
	// *********************************************************************
	// * Set the holder for the NodeCommodity to point to this 
	// * NodeBroker object.
	// *********************************************************************
	node.setHolder(this);

	// *********************************************************************
	// * Set the next pointer for the NodeCommodity to NULL.
	// *********************************************************************
	node.setNext(NULL);

	// *********************************************************************
	// * Next, increment the nodeCnt to reflect the presence of a new node.
	// *********************************************************************
	nodeCnt++;
	
	// *********************************************************************
	// * Finally, add the new node to the end of the list.  You'll note
	// * that since the MasterBroker sorts the nodes upon insertion and 
	// * extraction, that it is unnecesary to recompute the weight of the
	// * node to insert it into this list.
	// *********************************************************************
	if(nodes==NULL) nodes = &node;
	else            nodeTail->setNext(&node);
	nodeTail = &node;
	}


// *****************************************************************************
// * NodeBroker::requestNode :
// *	This method will search through the list of nodes to find one
// *	that represents a specific type.  If the client has this type
// *	of node, then it will provide it to the requestor - otherwise,
// *	a NULL is returned.
// *****************************************************************************
NodeCommodity * NodeBroker::requestNode ( size_t type )
	{
	NodeCommodity * result = NULL;
	
	if(type<SchedStatic::MAXSUBSETS && hasType(type))
		{
		NodeCommodity * prevPtr = NULL;
		NodeCommodity * currPtr = nodes;
		
		// *************************************************************
		// * Set the typeMask to 0 and recalculate it as you walk 
		// * through the list until you find a matching node.
		// *************************************************************
		typeMask = 0;
		while(currPtr && !currPtr->hasType(type))
			{
			// *****************************************************
			// * NOTE : The typeMask provided by the NodeCommodity 
			// *        has already been converted from subset 
			// *        indices to a bitmask... Direct ORing is 
			// *        allowed.
			// *****************************************************
			typeMask |= currPtr->getType();
			prevPtr   = currPtr;
			currPtr   = prevPtr->getNext();
			}
		// *************************************************************
		// * If a matching node was found, remove it from the list
		// * and place its address in the result variable.
		// *************************************************************
		if(currPtr)
			{
			result = currPtr; 
			nodeCnt--;
			if(prevPtr) prevPtr->setNext(currPtr->getNext());
			else nodes = currPtr->getNext();
			currPtr = currPtr->getNext();

			// *****************************************************
			// * Set the holder for the NodeCommodity to NULL - 
			// * indicating that the Commodity is an orphan.
			// *****************************************************
			result->setHolder(NULL);
			
			// *****************************************************
			// * Set the next pointer for the NodeCommodity to NULL.
			// *****************************************************
			result->setNext(NULL);
			}
		// *************************************************************
		// * Continue walking through the list until you have 
		// * obtained the typeMasks associated with all of the 
		// * nodes that remain inside of the NodeBroker object.
		// *************************************************************
		while(currPtr!=NULL)
			{
			// *****************************************************
			// * NOTE : The typeMask provided by the NodeCommodity 
			// *        has already been converted from subset 
			// *        indices to a bitmask... Direct ORing is 
			// *        allowed.
			// *****************************************************
			typeMask |= currPtr->getType();
			currPtr = currPtr->getNext();
			}
		}

	// *********************************************************************
	// * Return the NodeCommodity that was obtained from the list.
	// *********************************************************************
	return result;
	}

// *****************************************************************************
// * NodeBroker::revokeNode :
// *	This method should only be called when the type associated with
// *	a NodeCommodity has been drastically changed.  This will walk
// *	through the list of NodeCommodities to find the one that has
// *	the specified index and will remove it from the list.
// *****************************************************************************
NodeCommodity * NodeBroker::revokeNode ( size_t nodeIdx )
	{
	NodeCommodity * result = NULL;
	
	NodeCommodity * prevPtr = NULL;
	NodeCommodity * currPtr = nodes;
		
	// *********************************************************************
	// * Set the typeMask to 0 and recalculate it as you walk 
	// * through the list until you find a matching node.
	// *********************************************************************
	typeMask = 0;
	while(currPtr && currPtr->getNode().getIndex()!=nodeIdx)
		{
		// *************************************************************
		// * NOTE : The typeMask provided by the NodeCommodity 
		// *        has already been converted from subset 
		// *        indices to a bitmask... Direct ORing is 
		// *        allowed.
		// *************************************************************
		typeMask |= currPtr->getType();
		prevPtr   = currPtr;
		currPtr   = prevPtr->getNext();
		}
	// *********************************************************************
	// * If a matching node was found, remove it from the list
	// * and place its address in the result variable.
	// *********************************************************************
	if(currPtr)
		{
		result = currPtr; 
		nodeCnt--;
		if(prevPtr) prevPtr->setNext(currPtr->getNext());
		else nodes = currPtr->getNext();
		currPtr = currPtr->getNext();

		// *************************************************************
		// * Set the holder for the NodeCommodity to NULL - 
		// * indicating that the Commodity is an orphan.
		// *************************************************************
		result->setHolder(NULL);

		// *************************************************************
		// * Set the next pointer for the NodeCommodity to NULL.
		// *************************************************************
		result->setNext(NULL);
		}
	// *********************************************************************
	// * Continue walking through the list until you have 
	// * obtained the typeMasks associated with all of the 
	// * nodes that remain inside of the NodeBroker object.
	// *********************************************************************
	while(currPtr!=NULL)
		{
		// *************************************************************
		// * NOTE : The typeMask provided by the NodeCommodity 
		// *        has already been converted from subset 
		// *        indices to a bitmask... Direct ORing is 
		// *        allowed.
		// *************************************************************
		typeMask |= currPtr->getType();
		currPtr = currPtr->getNext();
		}

	// *********************************************************************
	// * Return the NodeCommodity that was obtained from the list.
	// *********************************************************************
	return result;
	}
	
// *****************************************************************************
// * NodeBroker::refresh :
// *	Because a NodeCommodity's characteristics may be modified 
// *	externally, each of the brokers must be refreshed prior to
// *	use to ensure that their typeMask is up to date.
// *****************************************************************************
void NodeBroker::refresh ( void )
	{
	nodeCnt  = 0;
	typeMask = 0;
	
	for(NodeCommodity * node = nodes; node!=NULL; node=node->getNext())
		{
		nodeCnt++;
		// *************************************************************
		// * NOTE : The typeMask provided by the NodeCommodity 
		// *        has already been converted from subset 
		// *        indices to a bitmask... Direct ORing is 
		// *        allowed.
		// *************************************************************
		typeMask |= node->getType();
		}
	}
	
// ****************************************************************************
// * NodeBroker::asciiDump :
// *	This method allows the caller to generate an ascii text
// *	diagnostic dump of the contents of the NodeBroker object. 
// ****************************************************************************
void NodeBroker::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;

	fprintf(fp, "%sNodeBroker for Node Type %i\n", 
		prefixString, typeIndex);
	fprintf(fp, "%s   => Node Count     : %i\n", 
		prefixString, nodeCnt);
	fprintf(fp, "%s   => Nodes Required : %i\n",
		prefixString, nodeReq);
	fprintf(fp, "%s   => Procs Per Node : %i\n",
		prefixString, procCnt);
	fprintf(fp, "%s   => Nodes :\n",
		prefixString );

	for(NodeCommodity * node = nodes; node!=NULL; node=node->getNext())
	
	for(int idx=0; idx<SchedStatic::MAXSUBSETS; idx++)
		{
		fprintf(fp, "%s         %s\n", prefixString, node->getNode().getName());
		}
	}


// *****************************************************************************
// * MasterBroker::MasterBroker :
// *	This is the initializer for the MasterBroker object.
// *****************************************************************************
MasterBroker::MasterBroker ( void )
	: NodeBroker()
	{
	}

// *****************************************************************************
// * MasterBroker::issueNode :
// *	This method is used to give a node to the MasterBroker object. The
// *	method will install the NodeCommodity object to the end of the
// *	list, will update the typeMask to represent all capabilities
// *	that are provided by this object and will increment the 
// *	nodeCnt variable.
// *****************************************************************************
void MasterBroker::issueNode ( NodeCommodity & node )
	{
	// *********************************************************************
	// * First logically OR the local typeMask with the capabilities
	// * provided by the node.  This will be used to express all of the 
	// * capabilties that are currently available by NodeCommodities
	// * within this MasterBroker.
	// *
	// * NOTE : The typeMask provided by the NodeCommodity has already
	// *        been converted from subset indices to a bitmask... Direct
	// *	    ORing is allowed.
	// *********************************************************************
	typeMask |= node.getType();
	
	// *********************************************************************
	// * Set the holder for the NodeCommodity to point to this 
	// * MasterBroker object.
	// *********************************************************************
	node.setHolder(this);

	// *********************************************************************
	// * Set the next pointer for the NodeCommodity to NULL.
	// *********************************************************************
	node.setNext(NULL);

	// *********************************************************************
	// * Next, increment the nodeCnt to reflect the presence of a new node.
	// *********************************************************************
	nodeCnt++;
	
	// *********************************************************************
	// * Here is where this code differs from the regular NodeBroker.  It 
	// * will use the current processorCnt - if one exists - to compute
	// * the weight of this NodeCommodity.  If the result is valid, then the
	// * Node will be inserted into the list at the correct location.
	// *
	// * Otherwise, the node will be appended to the end of the list.
	// *********************************************************************
	if(procCnt && node.computeWeight(procCnt)>=0)
		{
		NodeCommodity * curr = nodes;
		NodeCommodity * prev = NULL;
		
		while(curr && node.getWeight() > curr->getWeight())
			{
			prev = curr;
			curr = curr->getNext();
			}
		if(prev) prev->setNext(&node);
		else     nodes = &node;
		node.setNext(curr);
		if(!curr) nodeTail = &node;
		}
	// *********************************************************************
	// * If the weight for this node cannot be computed or is invalid, 
	// * simply apply it to the end of the list and continue.
	// *********************************************************************
	else	{
		if(nodes==NULL) nodes = &node;
		else            nodeTail->setNext(&node);
		nodeTail = &node;
		}
	}


// *****************************************************************************
// * MasterBroker::requestNode :
// *	This method will sort the list of NodeCommodities based on the number
// *	of processors requested (if necessary) and will then search through
// *	the nodes list to find the first node with a matching type.
// *****************************************************************************
NodeCommodity * MasterBroker::requestNode ( size_t type, size_t nProcs )
	{
	// *********************************************************************
	// * In the event that the number of processors specified is not equal 
	// * to the processor count... It will be necessary to resort the list
	// * based on the updated number of processors.
	// *********************************************************************
	if(nProcs!=procCnt)
		{
		NodeCommodity * sorted     = NULL;
		NodeCommodity * sortedTail = NULL;
		NodeCommodity * node;
		NodeCommodity * curr;
		NodeCommodity * prev;
		
		while(nodes!=NULL)
			{
			node  = nodes; 
			nodes = node->getNext();
			node->setNext(NULL);
			node->computeWeight(nProcs);
			
			for(prev=NULL, curr=sorted; 
			    curr && curr->getWeight() < node->getWeight();
			    prev=curr, curr=curr->getNext());

			if(prev) prev->setNext(node);
			else	 sorted = node;
			node->setNext ( curr );
			if(!curr) sortedTail = node;
			}

		nodes    = sorted;
		nodeTail = sortedTail;
		}

	// *********************************************************************
	// * Following the sort, call the regular NodeBroker to provide the
	// * next node from the list.
	// *********************************************************************
	return NodeBroker::requestNode(type);
	}
