// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The file ClusterNode.h contains the class definition for the ClusterNode
// 	class.  
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _CLUSTER_NODE_H_
#define _CLUSTER_NODE_H_ 1

#include <stdlib.h>		// Needed for size_t
#include <PBS_Node.h>		// Needed for class PBS_Node
#include <FreeProcMatrix.h>	// Needed for class FreeProcMatrix
#include <PBS_KeywordTable.h>	// Needed for class PBS_KeySet

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class CombinedTimeSlots;
class SimpleProcessor;
class RangeList;
class RangeDefinition;

// ******************************************************************************
// * class ClusterNode:
// *	The ClusterNode object is an interface that is used by the PBS_Matrix
// *	to perform match making on the cluster.  This object can provide a 
// *	linked list of free time period of a specific length which can
// *	then be used to determine the optimal time to execute a job.  The 
// *	scheduling system can then use this obejct to claim one of those
// *	free time period for the job.
// *****************************************************************************
class ClusterNode 
{
protected:
	// *********************************************************************
	// * size_t index :
	// *	This is an index number that uniquely identfies this node among
	// *	the other nodes in the system.
	// *********************************************************************
	size_t index;
	
	// *********************************************************************
	// * size_t nProcs:
	// *	This is the number of processors that this cluster node contains
	// *********************************************************************
	size_t nProcs;
	
	// *********************************************************************
	// * PBS_Node & node:
	// *	This is the PBS_Node object that this ClusterNode represents.
	// *	The PBS_Node objects are generated from data obtained 
	// *	dynamically from the PBS_Server.
	// *********************************************************************
	PBS_Node & node;
	
	// *********************************************************************
	// * PBS_KeySet keys :
	// *	This object represents all of the keywords that have been 
	// *	defined for the PBS_Node.  The PBS_KeySet object will be used
	// *	to compare the capabilities of the node with teh requirements of
	// *	the requesting job to determine if the node matches the need.
	// *********************************************************************
	PBS_KeySet keys;
	
	// *********************************************************************
	// * CombinedTimeSlots ** slots :
	// *	This is an array of CombinedTimeSlots objects. There will be 
	// *	one CombinedTimeSlots object for each possible number of
	// *	processors that can be requested simultaneously.  (If the node 
	// *	has one processor, then there will be only one of these objects.
	// *	If the node has 8 processors then there will be 8 of them.)
	// *
	// *	Note that a sufficient number of pointers is allocated when the
	// *	object is created, however, the actual objects will not be
	// *	allocated until the caller requests 'that' number of concurrent
	// *	processors from the ClusterNode.
	// *********************************************************************
	CombinedTimeSlots ** slots;
	
	// *********************************************************************
	// * SimpleProcessor ** procs:
	// *	This is a pointer to the SimpleProcessor objects representing 
	// *	the individual processors on the ClusterNode.
	// *********************************************************************
	SimpleProcessor ** procs;
	
	// *********************************************************************
	// * FreeProcMatrix pMatrix :
	// *	This object is populated with the list of all periods of 
	// *	availability on all of the processors on this node.  The 
	// *	FreeProcMatrix can then return a list of when specific 
	// *	combinations of these processors will be available concurrently.
	// *
	// *	Whenever the user makes a request to claim or incorporate
	// *	a job on the ClusterNode, this object will be cleared and
	// *	the dirty flag will be set.  It's data will be regenerated 
	// *	the next time the caller requests a range of free time.
	// *********************************************************************
	FreeProcMatrix pMatrix;
	
	// *********************************************************************
	// * int dirty :
	// *	This flag indicates that the data stored in the object has 
	// *	been made out of date by the addition of a job.
	// *********************************************************************
	int dirty;
	
	// *********************************************************************
	// * ClusterNode::refresh :
	// *	This method is called to cause the ClusterNode to regenerate
	// *	it's FreePeriodMatrix.  This matrix is then used to refresh the
	// *	list of CombinedTimeSlots objects.
	// *********************************************************************
	void refresh ( void );
	
	// *********************************************************************
	// * ClusterNode::mark :
	// *	This method is called internally to set the dirty flag to TRUE
	// *	and then to mark all of the CombinedTimeSlots object as dirty.
	// *********************************************************************
	void mark ( void );
	
public:
	// *********************************************************************
	// * ClusterNode::ClusterNode :
	// *	This is the public constructor for the object. It will 
	// *	initialize all internal variables, allocate each of the 
	// *	SimpleProcessor objects and prepare the object for immediate 
	// *	utilization.
	// *********************************************************************
	ClusterNode ( PBS_Node &, PBS_KeywordTable &, size_t );
	
	// *********************************************************************
	// * ClusterNode::~ClusterNode :
	// *	This is the destructor for the ClusterNode object.  It will 
	// *	delete any memory or sub-objects that were created during its
	// *	lifecycle.
	// *********************************************************************
	~ClusterNode ( void );
	
	// *********************************************************************
	// * These are the inline accessor methods that are defined to obtain
	// * information about the internal components of the class.
	// *********************************************************************
	inline const char       * getName    ( void ) { return node.getName(); } 
	inline size_t             getIndex   ( void ) { return index; }
	inline SimpleProcessor ** getProcs   ( void ) { return procs; }	
	inline size_t             getProcCnt ( void ) { return nProcs; }
	inline PBS_KeySet &       getKeys    ( void ) { return keys; }
	
	// *********************************************************************
	// * ClusterNode::claim :
	// *	This method is called to claim a period of time on a specified
	// *	number of processors.  The period of time MUST have been 
	// *	calculated from the data returned by the getRange method.  This
	// *	method will return an array of character strings that identify
	// *	the processors that have been allocated to the job.  It is the
	// *	responsibility of the caller to delete the array - but, not
	// *	the strings.
	// *********************************************************************
	char ** claim ( char *, size_t, time_t, time_t );
	
	// *********************************************************************
	// * ClusterNode::incorporate :
	// *	The incorporate method is called to add the statistics for
	// *	a running job to the underlying SimpleProcessor.
	// *********************************************************************
	void incorporate ( char *, char *, time_t, time_t );
	void incorporate ( size_t, char *, time_t, time_t );
	
	// *********************************************************************
	// * ClusterNode::available :
	// *	This method will check each of the underlying SimpleProcessor
	// *	objects and will return TRUE if any of the processors is not
	// *	either running a job or scheduled to immediately start a job.
	// *********************************************************************
	int available ( void );
	
	// *********************************************************************
	// * ClusterNode::availableTime :
	// *	This method will return the longest time slot that is 
	// *	immediately available on any ONE processor in the ClusterNode.
	// *********************************************************************
	time_t availableTime ( void );

	// *********************************************************************
	// * ClusterNode::getRange :
	// *	This method is called to get a list of free times when a 
	// *	specified number of processors are concurrently available. The
	// *	caller will then select the best time from this list and will
	// *	call the claim method to reserve that period of time.
	// *********************************************************************
	RangeList * getRange ( RangeDefinition & );

	// *********************************************************************
	// * ClusterNode::getUtilization :
	// *	This method will return a double value that represents the 
	// *	percentage of processors USED on this ClusterNode at a specific
	// *	time.  This value will be used in concert with Range objects
	// *	to provide better load balancing in a modestly loaded system.
	// *********************************************************************
	double getUtilization ( time_t startTime );

	// *********************************************************************
	// * ClusterNode::getUtilization :
	// *	This method will compute the percentage of this nodes processor 
	// *	resources that are USED during the specified time period.  A 
	// *	node that is inuse on all processors during the period will 
	// *	have a utilization of 1.0, while a node that has no running jobs 
	// *	during the interval will have a utilization of 0.
	// *********************************************************************
	double getUtilization ( time_t startTime, time_t duration );
	
	// *********************************************************************
	// * ClusterNode::getGapSize :
	// *	This method will retrieve the minimum amount of unused time (in 
	// *	seconds) that will be trapped if a caller claims 'cnt' 
	// *	processors at the specified 'startTime'.  The trick, of course, 
	// *	is to utilize nodes in such a way as to minimize the size of 
	// *	these gaps. Therefore, as a preliminary step it will be 
	// *	necessary to identify only those FreeTimeSlots that start at or 
	// *	before the specified startTime and have sufficient duration 
	// *	after that startTime.
	// *********************************************************************
	double getGapSize ( size_t cnt, time_t startTime, time_t duration );

	// *********************************************************************
	// * ClusterNode::reserveTime :
	// *	This method is called when the system needs to reserve a block
	// *	of time to prevent it from being utilized by a job.  This method
	// *	will walk through the list of processor objects and reserve
	// *	the specified period of time on each of them.
	// *********************************************************************
	void reserveTime ( time_t start, time_t duration );
	
	// *********************************************************************
	// * ClusterNode::asciiDump :
	// *	This method allows the caller to generate an ascii text
	// *	diagnostic dump of the contents of the ClusterNode object. 
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _CLUSTER_NODE_H_ */
