// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	CombinedTimeSlots.h contains the class header, method and attribute
// 	definitions for the CombinedTimeSlot class.  
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _COMBINED_TIME_SLOTS_H_
#define _COMBINED_TIME_SLOTS_H_ 1

#include <RangeList.h>		// Needed by inline methods for class RangeList
#include <stdlib.h>		// Needed for size_t

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class FreeTimeSet;

// *****************************************************************************
// * class CombinedTimeSlots :
// *	This class represents a collection of available time slots for a 
// *	specific number of processors on ONE node.  The class is composed of
// *	a list of distinct FreeTimeSet objects that each represent a period
// *	of availability on the specified number of processors.
// *
// *	Whenever a new job is allocated to a node, the CombinedTimeSlots object
// *	becomes dirty and then must be refreshed prior to being used to 
// *	acquire further information.
// *****************************************************************************
class CombinedTimeSlots 
{
private:
	// *********************************************************************
	// * size_t size:
	// *	This is the number of processors that must have overlapping 
	// *	free time in order to represent a period of availability. If
	// *	more than the required number of processors is available, then
	// *	only the ones that contribute to the longest duration of 
	// *	availability will be used.
	// *********************************************************************
	size_t size;
	
	// *********************************************************************
	// * int dirty:
	// *	This flag - when set to non-zero - indicates that a new job has
	// *	been allocated to one or more of the processors that are 
	// *	represented by this CombinedTimeSlots object.  Consequently, 
	// *	the object must refresh its information before any further 
	// *	availability information can be provided.
	// *********************************************************************
	int dirty;
	
	// *********************************************************************
	// * FreeTimeSet * slots:
	// *	The slots object is a pointer to a list of different periods
	// *	of availability that occur concurrently on 'size' number
	// *	of processors on the node.  The CombinedTimeSlot object
	// *	generates these objects to describe and supply the periods
	// *	of availability on the node.
	// *********************************************************************
	FreeTimeSet * slots; 

public:
	// *********************************************************************
	// * CombinedTimeSlots::CombinedTimeSlots :
	// *	This is the constructor for the object.  It is provided with 
	// *	the number of processors that are required to represent a 
	// *	quorum.
	// *********************************************************************
	CombinedTimeSlots ( size_t );
	
	// *********************************************************************
	// * CombinedTimeSlots::~CombinedTimeSlots :
	// *	This is the destructor for the CombinedTimeSlots object.  It
	// *	is responsible for deleteing all of the FreeTimeSet objects 
	// *	that it has allocated to track the periods of availability.
	// *********************************************************************
	~CombinedTimeSlots ( void );
	

	// *********************************************************************
	// * CombinedTimeSlots::getRange :
	// *	This method will populate the RangeList with all periods of 
	// *	availability that match the specified duration.  
	// *********************************************************************
	inline void getRange ( RangeList & list ) { list.addRanges(slots); }

	// *********************************************************************
	// * CombinedTimeSlots::getGroupSize :
	// *	This method retrieves the size attribute of the object.
	// *********************************************************************
	inline size_t getGroupSize ( void ) { return size; }

	// *********************************************************************
	// * CombinedTimeSlots::isDirty :
	// *	This method returns the currentstate of the dirty flag.
	// *********************************************************************
	inline int isDirty ( void ) { return dirty; }
	
	// *********************************************************************
	// * CombinedTimeSlots::mark :
	// *	This method sets the dirty flag on the CombinedTimeSlots object.
	// *	Once the object has been 'soiled', only the refresh method can
	// *	be used to restore it.
	// *********************************************************************
	inline void mark ( void ) { dirty = 1; }
	
	// *********************************************************************
	// * CombinedTimeSlots::refresh :
	// *	This method is called to force the CombinedTimeSlots object
	// *	to reload its contents from the data stored in the owning
	// *	ClusterNode object.  If the dirty flag is set, this method is 
	// *	executed prior to allocating any new periods of availability 
	// *	because the underlying FreeTimeSets are corrupted each time a 
	// *	job is allocated to the node.
	// *********************************************************************
	void refresh ( FreeTimeSet * Slots );
	

	// *********************************************************************
	// * CombinedTimeSlots::asciiDump :
	// *	This method performs an ascii diagnostic dump of the contents of
	// *	the CombinedTimeSlots object.  This includes a list of all 
	// *	underlying periods of availability and the associated 
	// *	processors.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _COMBINED_TIME_SLOTS_H_ */
