// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	FreeProcMatrix.h contains the class definitions for the methods and
// 	attributes of the classes used to perform free time sorting for 
//	time slots within a single processor.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _FREE_PROC_MATRIX_H_
#define _FREE_PROC_MATRIX_H_ 1

#include <stdlib.h>		// Needed for size_t
#include <stdio.h>		// Needed for FILE and stdout

class FreeTimeSlot;
class FreeTimeSet;
class SimpleProcessor;

// *****************************************************************************
// * class FreeProcMatrix :
// *	This class is the manager class for the FreeProcPeriod objects.  It 
// *	contains the hitching point that is used to attach the head of the 
// *	FreeProcPeriod list.  It also contains the methods used to sort the 
// *	objects and retrieve FreeTimeSlot objects that represented periods of
// *	availability across multiple processors on the same node.
// *****************************************************************************
class FreeProcMatrix 
{
protected:
	// *********************************************************************
	// * The FreeProcPeriod structure contains links to a FreeTimeSlot and
	// * an associated processor. This object also contains prev and next 
	// * links that are used to manage the arrangement of items in the
	// * list without requiring them to be physically moved.
	// *********************************************************************
	typedef struct FreeProcPeriod
	{
	public:
		struct FreeProcPeriod  * next;
		struct FreeProcPeriod  * prev;
		FreeTimeSlot           * slot;
		SimpleProcessor        * proc;
	} FreeProcPeriod;
	
	// *********************************************************************
	// * FreeProcPeriod * procs:
	// *	This is the pointer to which the array of FreeProcPeriod objects
	// *	will be allocated.
	// *********************************************************************
	FreeProcPeriod * procs;
	
	// *********************************************************************
	// * FreeProcPeriod * head:
	// *	This is the head of the list.  While the data objects are 
	// *	actually allocated to the procs array, this pointer will be used
	// *	to manage the linked-list, to prevent FreeProcPeriod objects
	// *	from being lost during usage.
	// *********************************************************************
	FreeProcPeriod * head;

	// *********************************************************************
	// * size_t procCnt :
	// *	This is the number of FreeProcPeriod objects that have actually
	// *	been stored in the procs array.
	// *********************************************************************
	size_t procCnt;
	
	// *********************************************************************
	// * size_t procMax :
	// *	This is the number of FreeProcPeriod objects that have been 
	// *	allocated to the array.
	// *********************************************************************
	size_t procMax;
	
	// *********************************************************************
	// * int resortFlag :
	// *	The resort flag indicates that a FreeProcPeriod object has been
	// *	added to the array and the array should be resorted prior to use.
	// *********************************************************************
	int resortFlag;
	
	// *********************************************************************
	// * int relinkFlag :
	// *	Indicates that an operation has occurred that could have 
	// *	impacted the linkage of the list.  This operation could be the 
	// *	addition of a new FreeProcPeriod, detachment of a FreeProcPeriod,
	// *	or a sort.
	// *********************************************************************
	int relinkFlag;

public:
	// *********************************************************************
	// * FreeProcMatrix::FreeProcMatrix :
	// *	This is the constructor for the FreeProcMatrix class.  It 
	// *	initializes internal variables and allocates a default number
	// *	of FreeProcPeriod objects to the procs array.
	// *********************************************************************
	FreeProcMatrix ( void );

	// *********************************************************************
	// * FreeProcMatrix::~FreeProcMatrix:
	// *	This is the destructor for the FreeProcMatrix object.  It will 
	// *	delete all of the FreeProcPeriod objects that were allocated
	// *	during its operation.
	// *********************************************************************
	~FreeProcMatrix ( void );
	
	// *********************************************************************
	// * FreeProcMatrix::clear :
	// *	This method sets the procCnt flag to 0 and clears the contents 
	// *	of all allocated FreeProcPeriod objects.  This method is 
	// *	performed in expectation of repopulating the object with new 
	// *	data.
	// *********************************************************************
	void clear ( void );
	
	// *********************************************************************
	// * FreeProcMatrix::detach :
	// *	This allows a FreeProcPeriod object to be detached from the 
	// *	linked list associated with the head variable, without being
	// *	removed from the allocated list.
	// *********************************************************************
	void detach ( FreeProcPeriod * period );
	
	// *********************************************************************
	// * FreeProcMatrix::append:
	// *	This method allows the caller to dynamically create and append
	// *	a new FreeProcPeriod object into the list.
	// *********************************************************************
	void append ( SimpleProcessor * Proc, FreeTimeSlot * Slot );


	// *********************************************************************
	// * FreeProcMatrix::append :
	// *	This method will append all of the FreeTimeSlot objects that 
	// *	are defined within a SimpleProcessor.
	// *********************************************************************
	void append ( SimpleProcessor * Proc );

	// *********************************************************************
	// * FreeProcMatrix::comp :
	// * 	This is the comparison function that is used by qsort. It 
	// * 	compares two FreeProcPeriod objects.
	// *
	// *	0  - The two objects are equal.
	// *	-1 - a is earlier/longer than b
	// *	1  - a is later/shorter than b
	// *********************************************************************
	static int comp ( const void * a, const void * b );
	
	// *********************************************************************
	// * FreeProcMatrix::sort : 
	// *	This is an implementation of the sort alogorithm that will be 
	// * 	used to sort the free time periods.
	// *********************************************************************
	void sort ( void );

	// *********************************************************************
	// * FreeProcMatrix::link :
	// *	This method will re-link the linkedlist of FreeProcPeriod 
	// *	objects from the sorted array.  Any FreeProcPeriod objects that
	// *	were detached or removed from the linked list will be restored
	// *	and the list will be ready to use.
	// *********************************************************************
	void link ( void );

	// *********************************************************************
	// * FreeProcMatrix::merge :
	// *	This method will call the sort routine to sort the regions of 
	// *	free time by startTime and duration.  It will then accumulate
	// *	these regions into periods of free time that span 'size' 
	// *	processors.  A pointer to a list of FreeTimeSet objects will 
	// *	be returned.
	// *
	// * Note that the FreeTimeSet objects generated by this method become
	// * the property of the caller and should be deleted when they are no 
	// * longer needed.
	// *********************************************************************
	FreeTimeSet * merge ( size_t size );

	// *********************************************************************
	// * FreeProcMatrix::asciiDump : 
	// *	This method is used to generate a diganostic ascii report of the
	// *	contents of the object.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};
#endif /* _FREE_PROC_MATRIX_H_ */
