// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	FreeTimeSet.cc realizes the methods defined within the FreeTimeSet
// 	class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <FreeTimeSet.h>	// Needed for local definitions
#include <SimpleProcessor.h>	// Needed for class SimpleProcessor

// *****************************************************************************
// * FreeTimeSet::freeList :
// *	This is the static definition for the freeList that is contained within
// *	the FreeTimeSet class.
// *****************************************************************************
FreeTimeSet * FreeTimeSet::freeList;

// *****************************************************************************
// * operator FreeTimeSet::new :
// *	This operator is called whenever a new FreeTimeSet object is 
// *	requested. The operator will first attempt to extract a FreeTimeSet
// *	object from those preallocated on the freeList.  If no FreeTimeSet
// *	objects are available from the freeList, then the freeList will be 
// *	repopulated.
// *****************************************************************************
void * FreeTimeSet::operator new ( size_t )
	{
	if(freeList==NULL)
		{
		int i;
		
		freeList = ::new FreeTimeSet[32];
		for(i=0; i<31; i++) freeList[i].next = &freeList[i+1];
		freeList[i].next = NULL;
		}

	FreeTimeSet * retval = freeList;
	freeList     = retval->next;
	retval->next = NULL;
	
	return (void *)retval;
	}
	
// *****************************************************************************
// * operator FreeTimeSet::delete :
// *	This is the delete operator for the FreeTimeSet class.  It will
// *	return the object to the freeList so that it can be provided (without
// *	reallocation) the next time the user requires one.
// *****************************************************************************
void FreeTimeSet::operator delete ( void * ptr )
	{
	FreeTimeSet * freeSet = (FreeTimeSet *)ptr;
	freeSet->next = freeList;
	freeList      = freeSet;
	}
	
	
// *****************************************************************************
// * FreeTimeSet::FreeTimeSet : 
// *	This is a private constructor for the FreeTimeSet object.  It will
// *	be called each time the FreeTimeSet::new operator has to repopulate
// *	the freeList.
// *****************************************************************************
FreeTimeSet::FreeTimeSet ( void )
	: nProcs(0), startTime(0), duration(0), next (NULL)
	{
	int idx; 
	for(idx=0; idx<SchedStatic::MAXPROC; idx++) procs[idx] = NULL;
	}
	
// *****************************************************************************
// * FreeTimeSet::FreeTimeSet :
// *	This is the public constructor for the FreeTimeSet object. It will
// *	be called when the user needs to allocate a new FreeTimeSet object.
// *****************************************************************************
FreeTimeSet::FreeTimeSet ( SimpleProcessor ** Procs, size_t NProcs, time_t StartTime, time_t Duration )
	: nProcs    ( NProcs ),
	  startTime ( StartTime ),
	  duration  ( Duration ),
	  next      ( NULL)
	{
	int idx; 
	if(nProcs>SchedStatic::MAXPROC) nProcs = SchedStatic::MAXPROC;
	for(idx=0; idx<SchedStatic::MAXPROC; idx++) procs[idx] = (idx<nProcs)?Procs[idx]:NULL;
	}
	
// *****************************************************************************
// * FreeTimeSet::~FreeTimeSet :
// *	This is the destructor for the FreeTimeSet object.  It will deallocate
// *	the memory that was allocated to the name string and will clear all 
// *	variables before the object is returned to the freeList.
// *****************************************************************************
FreeTimeSet::~FreeTimeSet ( void )
	{
	}
	
// *****************************************************************************
// * FreeTimeSet::asciiDump :
// *	This method is called to generate an ascii text report of the contents
// *	of the FreeTimeSet object.  This data will be reported to the file
// *	descriptor specified in the fp parameter.  Each line of the report will
// *	be prefixed by the string specified in the prefix parameter. The 
// *	compressed flag indicates that all data should be reported to the same
// *	output line.
// *****************************************************************************
void FreeTimeSet::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = (prefix?prefix:&c);
		
	fprintf(fp, "%sFree time on %i processors starts at %s for %i:%02i:%02i hours\n", 
		prefixString, nProcs, SchedStatic::getTimeString(startTime), 
		duration/3600, (duration%3600)/60, duration%60);
	int idx;
	for(idx=0; idx<nProcs; idx++)
		fprintf(fp, "%s - %s\n", 
			prefixString, procs[idx]?procs[idx]->getName():"Unknown");
	}
