// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	FreeTimeSet.h is the header file that defines the class methods and
//	attributes of the FreeTimeSet class.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------
#ifndef _FREE_TIME_SET_H_
#define _FREE_TIME_SET_H_ 1

#include <stdio.h>		// Needed for FILE and stdout
#include <SchedStatic.h>	// Needed for static initialization variables 

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class SimpleProcessor;

// *****************************************************************************
// * class FreeTimeSet :
// *	This class represents a collection of overlapping periods of 
// *	availability on one or more processors within a compute node.  The
// *	Scheduler will poll this list to identify the earliest period in
// *	which a job can be scheduled.
// *	
// *	In order to minimize reallocations, the class will maintain a 
// *	static freelist.  this freelist will be populated with an array
// *	of preallocated FreeTimeSet objects that will be provided to the 
// *	caller upon request.  When a user deletes a FreeTimeSet object, it
// *	will be returned to the freelist for later use.
// *****************************************************************************
class FreeTimeSet 
{
private:
	// *********************************************************************
	// * FreeTimeSet * freelist:
	// *	This is the freelist pointer that is used to maintain a pool of
	// *	available FreeTimeSet objects for allocation.
	// *********************************************************************
	static class FreeTimeSet * freeList;

	// *********************************************************************
	// * FreeTimeSet * next:
	// *	This pointer is used to maintain a linked list of FreeTimeSet
	// *	objects.  When the objects are not allocated to a user, this
	// *	next pointer is used to maintain them on the freelist.
	// *********************************************************************
	FreeTimeSet * next;
	
	// *********************************************************************
	// * time_t startTime:
	// *	This value represents the time that the period of availability 
	// *	begins on the specified processors.
	// *********************************************************************
	time_t startTime;
	
	// *********************************************************************
	// * time_t duration :
	// *	This is the length of the period of availability in seconds.
	// *********************************************************************
	time_t duration;

	// *********************************************************************
	// * size_t nProcs:
	// *	This is the number of processors that are represented by this
	// *	FreeTimeSet object.  There should be a pointer in the procs 
	// *	array for this many SimpleProcessor objects.
	// *********************************************************************
	size_t nProcs;
	
	// *********************************************************************
	// * SimpleProcessor * procs[MAXPROC]:
	// *	This is an array of MAXPROC SimpleProcessor pointers that is 
	// *	used to maintain the list of processors that are represented 
	// *	by this FreeTimeSet object.
	// *
	// *	By using the MAXPROC definition to prespecify the 
	// * 	maximum number of processors, the FreeTimeSet object can be 
	// *	used to support any number of processors supported by the 
	// *	system.  This elliminates the work required to maintain 
	// *	FreeTimeSet objects for different numbers of processors on 
	// *	nProc specific freelists.
	// *********************************************************************
	SimpleProcessor * procs[SchedStatic::MAXPROC];
	
	// *********************************************************************
	// * FreeTimeSet::FreeTimeSet :
	// *	This is the default constructor for the object.  Because it
	// *	requires no parameters it should only be called internally
	// *	to repopulate the freelist.
	// *********************************************************************
	FreeTimeSet ( void );
	
public:
	// *********************************************************************
	// * FreeTimeSet::operator new :
	// *	This is the new method for the FreeTimeSet object.  It will
	// *	provide a preallocated FreeTimeSet object from the freelist and
	// *	repopulate the freelist when necessary.
	// *********************************************************************
	void * operator new ( size_t size );
	
	// *********************************************************************
	// * FreeTimeSet::operator delete :
	// *	This is the delete method for the FreeTimeSet class.  When  
	// *	a FreeTimeSet object is deleted, this method will return it to 
	// *	the freelist where it can be reallocated to a new user at a
	// *	later time.
	// *********************************************************************
	void operator delete ( void * ptr );
	
	// *********************************************************************
	// * FreeTimeSet::FreeTimeSet :
	// *	This is the standard constructor for the object.  It is called
	// *	with the following parameters.
	// *
	// *	Procs :		This is an array of pointers to the 
	// *			SimpleProcessor objects that are to be 
	// *			represented by this FreeTimeSet.
	// *	NProcs :	This is the number of pointers provided in the
	// *			processors array.
	// *	StartTime :	This is the time that the period of availability
	// *			will begin.
	// *	Duration : 	This is the length of the period of availability
	// *********************************************************************
	FreeTimeSet ( SimpleProcessor **, size_t, time_t, time_t );
	
	// *********************************************************************
	// * FreeTimeSet::~FreeTimeSet :
	// *	This is the destructor for the FreeTimeSet object.  Because
	// *	this object is always restored to the freelist, this method
	// *	does nothing of consequence.
	// *********************************************************************
	~FreeTimeSet ( void );
	
	// *********************************************************************
	// * These are the inline accessor methods for the contents of the 
	// * FreeTimeSet object.
	// *********************************************************************
	inline void               setNext       ( FreeTimeSet * ptr ) { next = ptr; } 
	inline FreeTimeSet      * getNext       ( void ) { return next; }
	inline time_t             getStartTime  ( void ) { return startTime; }
	inline time_t             getEndTime    ( void ) { return startTime+duration; }
	inline time_t             getDuration   ( void ) { return duration; }
	inline size_t             getProcCnt    ( void ) { return nProcs; }
	inline SimpleProcessor ** getProcessors ( void ) { return procs; }	

	// *********************************************************************
	// * FreeTimeSet::asciiDump :
	// *	This method is used to generate a diagnostic ascii report of the
	// *	contents of the object. 
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _FREE_TIME_SET_H */
