// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The JobCandidate.cc file defines the non-inline methods that are
// 	identified for the JobCandidate class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <JobCandidate.h>	// Needed for local class definitions
#include <JobCandidateSubSet.h>	// Needed for JobCandidateSubSet definition
#include <SchedStatic.h>	// Needed for SchedStatic::stringToDuration
// *****************************************************************************
// * This is a utility string array that is used to collect the keyword strings
// * associated with a JobCandidateSubset prior to creating the object.
// *****************************************************************************
static const size_t  keyMatrixMaxCnt = 64;
static const size_t  keyMatrixMaxLen = 64;
static char keyMatrix[keyMatrixMaxCnt+1][keyMatrixMaxLen+1];

// *****************************************************************************
// * JobCandidate::JobCandidate :
// *	This is the constructor for the JobCandidate object.  It will 
// *	perform local data allocations and will descend into the bowels
// *	of the provided PBS_Job object to determine the specific job
// *	requirements.  A JobCandidateSubSet object will be created
// *	for each of these requirement definitions.  These subset objects
// *	will then be used to poll the ClusterNode for specific nodes 
// *	that are capable of supporting the job.
// *****************************************************************************
JobCandidate::JobCandidate ( PBS_Job & Job, PBS_KeywordTable & KeyTable )
	: next     (NULL),
	  duration (0),
	  subsets  (NULL),
	  subsetCnt(1),
	  job      (Job),
	  keyTable (KeyTable)	  
	{
	// *********************************************************************
	// * Read the duration from the job... This is actually the user 
	// * specified walltime resource.
	// *********************************************************************
	duration = job.getDuration();
	
	// *********************************************************************
	// * Get the "nodes" resource string from the job.  It will be
	// * parsed to determine the type of nodes required by the job.
	// *********************************************************************
	char * ptr   = (char *)job.getRequestedResources()->find("nodes");
	if(!ptr) ptr = (char *)job.getRequestedResources()->find("nodect");
	char * nodeString = strdup((ptr && *ptr)?ptr:"1");

	// *********************************************************************
	// * Count the number of pluses (+) in the string - this number will be
	// * 1 less than the total number of node type requested by the job.
	// *********************************************************************
	for(ptr=nodeString; ptr && *ptr; ptr++) if((*ptr)=='+') subsetCnt++;
	
	// *********************************************************************
	// * Allocate an array of JobCandidateSubSet pointers to hold the 
	// * JobCandidateSubSets.
	// *********************************************************************
	subsets = new JobCandidateSubSet * [subsetCnt];
	memset(subsets, 0, sizeof(JobCandidateSubSet *)*subsetCnt);
	
	// *********************************************************************
	// * Clear the keyMatrix before starting.
	// *********************************************************************
	memset(keyMatrix, 0, keyMatrixMaxCnt*keyMatrixMaxLen*sizeof(char));
	
	
	size_t  idx          = 0;
	size_t  commonKeyCnt = 0;
	size_t  uniqueKeyCnt = 0;

	// *********************************************************************
	// * Common keys are those keywords that are defined after the '#' 
	// * character. These keys will apply to each of the subsets that are
	// * defined by the nodes resource...  Consequently, these strings are
	// * collected first and are copied to the beginning of the keyMatrix.
	// *********************************************************************
	if((ptr = strchr(nodeString, '#'))!=NULL)	
		{
		*(ptr++)=0;
	
		while(commonKeyCnt<keyMatrixMaxCnt && ptr && *ptr)
			{
			while(isspace(*ptr)) ptr++;
			char * keyword = keyMatrix[commonKeyCnt];
			size_t charIdx = 0;
			
			while( *ptr && *ptr!=':' )
				{
				if((charIdx++)<keyMatrixMaxLen)
					*(keyword++) = *(ptr++); 
				}
			*keyword = 0;
			if(*keyMatrix[commonKeyCnt]) commonKeyCnt++;
			if(*ptr) ptr++;
			}
		}		
	
	// *********************************************************************
	// * Start collecting the keys for each set of nodes type.
	// *********************************************************************
	char * thisSet = nodeString;
	for(idx=0; idx<subsetCnt && thisSet; idx++)
		{
		size_t numNodes = 0;
		size_t numProcs = 0;
		
		// *************************************************************
		// * The plus character indicates the beginning of the next
		// * node type definition series.  Locate this character and
		// * use it's position to NULL terminate the current node type
		// * definition list.
		// *************************************************************
		char * nextSet  = strchr(thisSet, '+');
		if(nextSet) *(nextSet++)=0; 

		// *************************************************************
		// * Use the strtol function to convert the current node type
		// * definition to a long integer.  The ptr string will 
		// * receive a pointer to the first character that is not part 
		// * of the integer string - this should be a colon if keywords
		// * follow.
		// *************************************************************
		numNodes = strtol(thisSet, &ptr, 10);
		
		// *************************************************************
		// * Collect all of the keywords from the subset definition.
		// *************************************************************
		uniqueKeyCnt = commonKeyCnt;
		while(uniqueKeyCnt<keyMatrixMaxCnt && ptr && *ptr && *ptr!='+')
			{
			while(isspace(*ptr)) ptr++;
			char * keyword = keyMatrix[uniqueKeyCnt];
			size_t charIdx = 0;
			
			while( *ptr && *ptr!=':' && *ptr!='+' )
				{
				if((charIdx++)<keyMatrixMaxLen)
					*(keyword++) = *(ptr++);
				}
			*keyword = 0;

			if(*keyMatrix[uniqueKeyCnt])
				{
				// *********************************************
				// * Note: There is a special situation where
				// * thenumber of processors per node is 
				// * specified in the node specification.  In 
				// * this circumstance, we will read the number
				// * of processors required and then elliminate
				// * the entry.
				// *********************************************
				if(!strncasecmp(keyMatrix[uniqueKeyCnt], "ppn=", 4))
					{
					numProcs = atoi(keyMatrix[uniqueKeyCnt]+4);
					*keyMatrix[uniqueKeyCnt] = 0;
					}
				else uniqueKeyCnt++;
				}
			if(*ptr==':') ptr++;
			}
		// *************************************************************
		// * Now is the time to verify that the numProcs and numNodes 
		// * variables have been populated properly. Both of these
		// * values will be set to 1 by default.
		// *************************************************************
		if(!numNodes) numNodes = 1;
		if(!numProcs) numProcs = 1;
		
		// *************************************************************
		// * Create a new JobCandidateSubSet at the current index.
		// *************************************************************
		subsets[idx] = new JobCandidateSubSet(*this, duration, numNodes, numProcs, keyTable);
		
		// *************************************************************
		// * Add all of the keywords that have been collected - if any
		// * to the new JobCandidateSubSet object.
		// *************************************************************
		while(uniqueKeyCnt > 0) subsets[idx]->addKey(keyMatrix[--uniqueKeyCnt]);
			
		// *************************************************************
		// * Set the thisSet to the nextSet ptr to allow process the
		// * next node type definition (if any).
		// *************************************************************
		thisSet = nextSet; 
		}

	// *********************************************************************
	// * Set the subsetCnt variable to the highest node type definition that
	// * was processed by the loop.
	// *********************************************************************
	subsetCnt = idx;
	
	// *********************************************************************
	// * Free the memory allocated to the variable nodeString.
	// *********************************************************************
	free (nodeString);
	}

// *****************************************************************************
// * JobCandidate::~JobCandidate :
// *	This is the destructor for the JobCandidate object.  It will
// *	delete all locally allocated memeory and will remove delete
// *	each of the JobCandidateSubSet objects that werew allocated to 
// *	describe the jobs requirements.
// *****************************************************************************
JobCandidate::~JobCandidate ( void )
	{
	while(subsetCnt > 0) delete subsets[--subsetCnt];
	delete[] subsets;
	}

// *****************************************************************************
// * JobCandidate::asciiDump :
// *	This method allows the caller to generate an ascii text dump of
// *	the contents of the JobCandidate object.  This data may be used
// *	for either debugging or logging purposes.
// *****************************************************************************
void JobCandidate::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = (prefix?prefix:&c);
	char * subPrefixString = new char[strlen(prefixString)+6];
	sprintf(subPrefixString, "%s     ", prefixString);
	
	fprintf(fp, "%sJob Candidate : %s\n", prefixString, getName());
	for(int idx=0; idx<subsetCnt; idx++)
		{
		fprintf(fp, "%s  => Subset %i\n", prefixString, idx);
		subsets[idx]->asciiDump(fp, subPrefixString);
		}
	delete subPrefixString;
	}
