// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	The JobCandidate.h header file defines the contents of the JobCandidate
// 	class.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _JOB_CANDIDATE_H_
#define _JOB_CANDIDATE_H_ 1

#include <stdlib.h>	// Needed for time_t
#include <stdio.h>	// Needed for FILE structure and stdout
#include <PBS_Job.h>	// Needed for PBS_Job definition used by inline methods

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class JobCandidateSubSet;
class PBS_KeywordTable;

// *****************************************************************************
// * class JobCandidate:
// *	The JobCandidate object is used to describbe the requirements of a 
// *	job in a manner that can be used to best identify where the job
// *	should run.  The JobCandidate object contains a collection of one of
// *	more JobCandidateSubSet objects that define specific requirements of the
// *	job.  These JobCandidateSubSet objects are passed (as RangeDefinitions) 
// *	to the ClusterNode objects to locate a list of nodes that can support
// *	the job.
// *****************************************************************************
class JobCandidate 
{
protected:
	// *********************************************************************
	// * JobCandidate * next:
	// *	This is a pointer to the next JobCandidate object in the list.
	// *********************************************************************
	JobCandidate * next;

	// *********************************************************************
	// * time_t duration:
	// *	This is the projected walltime for the job as specified in the
	// *	requested resources component of the PBS_Job object.
	// *********************************************************************
	time_t duration;
	
	// *********************************************************************
	// * size_t subsetCnt :
	// *	This is the number of subset definitions that exist for this
	// *	job candidate.
	// *********************************************************************
	size_t subsetCnt;
	
	// *********************************************************************
	// * PBS_KeywordTable & keyTable:
	// *	This is the PBS_KeywordTable that will be used to extract the
	// *	codes that identify the job requirement strings for each subset.
	// *	This reference is provdied to the object through a parameter to 
	// *	the constructor.
	// *********************************************************************
	PBS_KeywordTable & keyTable;
	
	// *********************************************************************
	// * PBS_Job & job :
	// *	This is the PBS_Job definition object that was retreived from the 
	// *	PBS_Server data object.  This object will be probed for the 
	// *	definitions that will be used to create subset objects.
	// *********************************************************************
	PBS_Job & job;	
	
	// *********************************************************************
	// * JobCandidateSubSet ** subsets:
	// *	This is a pointer to an array of JobCandidateSubSet objects
	// *	that describe the individual job requirements.  Their will 
	// *	be 'subsetCnt' of these objects created during construction.
	// *********************************************************************
	JobCandidateSubSet ** subsets;

public:
	// *********************************************************************
	// * JobCandidate::JobCandidate :
	// *	This is the constructor for the JobCandidate object.  It will 
	// *	perform local data allocations and will descend into the bowels
	// *	of the provided PBS_Job object to determine the specific job
	// *	requirements.  A JobCandidateSubSet object will be created
	// *	for each of these requirement definitions.  These subset objects
	// *	will then be used to poll the ClusterNode for specific nodes 
	// *	that are capable of supporting the job.
	// *********************************************************************
	JobCandidate ( PBS_Job & Job, PBS_KeywordTable & KeyTable );
	
	// *********************************************************************
	// * JobCandidate::~JobCandidate :
	// *	This is the destructor for the JobCandidate object.  It will
	// *	delete all locally allocated memeory and will remove delete
	// *	each of the JobCandidateSubSet objects that werew allocated to 
	// *	describe the jobs requirements.
	// *********************************************************************
	~JobCandidate ( void );	 

	// *********************************************************************
	// * Inline accessor functions:
	// *	These methods are provided as an interface to the internals
	// *	of the JobCandidate object.
	// *********************************************************************
	inline JobCandidate *       getNext      ( void ) { return next; }
	inline PBS_Job &            getJob       ( void ) { return job; }
	inline const char *         getName      ( void ) { return (const char *)job.getName(); }
	inline time_t               getDuration  ( void ) { return duration; }
	inline PBS_KeywordTable &   getKeyTable  ( void ) { return keyTable; }
	inline size_t               getSubsetCnt ( void ) { return subsetCnt; }
	inline JobCandidateSubSet * getSubset    ( size_t idx ) { return (idx<subsetCnt?subsets[idx]:NULL); } 

	// *********************************************************************
	// * Inline manipulator functions:
	// *	These methods are provided as an interface to alter the contents 
	// *	of the JobCandidate object.
	// *********************************************************************
	inline void setNext ( JobCandidate * Next ) { next = Next; }

	// *********************************************************************
	// * JobCandidate::asciiDump :
	// *	This method allows the caller to generate an ascii text dump of
	// *	the contents of the JobCandidate object.  This data may be used
	// *	for either debugging or logging purposes.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};	

#endif /* _JOB_CANDIDATE_H_ */
