// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	JobDeployment.cc contains the method declarations for the JobDeployment
// 	class.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <JobDeployment.h>	// Needed for local class definitions
#include <string.h>		// Needed for strdup
#include <stdlib.h>		// Needed for malloc, realloc, free
#include <SchedStatic.h>	// Needed for timeString conversion

// *****************************************************************************
// * JobDeployment::JobDeployment :
// *	This is the standard constructor for the object.  It will 
// *	receive a startTime, job name and a count of processors. The
// *	names of the processors will be specified at a later time
// *	using the addProc method.
// *****************************************************************************
JobDeployment::JobDeployment ( time_t Start, time_t Duration, char * Job, size_t MaxProcs = 16)
	: startTime(Start),
	  duration (Duration),
	  job      (strdup(Job)),
	  maxProcs (MaxProcs>0?MaxProcs:16),
	  numProcs (0),
	  procs    (NULL),
	  next     (NULL),
	  maxNodes (MaxProcs>0?MaxProcs:16),
	  numNodes (0),
	  nodeList (NULL)
	{
	procs = (char **)malloc(maxProcs*sizeof(char *));
	nodeList = new NodeList[maxNodes];
	}

// *****************************************************************************
// * JobDeployment::~JobDeployment :
// *	This is the destructor for the object.  It will release any
// *	memory that was allocated during the objects lifetime.
// *****************************************************************************
JobDeployment::~JobDeployment ( void )
	{
	// *********************************************************************
	// * The job pointer must be release using free because it was allocated
	// * using strdup.
	// *********************************************************************
	if(job) free(job);
	
	if(nodeList) delete[] nodeList;
	
	for(int idx=0; idx<numProcs; idx++) 
		if(procs[idx]!=NULL) free(procs[idx]);

	free(procs);
	}

// *****************************************************************************
// * JobDeployment::addProc :
// *	This method is used to add a new processor name to the list
// *	of processor names in the JobDeployment object. If the number
// *	of processor names exceeds the maximum value that was specified
// *	in the constructor, then the array size will be increased to 
// *	accomodate the new amount.
// *****************************************************************************
void JobDeployment::addProc ( char * proc )
	{
	if(proc && *proc)
		{
		char   procBuf[64];
		size_t idx;

		// *************************************************************
		// * Add the processor by name to the list of processors.
		// *************************************************************
		if(numProcs+1>=maxProcs)
			{
			maxProcs*=2;
			procs = (char **)realloc(procs, maxProcs*sizeof(char *)); 
			}
		procs[numProcs++] = strdup(proc);

		// *************************************************************
		// * Copy the node name portion of the proc string into the 
		// * procBuf buffer.
		// *************************************************************
		strcpy(procBuf, proc);
		char *ptr = procBuf;
		while(*ptr && *ptr!='/') ptr++;
		*ptr = 0;
		
		// *************************************************************
		// * Locate the entry in the nodeList if it exists.
		// *************************************************************
		for(idx=0; 
		    idx<numNodes && strcmp(procBuf, nodeList[idx].node);
		    idx++);
		
		if(idx==numNodes)
			{
			if(numNodes+1 >= maxNodes)
				{
				maxNodes *= 2;
				NodeList * temp = new NodeList[maxNodes]; 
				for(size_t i=0; i<numNodes; i++)
					{
					strcpy(temp[i].node, nodeList[i].node);
					temp[i].procs = nodeList[i].procs;
					}
				delete[] nodeList;
				nodeList = temp;
				}
			strcpy(nodeList[idx].node, procBuf);
			nodeList[idx].procs = 1;
			numNodes++;
			}
		else nodeList[idx].procs++;
		}
	}

// *****************************************************************************
// * JobDeployment::getNodeStr :
// *	This method will retrieve a server compatible list of nodes
// *	where this job can be deployed. This string will be used to modify the 
// * 	neednodes resource before the job is executed.
// *
// *	The string will be of the format: hpc5a:ppn=2+hpc6a:ppn=1
// *****************************************************************************
const char * JobDeployment::getNodeStr ( void )
	{
	static char   c       = 0;
	static char * nodeStr = NULL;
	static size_t maxLen  = 0;
	
	// *********************************************************************
	// * I am providing a maximum of 75 characters per node name in the 
	// * allocation.  It is broken out as such:
	// *	Node Name = 64
	// *	":ppn="   =  5
	// *	"xxxx"    =  5 (assuming no more than 99999 procs/nodes)
	// *	"+"       =  1
	// *********************************************************************
	if(nodeStr) *nodeStr=0;
	
	if(numNodes)
		{
		if(maxLen<75*numNodes)
			{
			if(nodeStr) delete[] nodeStr;
			maxLen   = 75*numNodes;
			nodeStr  = new char[maxLen];
			*nodeStr = 0;
			}
		for(size_t idx=0; idx<numNodes; idx++)
			{
			char * ptr = nodeStr+strlen(nodeStr);
			sprintf(ptr, "%s:ppn=%i%c", 
				nodeList[idx].node, 
				nodeList[idx].procs,
				((idx+1)<numNodes)?'+':'\0');
			}
		}
	// *********************************************************************
	// * Return the node string or an empty string if no nodes were defined.
	// *********************************************************************
	return (const char *)(numNodes?nodeStr:&c);
	}
	

// *****************************************************************************
// * JobDeployment::comp :
// *	This is the comparison function that will be used by the sort method to
// *	sort the array of JobDeployment pointers.
// *****************************************************************************
int JobDeployment::comp ( const void * s1, const void * s2 )
	{
	int result = 0;
	
	if(s1 && s2)
		{
		JobDeployment * a = *(JobDeployment **)s1;
		JobDeployment * b = *(JobDeployment **)s2;
		
		result = a->getStartTime()-b->getStartTime();
		}
	else if (s1) result = -1;
	else if (s2) result =  1;
	
	return (result<0)?-1:(result?1:0);
	}
	

// *****************************************************************************
// * JobDeployment::sort :
// *	This is a static method that will sort an array of JobDeployment
// *	object based on start time and duration. This object is given
// *	a reference to a pointer to allow it to alter the head of the
// *	list of necessary.
// *****************************************************************************
void JobDeployment::sort ( JobDeployment * &head )
	{
	size_t idx=0, cnt = 0;
	JobDeployment * ptr = head;

	// *********************************************************************
	// * If fewer than one JobDeployment objects were provided to the array
	// * then simply return without doing anything.
	// *********************************************************************
	if(!head || !head->getNext()) return;
	
	// *********************************************************************
	// * Count the number of JobDeployment objects in the list.
	// *********************************************************************
	while(ptr!=NULL)
		{
		cnt++;
		ptr = ptr->getNext();
		}
	
	// *********************************************************************
	// * Generate an array of pointers and install the jobs in the array.
	// *********************************************************************
	JobDeployment ** jobArray = new JobDeployment * [cnt];
	
	for(idx=0, ptr=head; ptr!=NULL && idx<cnt; ptr=ptr->getNext(), idx++)
		{
		jobArray[idx] = ptr;
		}
	
	// *********************************************************************
	// * Sort the jobs using the qsort algorithm.
	// *********************************************************************
	qsort(jobArray, cnt, sizeof(JobDeployment *), JobDeployment::comp);
	
	// *********************************************************************
	// * Restore each of the jobArray objects to the linked list in the
	// * new order.
	// *********************************************************************
	head = jobArray[0];
	for(idx=0; idx<(cnt-1); idx++) jobArray[idx]->setNext(jobArray[idx+1]);
	jobArray[idx]->setNext(NULL);
		
	// *********************************************************************
	// * Delete the array of JobDeployment pointers that was allocated.
	// *********************************************************************
	delete [] jobArray;
	}
	
// *****************************************************************************
// * JobDeployment::asciiDump :
// *	This method generates an ascii text report of the contents of
// *	the JobDeployment object.
// *****************************************************************************
void JobDeployment::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = prefix?prefix:&c;
	
	fprintf(fp, "%sJob %-16.16s scheduled from %s to ", 
		prefixString, 
		job, 
		SchedStatic::getTimeString(startTime));

	fprintf(fp, "%s on processor",
		SchedStatic::getTimeString(startTime+duration));
		
	for(int idx=0; idx<numProcs; idx++) 
		if(procs[idx]) fprintf(fp, "%c%s", ((idx>0)?',':' '), procs[idx]);

	fprintf(fp, "\n");
	}
