// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	JobDeployment.h contains the class definition for the JobDeployment
// 	class.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _JOB_DEPLOYMENT_H_
#define _JOB_DEPLOYMENT_H_ 1

#include <time.h>		// Needed for time_t
#include <stdio.h>		// Needed for FILE, stdout

// *****************************************************************************
// * class JobDeployment :
// *	This class represents a job that has been started or is scheduled to
// *	start on the cluster.  The class contains a start time, job name,
// *	number of processors, and a list of processors that the job will/is
// *	using.
// *
// *	Note that the list of processors may span multiple nodes depending on
// *	the job requirements.
// *****************************************************************************
class JobDeployment 
{
protected:
	// *********************************************************************
	// * JobDeployment * next :
	// *	This is a pointer to the next JobDeployment object in the linked
	// * 	list.
	// *********************************************************************
	JobDeployment * next;
	
	// *********************************************************************
	// * time_t startTime :
	// *	This is the time that the job is scheduled to start.
	// *********************************************************************
	time_t startTime;
	
	// *********************************************************************
	// * time_t duration :
	// *	This is the scheduled duration of the job.
	// *********************************************************************
	time_t duration;
	
	// *********************************************************************
	// * char * job :
	// *	This is the name of the job.
	// *********************************************************************
	char * job;
	
	// *********************************************************************
	// * size_t maxProcs :
	// *	This is the number of string pointers that have been allocated
	// *	to the procs array.  The caller can specify the maximum number
	// *	of processors in the constructor - the size of this variable
	// *	will be increased as necessary when new processor names are
	// *	added to the object.
	// *********************************************************************
	size_t maxProcs;
		
	// *********************************************************************
	// * size_t numProcs :
	// *	This is the number of strings that have been assigned to the
	// *	procs array. 
	// *********************************************************************
	size_t numProcs;
	
	// *********************************************************************
	// * char ** procs :
	// *	This is the list of processor names that the job will be 
	// *	executing on.
	// *********************************************************************
	char ** procs;
	
	// *********************************************************************
	// * class NodeList * nodeList :
	// *	This is a list of the specific nodes and the number of 
	// *	processors specified for each. This structure will be used 
	// *	to generate the nodeStr when it is time to deploy the job. 
	// *********************************************************************
	class NodeList
	{
	public:	
		char node[64];
		size_t procs;
		
		NodeList ( void ) { *node=0; procs=0; }	
	};
	
	NodeList * nodeList;
	
	// *********************************************************************
	// * size_t nodeCnt :
	// *	This is the number of nodes that have been requested for the
	// *	job.
	// *********************************************************************
	size_t numNodes;
	
	// *********************************************************************
	// * size_t maxNodes :
	// *	This is the number of NodeList objects that have been allocated.
	// *********************************************************************
	size_t maxNodes;
	
	// *********************************************************************
	// * JobDeployment::comp :
	// *	This is the comparison function that will be used by the sort 
	// * 	method to sort the array of JobDeployment pointers.
	// *********************************************************************
	static int comp ( const void * s1, const void * s2 );
	
public:
	// *********************************************************************
	// * JobDeployment::JobDeployment :
	// *	This is the standard constructor for the object.  It will 
	// *	receive a startTime, job name and a count of processors. The
	// *	names of the processors will be specified at a later time
	// *	using the addProc method.
	// *********************************************************************
	JobDeployment ( time_t Start, time_t Duration, char * Job, size_t MaxProcs = 16);
	
	// *********************************************************************
	// * JobDeployment::~JobDeployment :
	// *	This is the destructor for the object.  It will release any
	// *	memory that was allocated during the objects lifetime.
	// *********************************************************************
	~JobDeployment ( void );
	
	// *********************************************************************
	// * These are the inline accessor methods used by the object
	// *********************************************************************
	inline void            setNext      ( JobDeployment * Next ) { next = Next; }
	inline JobDeployment * getNext      ( void ) { return next; }
	inline time_t          getStartTime ( void ) { return startTime; }
	inline time_t          getEndTime   ( void ) { return startTime+duration; }
	inline time_t          getDuration  ( void ) { return duration; }
	inline char *          getJob       ( void ) { return job; }
	inline size_t          getProcCnt   ( void ) { return numProcs; }
	inline const char **   getProcs     ( void ) { return (const char **)procs; }
	
	// *********************************************************************
	// * JobDeployment::addProc :
	// *	This method is used to add a new processor name to the list
	// *	of processor names in the JobDeployment object. If the number
	// *	of processor names exceeds the maximum value that was specified
	// *	in the constructor, then the array size wil be increased to 
	// *	accomodate the new amount.
	// *********************************************************************
	void addProc ( char * );
	
	// *********************************************************************
	// * JobDeployment::getNodeStr :
	// *	This method will retrieve a server compatible list of nodes
	// *	where this job can be deployed. This string remains the property
	// *	of the JobDeployment CLASS and should not be deleted, modified,
	// *	or accessed between calls to the getNodeStr function.
	// *********************************************************************
	const char * getNodeStr ( void );
	
	// *********************************************************************
	// * JobDeployment::sort :
	// *	This is a static method that will sort an array of JobDeployment
	// *	object based on start time and duration. This object is given
	// *	a reference to a pointer to allow it to alter the head of the
	// *	list of necessary.
	// *********************************************************************
	static void sort ( JobDeployment * &head );
	
	// *********************************************************************
	// * JobDeployment::asciiDump :
	// *	This method generates an ascii text report of the contents of
	// *	the JobDeployment object.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _JOB_DEPLOYMENT_H_ */
