// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	MatchMatrix.cc realizaties the method and class defined in the 
// 	MatchMatrix.h header file.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <MatchMatrix.h>	// Required for local class and data definitions
#include <ClusterNode.h>	// Needed for ClusterNode class definition
#include <JobDeployment.h>	// Needed for JobDeployment class definition
#include <JobCandidate.h>	// Needed for JobCandidate class definition
#include <PBS_Server.h>		// Needed for PBS_Server class defintions
#include <SchedStatic.h>	// Needed for SchedStatic::stringToDuration
#include <PBS_Log.h>

// *****************************************************************************
// * MatchMatrix::MatchMatrix :
// *	This is the constructor for the MatchMatrix class.  It performs 
// *	very little actual initialization because, unlike the MatchMatrix object,
// *	most of the data it contains is not static between scheduling cycles.
// *	Consequently, the startCycle method is executed at the beginning of
// *	each scheduling cycle to update the object with fresh data.  At the end
// *	of the scheduling cycle the endCycle method is executed to release any
// *	data that is transient in nature.
// *****************************************************************************
MatchMatrix::MatchMatrix ( void ) 
	: scheduled  ( NULL ),
	  nodes      ( NULL ),
	  nNodes     ( 0 ),
	  jobs       ( NULL ),
	  range      ( ),
	  keyTable   ( ),
	  nodeMgr    ( ),
	  scheduling ( 0 )
	{
	}
	
// *****************************************************************************
// * MatchMatrix::~MatchMatrix :
// *	This is the destructor for the MatchMatrix object.  It is responsible
// *	for freeing any memeory that was allocated to the object during its
// *	lifetime.  Since the object is cyclic, all of the data deletion is 
// *	performed by the endCycle method.
// *****************************************************************************
MatchMatrix::~MatchMatrix ( void )
	{
	endCycle();
	}
	
// *****************************************************************************
// * MatchMatrix::startCycle :
// *	This method is called at the beginning of each scheduling cycle to 
// *	initialize this object with fresh data from the PBS_Server object.
// *	This data will be used during the scheduling cycle to select the best
// *	place and time for scheduling queued jobs.
// *	
// *	Upon completion, this method will return an integer indicating the 
// *	number of processors that are available for immediate jobs deployment.
// *	If there are no processors available, then it is pointless to process
// *	new jobs because there is no way that a job can be deployed.
// *****************************************************************************
int MatchMatrix::startCycle ( PBS_Server & server )
	{
	// *********************************************************************
	// * static size_t NodeOffset :
	// *	This value is provided to allow the MatchMatrix to walk through
	// *	the available nodes in a round-robin style.  This way, if the
	// *	system is lightly burdened, the MatchMatrix will tend to 
	// *	distribute the jobs across all nodes rather than merely focusing
	// *	on the first node in the list.
	// *********************************************************************
	static size_t nodeOffset = 0;
	
	// *********************************************************************
	// * The available variable is used to accumulate a count of the number
	// * of processors that are immediately available.
	// *********************************************************************
	size_t available = 0;
	
	// *********************************************************************
	// * If the system is already scheduling - indicating that the endCycle
	// * method was not called after the last scheduling cycle.  Call the 
	// * endCycle method to free any transient data objects.
	// *********************************************************************
	if(scheduling) endCycle();
	
	// *********************************************************************
	// * Read the list of available node from the PBS_Server object.
	// *********************************************************************
	PBS_NodeList & nodeList = (PBS_NodeList &)server;
	size_t         nodeIdx  = 0;
	size_t         maxIdx   = nodeList.getNodeCount();
	PBS_Node    *  node     = NULL;
	
	// *********************************************************************
	// * If the nodeOffset is greater than the number of nodes available, 
	// * then reset it to 0.
	// *
	// * Because the loadBalancing mechanism is switchable, I will only
	// * increment the startNode when loadBalancing is turned on.
	// *********************************************************************
	if(nodeOffset>=maxIdx || !SchedStatic::getBalanceFlag()) nodeOffset = 0;
	
	// *********************************************************************
	// * Set the number of nodes to 0 to indicate that none have been 
	// * allocated.
	// *********************************************************************
	nNodes = 0; 
	
	// *********************************************************************
	// * Allocate an array of ClusterNode pointers that is sufficient to 
	// * hold all entries - regardless of their state.
	// *********************************************************************
	nodes = new ClusterNode * [ maxIdx ];
	
	// *********************************************************************
	// * Note: I am creating a local macro definition here that I will use
	// * to index the NodeList - while vile, this will make the while
	// * loop easier to read and more understandable.
	// *********************************************************************
	#define NODEINDEX ((nodeOffset+nodeIdx<maxIdx)?(nodeOffset+nodeIdx):(nodeOffset+nodeIdx-maxIdx))

	// *********************************************************************
	// * Create a ClusterNode object for each node that is not down, 
	// * offline, reserved, or in an unknown state. Chain the new 
	// * ClusterNode objects into a list.
	// *********************************************************************
	while((node=nodeList[NODEINDEX])!=NULL && nodeIdx<maxIdx)
		{
		PBS_Node::NodeState state = node->getState();
		
		if(state!=PBS_Node::UNKNOWN_STATE &&
		   state!=PBS_Node::RESERVED      &&
		   state!=PBS_Node::DOWN          &&
		   state!=PBS_Node::OFFLINE)
			{
			ClusterNode * cNode = new ClusterNode(*node, keyTable, nNodes);
			
			// *****************************************************
			// * Integrate all running jobs into the ClusterNode.
			// * This is necesary so that the MatchMatrix does not
			// * attempt to overlay the execution of queued jobs on
			// * nodes that are already executing jobs.
			// *****************************************************
			char ** jobList = (char **)node->getJobs();
			if(jobList!=NULL) 
				{
				char * jobName;
				
				for(size_t jIdx=0; 
				   (jobName=jobList[jIdx])!=NULL && 
				   (jobName=strrchr(jobName, '/'))!=NULL &&
				   *(++jobName)!=0; 
				   jIdx++)
					{
					PBS_Job * job = server.getJob(jobName);
					
					if(job)	{
						size_t procIdx      = atoi(jobList[jIdx]);
						time_t duration     = job->getDuration();
						time_t startTime    = job->getTimeStarted();

						if(startTime)
							{
							// *********************
							// * In the event that a 
							// * running job has 
							// * exceeded it's
							// * specified duration,
							// * set the duration to
							// * 1 minute into the 
							// * future.
							// **********************
							if(startTime+duration <= SchedStatic::getNow())
								duration = (SchedStatic::getNow()-startTime)+60;
								
							cNode->incorporate(procIdx, 
									   (char *)job->getName(), 
									   startTime, 
									   duration);
							} 
						}
					}
				}
				
			// *****************************************************
			// * Assign the ClusterNode object into the array and
			// * count the number of available processors.
			// *****************************************************
			nodes[nNodes++] = cNode;
			available += cNode->available();
			}
		nodeIdx++;
		}

	// *********************************************************************
	// * Undefine the NODEINDEX macro and increment the static nodeOffset.
	// *********************************************************************
	#undef NODEINDEX
	nodeOffset++;

	// *********************************************************************
	// * Create a JobDeployment object for each job that is actively 
	// * executing in the system.
	// *********************************************************************
	PBS_JobList runningJobs ( server.getRunningJobs() );
	JobDeployment * tail = NULL;
	
	for(size_t jIdx=0; runningJobs[jIdx]!=NULL; jIdx++)
		{
		char ** execHost = (char **)runningJobs[jIdx]->getExecutionHosts();
		if(execHost)
			{
			JobDeployment * jd = 
				new JobDeployment(
					runningJobs[jIdx]->getTimeStarted(),
					runningJobs[jIdx]->getDuration(),
					(char *)runningJobs[jIdx]->getName());
			while(*execHost!=NULL) jd->addProc(*(execHost++));
			
			if(tail) tail->setNext(jd);
			else     scheduled = jd;
			tail = jd;
			}
		}

	// *********************************************************************
	// * Pass the list of available ClusterNodes to the startCycle method
	// * of the NodeSetManager.  The nodes associated with a specific
	// * queue may then be extracted as an array for later job scheduling
	// * and execution.
	// *********************************************************************
	nodeMgr.startCycle(nodes, nNodes);

	// *********************************************************************
	// * Set the scheduling variable to TRUE to indicate that the startCycle
	// * method has been executed and the system has been initialized with
	// * free data.
	// *********************************************************************
	scheduling = 1;
	return available;
	}
	
// *****************************************************************************
// * MatchMatrix::endCycle :
// *	This method is called at the end of the scheduling cycle to release any
// *	memory that may have been allocated during processing.
// *****************************************************************************
void MatchMatrix::endCycle ( void )
	{
	// *********************************************************************
	// * Clear the RangeTable object.
	// *********************************************************************
	range.clear();

	// *********************************************************************
	// * Delete all JobCandidate objects from the list.
	// *********************************************************************
	while (jobs!=NULL)
		{
		JobCandidate *cJob = jobs;
		jobs = cJob->getNext();
		delete cJob;
		}

	// *********************************************************************
	// * Delete all JobDeployment objects from the list.
	// *********************************************************************
	while (scheduled!=NULL)
		{
		JobDeployment * cSched = scheduled;
		scheduled = cSched->getNext();
		delete cSched;
		}

	// *********************************************************************
	// * Call the endCycle method of the NodeSetManager to release all 
	// * references to the externally allocated ClusterNode objects.
	// *********************************************************************
	nodeMgr.endCycle();

	// *********************************************************************
	// * Delete all ClusterNode objects from the list.
	// *********************************************************************	
	if (nodes!=NULL)
		{
		while(nNodes>0) delete nodes[--nNodes];
		delete[] nodes;
		nodes = NULL;
		}
	
	// *********************************************************************
	// * Set the scheduling variable to FALSE (0).
	// *********************************************************************
	scheduling = 0;
	}
	
// *****************************************************************************
// * MatchMatrix::processJobs :
// *	This method is provided with a list of PBS_Job objects that have been
// *	sorted based on priority. The MatchMatrix object must find the best
// *	location and time for each of these jobs and then generate a 
// *	JobDeployment object that describes that data.  At the end of the
// *	method the JobDeployment object will be retuned to the caller and may
// *	be used to deploy queued jobs.
// *
// *	The JobDeployment object is a transient list and is only valid during
// *	the scheduling cycle.  If the caller wishes to retain/use this 
// *	information between scheduling cycles then the extractJobs method should
// *	be used to extract the JobDeployment list from the MatchMatrix.
// *
// *	A new flag has been added to this function: quick
// *	If the quick flag IS NOT set (zero) then the method
// *	will perform a complete inspection and perform match making
// *	on all jobs... If quick IS set (non-zero), then the
// *	processJobs method will only run until available nodes are
// *	depleted.
// *****************************************************************************
JobDeployment * MatchMatrix::processJobs ( PBS_Job ** Jobs, size_t jobCnt, int quick )
	{
	// *********************************************************************
	// * The scheduling cycle must have been initiated with the startCycle
	// * method in order for this method to be called.
	// *********************************************************************
	if(scheduling && Jobs && jobCnt)
		{
		// *************************************************************
		// * time_t shortest:
		// *	This value will be used to hold the duration of the 
		// *	shortest job that has been submitted for processing.
		// *	If the processJob's method is running in
		// *	quick mode, then once the time available
		// *	from NOW on any node is less than shortest, then 
		// *	job processing can stop.
		// *
		// *	This value will be initialized with the duration of the
		// *	first node.
		// *************************************************************
		time_t shortest = Jobs[0]->getDuration();
		
		
		// *************************************************************
		// * time_t longest :
		// *	This value represents the duration of the longest
		// *	immediately available processor.  When operating in
		// *	quick mode, this value will be recomputed at 
		// *	the end of processing each job (if the job was 
		// *	scheduled) for immediate processing.
		// *************************************************************
		time_t longest = 0;
		
		// *************************************************************
		// * The jobTail pointer is the last JobCandidate in the jobs 
		// * linked list. Strat with the jobs pointer and while the 
		// * next pointer is not NULL, walk to the last entry.
		// *************************************************************
		JobCandidate * jobTail = jobs;
	
		while(jobTail && jobTail->getNext()!=NULL)
			{
			jobTail = jobTail->getNext();
			}

		// *************************************************************
		// * Create a JobCandidate object for each PBS_Job that is 
		// * provided in the array and add them to the linked list of
		// * jobs.  We are using the jobTail object to append each job 
		// * to the end of the list in order to retain the ordering that
		// * was created by the caller.
		// *************************************************************
		for(int idx=0; idx<jobCnt; idx++)
			{
			if(Jobs[idx])
				{
				// *********************************************
				// * Determine if this job is shorter than all 
				// * of the other jobs before it.
				// *********************************************
				if(Jobs[idx]->getDuration()<shortest)
					{
					shortest = Jobs[idx]->getDuration();
					}
					
				// *********************************************
				// * Create a new JobCandidate object and add
				// * it to the list.
				// *********************************************
				JobCandidate * newJob = 
					new JobCandidate(*Jobs[idx], keyTable);
				
				if(jobs) jobTail->setNext(newJob);
				else jobs = newJob;
				
				jobTail = newJob;
				}
			}
		
		// *************************************************************
		// * Walk to the end of the JobDeployment object in order to 
		// * determine its tail... All new JobDeployment objects will
		// * be added to the end of the list in order to maintain the
		// * used specified order.
		// *************************************************************
		JobDeployment * schedTail = scheduled;
		
		while(schedTail && schedTail->getNext())
			schedTail = schedTail->getNext();
			
		// *************************************************************
		// * Now that the JobCandidate list has been created, it is
		// * time to begin walking through the candidates and finding
		// * out where and when they can be executed - optimally.
		// *************************************************************
		JobCandidate  * job = jobs;
		JobDeployment * deployment;

		// *************************************************************
		// * If we are running in quick mode, then this
		// * flag will signal when one or more of the following 
		// * conditions have been met:
		// *
		// * 1) There are no more immediately available time slots on
		// *    any of the cluster nodes.
		// *
		// * 2) The job duration specified by shortest will not fit
		// *    in the longest immediately available time slot.
		// *
		// * If we ARE running in non-quick mode, then this flag
		// * will not be used and the method will continue until all
		// * jobs are given a reservation.
		// *************************************************************
		int done = 0;
		
		if(quick && (longest=availableTime())<=shortest) done=1;
		
		while(!done && job!=NULL)
			{
			// *****************************************************
			// * Clear the RangeTable object and then initialize
			// * it with the current JobCandidate object.
			// *****************************************************
			range.clear();
			range.init(job);

			// *****************************************************
			// * Ask the NodeSetManager for the list of ClusterNode
			// * objects that are associated with the queue that 
			// * this job originated from. Only these nodes will be
			// * used to attempt job scheduling.
			// *****************************************************
			size_t         jNodeCnt = 0;
			ClusterNode ** jNodes   = nodeMgr.getNodes(job->getJob().getQueueName(), jNodeCnt);
			
			// *****************************************************
			// * If the current job can be deployed immediately on
			// * the existing nodes, then add it to the end of the
			// * scheduled list.
			// *****************************************************
			if(jNodes && jNodeCnt && (deployment = range.process(jNodes, jNodeCnt))!=NULL)
				{				
				// *********************************************
				// * If we are running in quick mode, we should 
				// * determine if the deployment of this job
				// * has resulted in a decrease in the length
				// * of time on the longest immediately 
				// * available time slot.  We should update
				// * the longest variable and the done flag
				// * in accordance with these results.
				// *********************************************
				if(quick && (deployment->getStartTime()-SchedStatic::getNow())<longest)
					{
					if((longest=availableTime())<=shortest) 
					  	{
						done=1;
						}
					}	
						
				// *********************************************
				// * If we are running in quick mode any 
				// * JobDeployment that does not start 
				// * immediately will be deleted.
				// *********************************************
				if(quick && 
				   deployment->getStartTime()>SchedStatic::getNow())
					{
					delete deployment;
					}
				// *********************************************
				// * If we ARE running in non-quick mode or
				// * if the jobs starts immediately, we will
				// * add the JobDeployment to the list.
				// *********************************************
				else 	{
					if(scheduled) schedTail->setNext(deployment);
					else scheduled = deployment;
					schedTail = deployment;
					}
				}

			// *****************************************************
			// * Get a pointer to the next job in the linked list.
			// *****************************************************
			job = job->getNext();
			}

		// *************************************************************
		// * If we are running in non-quick mode, it means that the
		// * JobDeployment list will be populated with jobs that start
		// * at varying times... Call the sort method to sort the 
		// * jobs by startTime.
		// *************************************************************
		if(!quick) JobDeployment::sort(scheduled);
		}

	// *********************************************************************
	// * Return the list of JobDeployment objects to the caller.
	// *********************************************************************
	return scheduled;
	}
	
// *****************************************************************************
// * MatchMatrix::extractJobs :
// *	This method allows the caller to extract the list of JobDeployments 
// *	from the MatchMatrix.  After this method is called the scheduled list
// *	will be set to NULL and the list of JobDeployment objects will become
// *	the property of the caller.  The caller should delete the list of
// *	objects when it is no longer needed.
// *****************************************************************************
JobDeployment * MatchMatrix::extractJobs ( void )
	{
	JobDeployment * Scheduled = scheduled;
	scheduled = NULL;
	return Scheduled;
	}
	
// *****************************************************************************
// * MatchMatrix::available :
// *	This method will return TRUE if any single processor in the cluster
// *	is available for immediate deployment.
// *****************************************************************************
int MatchMatrix::available ( void )
	{
	size_t result = 0;
	if(scheduling)
		{
		for(int idx=0; !result && idx<nNodes; idx++)
			{
			result=nodes[idx]->available();
			}
		}
	return result;
	}
	
// *****************************************************************************
// * MatchMatrix::reserveTime :
// *	This method is called when the system needs to reserve a block
// *	of time to prevent it from being utilized by a job.  This method
// *	will walk through the list of processor objects and reserve
// *	the specified period of time on each of them.
// * 
// *	If no names are specified then time will be reserved
// * 	on all of the nodes.
// *****************************************************************************
void MatchMatrix::reserveTime ( time_t start, time_t duration, char ** nodeNames, size_t nodeCnt )
	{
	if(scheduling)
		{
		size_t nameIdx;
		size_t nodeIdx;

		// *************************************************************
		// * If no names are specified then time will be reserved
		// * on all of the nodes.
		// *************************************************************
		if(nodeNames==NULL)
			{
			for(nodeIdx=0; nodeIdx<nNodes; nodeIdx++)
				{
				nodes[nodeIdx]->reserveTime(start, duration);
				}
			}
		// *************************************************************
		// * Otherwise, make a reservation on the nodes that are 
		// * specified.
		// *************************************************************
		else for(nameIdx=0; nameIdx<nodeCnt && nodeNames[nameIdx]; nameIdx++)
			{
			for(int done=0, nodeIdx=0; nodeIdx<nNodes && !done; nodeIdx++)
				{
				if(!strcmp(nodeNames[nameIdx], nodes[nodeIdx]->getName()))
					{
					nodes[nodeIdx]->reserveTime(start, duration);
					done = 1;
					}
				}
			}
		}
	}

// *****************************************************************************
// * MatchMatrix::availableTime :
// *	This method will return the longest time slot that is 
// *	immediately available on any ONE processor in the cluster.
// *****************************************************************************
time_t MatchMatrix::availableTime ( void )
	{
	time_t result = 0;

	if(scheduling)
		{
		for(int idx=0; idx<nNodes; idx++)
			{
			time_t nodeTime = nodes[idx]->availableTime();
			if(nodeTime>result) result = nodeTime;
			}
		}

	return result;
	}
	
	
// *****************************************************************************
// * MatchMatrix::asciiDump :
// *	This method is called to generate a diagnostic ascii text dump of the 
// *	contents of the MatchMatrix object.  This method may be used for 
// *	debugging or for reporting purposes.  The FILE * paramater is the file 
// *	pointer where the data will be written.  The prefix parameter indicates
// *	a prefix strings which should be prepended to each line before it is
// *	printed.
// *****************************************************************************
void MatchMatrix::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = (prefix?prefix:&c);
	char * subPrefixString = new char [strlen(prefixString)+6];
	sprintf(subPrefixString, "%s     ", prefixString);
	
	fprintf(fp, "%sMatchMatrix Object :\n", prefixString);
	fprintf(fp, "%s  => Scheduled Job Deployments:\n", prefixString);
	for(JobDeployment * cSched = scheduled; cSched!=NULL; cSched=cSched->getNext())
		{
		cSched->asciiDump(fp, subPrefixString);
		}
	fprintf(fp, "%s  => Available Cluster Nodes :\n", prefixString);
	for(int nIdx=0; nIdx<nNodes; nIdx++)
		{
		nodes[nIdx]->asciiDump(fp, subPrefixString);
		}
	fprintf(fp, "%s  => Job Candidates :\n", prefixString);
	for(JobCandidate * cJob = jobs; cJob!=NULL; cJob = cJob->getNext())
		{
		cJob->asciiDump ( fp, subPrefixString );
		}
	delete[] subPrefixString;
	}
