// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
//	The MatchMatrix.h header file includes the class definition for the 
// 	MatchMatrix class.
// 
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#ifndef _MATCH_MATRIX_H_
#define _MATCH_MATRIX_H_ 1

#include <PBS_KeywordTable.h>	// Needed for class PBS_KeywordTable
#include <RangeTable.h>		// Needed for class RangeTable
#include <NodeSet.h>		// Needed for class NodeSetManager

// *****************************************************************************
// * Class declarations for classes that will be used or referenced locally. By
// * minimizing the inclusion of header files, many external dependencies can
// * be elliminated.
// *****************************************************************************
class JobDeployment;
class JobCandidate;
class ClusterNode;
class PBS_Server;
class PBS_Node;
class PBS_Job;

// *****************************************************************************
// * class MatchMatrix :
// *	The MatchMatrix class is the access point for the match making component
// *	of the underlord scheduler.  This class is provided a list fo all of the
// *	nodes, running jobs and queue jobs from the server.  It will then use
// *	this information to schedule the best possible utilization of the
// *	system (theoretically - actually it's the best possible utilization 
// *	algorithm that we could come up with).
// *****************************************************************************
class MatchMatrix
{
protected:
	// *********************************************************************
	// * JobDeployment * scheduled:
	// *	This is the list of scheduled or running jobs that have been
	// *	processed by the MatchMatrix. At the beginning of the scheduling
	// *	cycle, all running jobs are converted to JobDeployment objects.
	// *	As queued jobs are scheduled for execution, a JobDeployment 
	// *	object is generated for each of them.  
	// *********************************************************************
	JobDeployment * scheduled;
	
	// *********************************************************************
	// * NodeSetManager nodeMgr:
	// *	The NodeSetManager class is used to associated the available
	// *	nodes with a specific queue.  This code was added to the 
	// *	project to allow the system administrator to associate specific
	// *	nodes with a queue.  In the event that a queue is not referenced
	// *	in the scheduler configuration file, then the queue will have
	// *	access to all available processors.
	// *********************************************************************
	NodeSetManager nodeMgr;
	
	// *********************************************************************
	// * ClusterNode * nodes:
	// *	This is the list of all nodes that reside in the system. When
	// *	the scheduler cycle is started by the startCycle method. All
	// *	of the nodes provided in the PBS_Server object will be 
	// *	interpretted into ClusterNode objects.
	// *********************************************************************
	ClusterNode ** nodes;

	// *********************************************************************
	// * size_t nNodes;
	// *	This is the number of ClusterNodes that were allocated at the
	// *	beginning of the last cycle.
	// *********************************************************************
	size_t nNodes;
		
	// *********************************************************************
	// * JobCandidate * jobs:
	// *	This is a pointer to a linked list of JobCandidate objects.  
	// *	These objects are an interpretation of the PBS_Job object and
	// *	provide enhanced data and structure to improve scheduling.
	// *********************************************************************
	JobCandidate * jobs;

	// *********************************************************************
	// * RangeTable range:
	// *	This is a utility object used to compute the histogram for
	// *	scheduling a job that requires multiple nodes with varying
	// *	requirements.  This object provides the interface that selects
	// *	the earliest reasonable deployment time for the job.
	// *********************************************************************
	RangeTable range;
	
	// *********************************************************************
	// * PBS_KeywordTable keyTable :
	// *	This is the keyword table that will be used to convert key 
	// *	strings to integers.  The key integers are used to associate
	// *	a job with a corresponding node.
	// *********************************************************************
	PBS_KeywordTable keyTable;
	
	// *********************************************************************
	// * int scheduling :
	// *	This variable indicates whether the startCycle method has been
	// *	called and the system has been initialized from the PBS_Server
	// *	database object.  This variable is set to TRUE when the 
	// *	startCycle method is called and then restored to FALSE (0) 
	// *	when the endCycle method is executed.
	// *********************************************************************
	int scheduling;
	
public:
	// *********************************************************************
	// * MatchMatrix::MatchMatrix :
	// *	This is the construct for the Match Matrix object. It does 
	// *	only simple initialization because the bulk of the system
	// *	initialization is performed at the beginning of each scheduling 
	// *	cycle.
	// *********************************************************************
	MatchMatrix ( void );
	
	// *********************************************************************
	// * MatchMatrix::~MatchMatrix :
	// *	This is the destructor for the object.  It will delete any 
	// *	locally allocated memory prior to the deletion of the object.
	// *********************************************************************
	~MatchMatrix ( void );
	
	// *********************************************************************
	// * MatchMatrix::startCycle :
	// *	This method is called at the beginning opf each scheduling 
	// *	cycle.  The method is provdied with a reference to the 
	// *	PBS_Server object which is a database of the current state of 
	// *	the PBS Server.  The data from this object is then used to
	// *	perform the following initializations:
	// *
	// *	1) ClusterNode objects are allocated for every active node in
	// *	   the cluster.
	// *
	// *	2) Running jobs are registered on the nodes/processors where
	// *	   they are executing.
	// *
	// *	This method returns the number of processors that are currently
	// *	free and available for immediate job deployment.  If there are
	// *	no available processors, then it is pointless for the 
	// *	scheduling cycle to be executed.
	// *********************************************************************
	int startCycle ( PBS_Server & server );

	// *********************************************************************
	// * MatchMatrix::endCycle :
	// *	This method is called at the end of the scheduling cycle to free
	// *	any variables or data objects that may be transient between
	// *	scheduling cycles.
	// *********************************************************************
	void endCycle ( void );

	// *********************************************************************
	// * MatchMatrix::processJobs :
	// *	This method is given a sorted list of jobs that are queued and
	// *	ready for immediate deployment.  The processJobs method is then 
	// *	responsible for identifying where each of these jobs may be 
	// *	deployed to get the best fit...  
	// *
	// *	This method will return a list of JobDeployment objects that 
	// *	defines where and when each job should be executed.  This
	// *	JobDeployment object remains the property of the MatchMatrix
	// *	and will be deleted when the endCycle method is called.  If the 
	// *	caller wishes to retain a copy of this object, then the 
	// *	extractJobs method should be used.
	// *
	// *	A new flag has been added to this function: quick
	// *	If the quick flag IS NOT set (is zero) then the method
	// *	will perform a complete inspection and perform match making
	// *	on all jobs... If quick IS set (non-zero), then the
	// *	processJobs method will only run until available nodes are
	// *	depleted.
	// *********************************************************************
	JobDeployment * processJobs ( PBS_Job ** jobs, size_t jobCnt, int quick = 1);

	// *********************************************************************
	// * MatchMatrix::extractJobs :
	// *	This method is used to extract the list of JobCandidate objects
	// *	that has been created by the MatchMatrix.  This list can be
	// *	used for numerous purposes, but is primarlity intended to allow
	// *	the scheduler to perform ordered job deployment and to allow
	// *	the user to query for scheduled execution times between 
	// *	scheduling cycles.
	// *
	// *	Note:  Following this method, the local jobs scheduled object 
	// *	will be set to NULL and the list of JobDeployments will no 
	// *	longer be accessible from this object until more jobs are
	// *	processed.  The caller is responsible for deleteing the list
	// *	of JobDeployment objects that are returned by this method.
	// *********************************************************************
	JobDeployment * extractJobs ( void );
	
	// *********************************************************************
	// * MatchMatrix::available :
	// *	This method will return TRUE if any single processor in the 
	// * 	cluster is available for immediate deployment.
	// *********************************************************************
	int available ( void );
	
	// *********************************************************************
	// * MatchMatrix::availableTime :
	// *	This method will return the longest time slot that is 
	// *	immediately available on any processor in the cluster.
	// *********************************************************************
	time_t availableTime ( void );
	
	// *********************************************************************
	// * MatchMatrix::reserveTime :
	// *	This method is called when the system needs to reserve a block
	// *	of time to prevent it from being utilized by a job.  This method
	// *	will walk through the list of processor objects and reserve
	// *	the specified period of time on each of them.
	// * 
	// *	If no names are specified then time will be reserved
	// * 	on all of the nodes.
	// *********************************************************************
	void reserveTime ( time_t start, time_t duration, char ** nodeNames = NULL, size_t nodeCnt = 0);

	// *********************************************************************
	// * MatchMatrix::asciiDump :
	// *	This method is provided to allow the caller to generate an 
	// *	ascii text dump of the contents of the MatchMatrix object. This
	// *	dump can be used for diagnostic and debugging purposes, or to
	// *	create a record/log of scheduling information.
	// *********************************************************************
	void asciiDump ( FILE * fp = stdout, char * prefix = NULL );
};

#endif /* _MATCH_MATRIX_H_ */
