// -----------------------------------------------------------------------------
//  Copyright (c) 2000     Southeastern Universities Research Association,
//                         Thomas Jefferson National Accelerator Facility
// 
//  This software was developed under a United States Government license
//  described in the NOTICE file included as part of this distribution.
// -----------------------------------------------------------------------------
// 
//  Project Overview: 
// 	This source code was developed as a component of the UnderLord 
// 	Scheduler for the Parallel Batch System - A joint, wide-area computing
// 	initiative at Jefferson Lab and MIT.
// 
//  Description:
// 	RangeDefinition.cc is the source file that defines the methods that
// 	have been declared for the RangeDefinition class.
//
//  Author:  Walt Akers
// -----------------------------------------------------------------------------

#include <RangeDefinition.h>
#include <ClusterNode.h>

// *****************************************************************************
// * RangeDefinition::RangeDefinition :
// * 	This is the constructor for the object.  It must be provided
// *	with the following parameters. 
// *
// *	Duration : The walltime required of the nodes requested.
// *	NNodes   : The number of nodes required.
// *	NProcs   : The number of processors required.
// *	keyTable : The PBS_KeywordTable used to manage keywords.
// *****************************************************************************
RangeDefinition::RangeDefinition 
	( time_t Duration, size_t NNodes, size_t NProcs, PBS_KeywordTable & keyTable )
	: duration ( Duration ),
	  nNodes   ( NNodes ),
	  nProcs   ( NProcs ),
	  keys     ( keyTable )
	{
	}
	
// *****************************************************************************
// * RangeDefinition::RangeDefinition :
// *	This is the constructor for the object.  It will copy the
// *	pertinent contents from an existing RangeDefinition object.
// *****************************************************************************
RangeDefinition::RangeDefinition ( RangeDefinition & def)
	: duration(def.getDuration()),
	  nNodes  (def.getNodesReq()),
	  nProcs  (def.getProcsPerNode()),
	  keys    (def.getKeys())
	{
	}

// *****************************************************************************
// * RangeDefinition::~RangeDefinition :
// *	This is the destructor for the object.  Because this class will
// *	almost always be used as a base for other objects,  this is 
// *	a virtual place holder to allow the child class's constructor
// *	to be called.
// *****************************************************************************
RangeDefinition::~RangeDefinition ( void )
	{
	}
	
// *****************************************************************************
// * RangeDefinition::operator == :
// *	This method will compare the job requirements specified in this 
// *	RangeDefinition with a ClusterNode objects to determine if they
// *	are compatible.
// *****************************************************************************
int RangeDefinition::operator == ( ClusterNode & node)
	{
	return (node.getProcCnt()>=nProcs && PBS_KeySet::compare(keys, node.getKeys())>=0);
	}

// *****************************************************************************
// * RangeDefinition::asciiDump :
// *	This methdo allows the caller to generate an ascii text report of the 
// * 	contents of the RangeDefiniton object.  This data may be used for either
// *	debugging or reporting.
// *****************************************************************************
void RangeDefinition::asciiDump ( FILE * fp, char * prefix )
	{
	char c = 0;
	char * prefixString = (prefix?prefix:&c);
	char * subPrefixString = new char [strlen(prefixString)+13];
	sprintf(subPrefixString, "%s%12c", prefixString, 32);
	
	fprintf(fp, "%sRange Definition :\n", prefixString);
	fprintf(fp, "%s    - Job Duration    : %i:%i:%i\n", prefixString, duration/3600, (duration%3600)/60, duration%60); 
	fprintf(fp, "%s    - Nodes Required  : %i\n", prefixString, nNodes);
	fprintf(fp, "%s    - Processors/Node : %i\n", prefixString, nProcs);
	fprintf(fp, "%s    - Keys :\n", prefixString);	
	keys.asciiDump(fp, subPrefixString);

	delete subPrefixString;
	}
