#include <UnderLord.h>
#include <DurationStage.h>
#include <JobAgeStage.h>
#include <UserShareStage.h>
#include <UserPriorityStage.h>
#include <QueuePriorityStage.h>
#include <signal.h>
#include <unistd.h>

extern "C" {
        #include <sched_cmds.h>
	int die (int);
}

// *****************************************************************************
// * These are global variables that are used by the schedinit and schedule
// * functions.
// *****************************************************************************
UnderLord              * underlord              = NULL;
PBS_AccountingDatabase * serverDB               = NULL;
char                   * serverName             = NULL;
int                      reinitFlag             = 0;
void                   (*oldSIGHUPHandler)(int) = NULL;

// *****************************************************************************
// * HUPHandler :
// *	This is the signal handler that is being installed to catch the HUP
// *	signal.  It will call the original HUP handler and will set a flag
// *	telling my checuler code that it should reinitialize all variables at
// *	the next pass through the schedule function.
// *****************************************************************************
void SIGHUPHandler ( int flag )
	{
	static int inHandler = 0;
	
	if(!inHandler)
		{
		// *************************************************************
		// * Set the local inHandler flag to prevent simultaneous 
		// * re-execution of this code.
		// *************************************************************
		inHandler = 1;

		// *************************************************************
		// * If the oldSIGHUPHandler variable contains a function 
		// * address and the function address is not the same as this 
		// * function, then call it to ensure that the other handler is 
		// * exercised.
		// *************************************************************
		if(oldSIGHUPHandler!=NULL) (*oldSIGHUPHandler)(flag);

		// *************************************************************
		// * Set the reinitFlag to cause the scheduler to be 
		// * reinitialized on the next pass through the schedule
		// * function.
		// *************************************************************
		reinitFlag = 1;
		
		// *************************************************************
		// * Clear the inHandler flag.
		// *************************************************************
		inHandler = 0;
		}
	}

// *****************************************************************************
// * reinitSched :
// *	This function will be called during the next pass through the schedule
// *	function after a SIGHUP has been caught.  It will cause all of the 
// *	system parameters to be reloaded from disk.
// *****************************************************************************
void reinitSched ( void )
	{
	LogRecord ("schedinit", "Initializing scheduler");
		
	// *********************************************************************
	// * Clear any temporarily assigned variables and reload the 
	// * configuration file from disk.
	// *********************************************************************
	PBS_Parameters & params = PBS_Parameters::attach();
	params.reload();
	
	// *********************************************************************
	// * If an underlord object has already been created - delete it so it
	// * will be recreated on the next pass through the schedule event.
	// *********************************************************************
	if(underlord)
		{
		delete underlord;
		underlord = NULL;
		}	

	// *********************************************************************
	// * If a serverDB object has already been created - delete it to force
	// * a reload of the accounting data from the new source.
	// *********************************************************************
	if(serverDB)
		{
		delete serverDB;
		serverDB = NULL;
		}
	
	// *********************************************************************
	// * Reset the reinitFlag to prevent the scheduler from being 
	// * reinitialized unnecessarily during subsequent scheduling cycles.
	// *********************************************************************
	reinitFlag = 0;	
	}

// *****************************************************************************
// * schedinit:
// *	This method deletes all of the objects that are used to manage the 
// *	scheduler so they can be reinitialized on the next pass through
// *	schedule.  Because repopulation of the server database is such a 
// *	protracted process - it is only created once.
// *****************************************************************************
extern "C" int schedinit( int argc, char ** argv)
	{
	// *********************************************************************
	// * Turn on remote logging to force all messages to be transmitted
	// * through the PBS logging functions.
	// *********************************************************************
	SetLogRemote();
	
	// *********************************************************************
	// * Ignore SIGPIPE signal.
	// *********************************************************************
	signal(SIGPIPE, SIG_IGN);
	
	// *********************************************************************
	// * Trap the SIGHUP signal and use it to reinitialize the scheduler.
	// *********************************************************************
	oldSIGHUPHandler = signal(SIGHUP, SIGHUPHandler);
	
	// *********************************************************************
	// * If the serverName has not previously been allocated, then generate
	// * it and use this information to create the server database.
	// *********************************************************************
	if(!serverName)
		{
		FILE * fp  = NULL;
		char * ptr = NULL;
		char   buf[1024];

		if(argc>1) ptr = argv[1];
		else if	((fp = fopen("/usr/spool/PBS/server_name", "r"))!=NULL &&
			(fgets(buf, 1024, fp)!=NULL))
			{
			// Remove the trailing carriage return
			if((ptr = strchr(buf, '\n'))!=NULL) *ptr = 0;
			ptr = buf;
			}
		if(ptr) serverName = strdup(ptr);
		else	{
			LogError("schedinit", "Server name not specified");
			die(0); 
			}
		}

	// *********************************************************************
	// * Call reinitSched to perform the operations that must be performed
	// * each time the scheduler is reinitialized.
	// *********************************************************************
	reinitSched();

	return 0;
	}

// *****************************************************************************
// * schedule: 
// *	This function gets called to start each scheduling cycle It will handle 
// *	the different cases that caused a scheduling cycle.
// *
// *	cmd - reason for scheduling cycle
// * 		SCH_ERROR		error
// *		SCH_SCHEDULE_NULL	NULL command
// *		SCH_SCHEDULE_NEW	A new job was submited
// *		SCH_SCHEDULE_TERM	A job terminated
// *		SCH_SCHEDULE_TIME	The scheduling interval expired
// *		SCH_SCHEDULE_RECYC	A scheduling recycle - this is ignored 
// *                                    because it is meant for a type of 
// *                                    scheduler which only runs one job 
// *                                    at a time. The server will send a
// *                                    recycle command to the scheduler if
// *                                    only one job is run.  There is no reason
// *                                    to run another scheduling cycle if this
// *                                    occurs.
// *		SCH_SCHEDULE_CMD	The server scheduling variable was set 
// *					or reset to true
// *		SCH_SCHEDULE_FIRST	the first cycle after the server starts
// *		SCH_CONFIGURE		perform scheduling configuration
// *		SCH_QUIT		have the scheduler quit
// *		SCH_RULESET		reread the scheduler ruleset
// *
// *	sd - connection descriptor to the pbs server
// *	
// *
// *	returns 0: continue calling scheduling cycles,
// *	returns 1: exit scheduler
// *****************************************************************************
extern "C" int schedule(int cmd, int sd)
	{
	if(reinitFlag) reinitSched();
	
	if(serverDB == NULL)
		{
		LogRecord ("schedule", "Creating Server Database");
		PBS_Parameters & params = PBS_Parameters::attach();
		char * logPath    = (char *)params.getConfigString
			("Scheduler", "accountingDir", "/usr/spool/PBS/server_priv/accounting");
		int    maxHistory = params.getConfigInt
			("Scheduler", "maxHistory", 14);
		int    maxWallHours = params.getConfigInt
			("Scheduler", "maxWallHours", 200);
				
		serverDB = new PBS_AccountingDatabase(serverName, maxHistory, maxWallHours, logPath);
		}
	if(underlord == NULL)
		{
		LogRecord ("schedule", "Creating UnderLord");
		underlord = new UnderLord();
		underlord->addStage(new DurationStage());
		underlord->addStage(new JobAgeStage());
		underlord->addStage(new UserPriorityStage());
		underlord->addStage(new UserShareStage(*serverDB));
		underlord->addStage(new QueuePriorityStage(underlord->getServer(), *serverDB));
		}
	
	switch(cmd)
		{
		case SCH_ERROR:
		case SCH_SCHEDULE_NULL:
		case SCH_RULESET:
		case SCH_SCHEDULE_RECYC:
		break;

		case SCH_SCHEDULE_NEW:
		case SCH_SCHEDULE_TERM:
		case SCH_SCHEDULE_FIRST:
		case SCH_SCHEDULE_CMD:
		case SCH_SCHEDULE_TIME:
		// *************************************************************
		// * Populate the PBS_Server object with the configuration of 
		// * the server, queues, nodes, and jobs.
		// *************************************************************
		underlord->getServer().loadData(sd);
			
		// *************************************************************
		// * Refresh the server database with the most recent data.
		// *************************************************************
		serverDB->update(underlord->getServer());

		// *************************************************************
		// * Call the executive to process any jobs that are due for 
		// * execution.
		// *************************************************************
		underlord->processJobs(sd);
		break;

		case SCH_CONFIGURE:
		break;

		case SCH_QUIT:
		return 1;		
		
		default:
		break;
		}
	return 0;
	}

